IFDEF TEST THEN open Dumper
END

open Common
open Alphabet 
open Transitions
open Term

type automaton= {trans: Transitions.delta; fstates: Stateset.t}

(* mapping a state to the corresponding term *)

type s2t= (state, term) Hashtbl.t

(* mapping a state to the corresponding variable for lhs automaton *)

type statevar = (state,variable) Hashtbl.t

(* Tree automaton copy *)
let copy (a: automaton): automaton=
  {trans= Transitions.copy a.trans;
   fstates= a.fstates}


let new_statevar()= Hashtbl.create Common.max_var 

let new_s2t()= Hashtbl.create Common.max_equation

let statevar_add (s: statevar) (s1: state) (v: variable): unit=
  try(
    if (Hashtbl.find s s1)=v then ()
    else failwith "Automaton:statevar_add: different variable association for a state"
  ) with
      Not_found -> Hashtbl.add s s1 v

let statevar_find = Hashtbl.find
    
let statevar_to_string t=
  List.fold_right (function (k,e) -> function s -> ((State.to_string k)^" -> "^
  (Variable.to_string e))^"\n"^s)
    (hashtbl_to_list t) ""

let s2t_to_string t=
  List.fold_right (function (k,e) -> function s -> ((State.to_string k)^" -> "^
  (Term.to_string e)^"\n"^s))
    (hashtbl_to_list t) ""



(* Run function guided by destination state. 
   if t=f(q1) and dest=q2, then it is NOT covered by a delta where
   f(q3) -> q2 and q1 -> q3. 

   However, if t=f(q1) dest=q2 and delta contains 
   f(q1) -> q3 and q3 -> q2 it is OK.  (The Transitions.get_trans is modulo epsilon)
*)

let rec oriented_run (t: term) (dest: state) (d: delta): bool=
(*  let _= print_string ("For t="^(Term.to_string t)^"\nwith delta="^(Transitions.to_string d)^"\n") in *)
  match t with
      Fsym(f,tl) -> 
	let all_trans= Transitions.get_trans dest f d in
	  List.exists 
	    (function (_,sl,_) -> oriented_run_list tl sl d)
	    all_trans
    | State(q) -> q=dest (* || (Transitions.is_pred q dest d)*)
    | Var(_) -> failwith "Automaton:oriented_run: non ground term!"

and oriented_run_list (tl: term list) (destl: state list) (d: delta): bool=
   match tl, destl with
       [],[] -> true
     | t::rem1, dest::rem2 -> 
	 (oriented_run t dest d) && (oriented_run_list rem1 rem2 d)
     | _,_ -> failwith "Automaton:oriented_run_list: size of lists is not the same!"



(* DEPRECATED? use norm_term and norm_term_list *)
(* Deterministic run: used for prior normalisation (no epsilon resolving) *)
(* chooses the first found state to normalize a transition *)

let rec det_run (t: term) (d: delta): term=
  match t with
      Fsym(f,[]) -> 
	(let relevant_trans= Transitions.get_trans_sym_constant f d in
	   match relevant_trans with
	       [] -> t
	     | (_,_,q)::_ -> State(q))
    | Fsym(f,tl) -> 
	(let res_tl= det_run_list tl d in
	   try(
	     let res_states= List.map Term.get_state res_tl in
	     let relevant_trans= 
	       List.flatten (List.map (function q-> Transitions.get_trans_sym_parameter f q d)
			       res_states) in
	       try(
		 State(match (List.find (function  (_,sl,_) -> sl=res_states) relevant_trans)
		       with (_,_,q) -> q)
	       )
	       with (* No transition matches *)
		   Not_found -> Fsym(f, res_tl)
	   )
	   with (* Transition is not fully normalizable *)
	       Term.Not_a_state -> Fsym(f, res_tl))


    | State(q) -> State(q) 
    | Var(_) -> failwith "Automaton:oriented_run: non ground term!"

and det_run_list (tl: term list) (d: delta): term list=
   match tl with
       [] -> []
     | t::rem1 -> 
	 (det_run t d)::(det_run_list rem1 d)


(* Non deterministic run: use any transition to normalize a term (no epsilon resolving) *)
(* gives every possible state in which a term can be rewritten or the empty list *)
(* Principle is to find all transitions rooted by the same top symbol (all
   f(q1...qn)->q for f(t1...tn) and then removing parameter states q...qn of transitions
   one after one, and keep only transitions such that they were matching t1...tn successively *)

let rec all_run (t: term) (d: delta): state list=
(*   let _= print_string ("Term= "^(Term.to_string t)^"\non delta\n") in *)
  match t with
      Fsym(f,[]) -> 
	(let relevant_trans= Transitions.get_trans_sym_constant f d in
	   match relevant_trans with
	       [] -> []
	     | l -> List.map (function  (_,_,q) -> q) l)
    | Fsym(f,tl) -> 
	let trans_for_f= Transitions.simple_get_trans_f f d in
	  all_run_list tl trans_for_f d
    | State(q) -> [q] 
    | Var("_") -> failwith "Automaton:all_run: anonymous variable cannot be used on top!"
    | Var(_) -> failwith "Automaton:all_run: free variable uninstantiated!"

and all_run_list (tl: term list) (l: external_symtrans list)(d: delta): state list=
(*  let _= print_string ("Term list= "^(String.concat ", " (List.map Term.to_string tl))^"\non l="^(Transitions.symextlist_to_string l)^"\n") in *)
  match tl with
      [] -> List.map (function (_,_,dest) -> dest) l (* all remaining trans are ok *)
    | Var("_")::rem -> (* anonymous variables can match any state i.e. keep all trans *)
	all_run_list rem (List.map (function (f,sl,dest) -> (f,List.tl sl,dest)) l) d
    | Var(x)::rem -> failwith ("Automaton:all_run: non ground term: variable "^x)
    | State(q)::rem ->
	all_run_list rem (List.fold_right 
			    (function (f,sl,dest) -> function l2 -> 
			       if List.hd sl=q then (f,List.tl sl,dest)::l2
			       else l2) l []) d
    | t::rem -> let sl= all_run t d in
	all_run_list rem (List.fold_right
			    (function (f,sl2,dest) -> function l2 ->
			       if List.mem (List.hd sl2) sl then (f,List.tl sl2,dest)::l2
			       else l2) l []) d


(* Normalisation of a term for adding in delta, during completion (flag variable false) *)
(* Also: normalisation of a term to produce a term_automaton: for left-hand
   side of rules, or initial term sets (flag variable true) *)

(* Add a term to a delta and returns the top state of this term 
   as well as the list of new external transitions resulting in
   this adding. Uses d to normalize t and adds transitions for t to d. Also  
   creates new states in d when necessary.

   The boolean wvar is a flag which is true when normalizing terms with
   variables (left-hand side of rules) *)

let rec normalize_term (t: term) (d: Transitions.delta) (wvar: bool): (state * external_trans list)=
  (match t with
       Fsym(f,[]) -> 
	 (let relevant_trans= Transitions.get_trans_sym_constant f d in
	    match relevant_trans with
		[] -> 
		  let s= Transitions.new_state d in
		  let _= Transitions.add (Trans(f,[],s)) d in
		    (s, [Trans(f,[],s)])
	      | (_,_,q)::_ -> (q,[]))
     | Fsym(f,tl) -> 
	 (let (res_state_list,translist)= normalize_term_list tl d wvar in
	    try(
	      let relevant_trans= 
		List.flatten (List.map (function q-> Transitions.get_trans_sym_parameter f q d)
				res_state_list) in
		try(
		  match (List.find (function  (_,sl,_) -> sl=res_state_list) relevant_trans)
		    
		  with (_,_,q) -> (q,translist)  (* A transition matches, no new state
						    is necessary *)
		)
		with (* No transition matches, a new state and a new transition are necessary *)
		    Not_found -> 
		      let s= Transitions.new_state d in
		      let _= Transitions.add (Trans (f,res_state_list,s)) d in
			(s, (Trans(f, res_state_list,s))::translist)
	    )
	    with (* Transition is not fully normalizable *)
		Term.Not_a_state -> failwith "Automaton:normalize_term: Subterms are not fully normalized!")
     | State(q) -> 
	 if wvar then failwith "Automaton:normalize_term: cannot be a state"
	 else (q, [])
     | Var(x) -> 
	 if wvar
	 then (* For variables, in the case where we normalize a lhs
	         we simply add a new constant (and state) if it is
	         not already present *)
	   (let relevant_trans= Transitions.get_trans_sym_constant x d in
	      match relevant_trans with
		  [] -> 
		    let s= Transitions.new_state d in
		    let _= Transitions.add (Trans(x,[],s)) d in
		      (s, [Trans(x,[],s)])
		| (_,_,q)::_ -> (q,[]))
	 else failwith "Automaton:normalize_term: cannot be a variable!")

and normalize_term_list (tl: term list) (d: delta) (wvar: bool): (state list * (external_trans list))=
   match tl with
       [t] -> let (s, tl)= normalize_term t d wvar in
	 ([s], tl)
     | t::rem1 -> 
	 let (s, tl)= normalize_term t d wvar in
	 let (sl2, tl2)= normalize_term_list rem1 d wvar in
	   (s::sl2, List.append tl tl2)
     | [] -> failwith "Automaton:normalize_term_list: empty list!"


(* Normalisation of a term for adding in delta, using only new states! *)

let rec normalize_new_states (t: term) (d:Transitions.delta): (state * external_trans list)=
  (match t with
       Fsym(f,[]) -> 
	 let s= Transitions.new_state d in
	 let _= Transitions.add (Trans(f,[],s)) d in
		    (s, [Trans(f,[],s)])
     | Fsym(f,tl) -> 
	 let (res_state_list,translist)= normalize_new_states_list tl d in
	 let s= Transitions.new_state d in
	 let _= Transitions.add (Trans (f,res_state_list,s)) d in
	   (s, (Trans(f, res_state_list,s))::translist)
     | State(q) -> (q, [])
     | Var(x) -> failwith "Automaton:normalize_term: cannot be a variable!")

and normalize_new_states_list (tl: term list) (d: delta): (state list * (external_trans list))=
   match tl with
       [t] -> let (s, tl)= normalize_new_states t d in
	 ([s], tl)
     | t::rem1 -> 
	 let (s, tl)= normalize_new_states t d in
	 let (sl2, tl2)= normalize_new_states_list rem1 d in
	   (s::sl2, List.append tl tl2)
     | [] -> failwith "Automaton:normalize_term_list: empty list!"



(* Normalisation of a term for adding in delta, during completion (flag variable
   false. This corresponds to oldstyle normalization, i.e. normalize f(t1...tn)
   ->q such that the top transition is of the form f(q1...qn) -> q. *)

let oldnormalize_term (newstates: bool)(t: term)(dest: state)(d: Transitions.delta) : external_trans list=
  match t with
      Fsym(f,[]) -> 
	if (fst (Transitions.add (Trans(f,[],dest)) d)) 
	then [Trans(f,[],dest)]
	else [] 
    | Fsym(f,tl) -> 
	let (res_state_list,translist)= 
	  if newstates then normalize_new_states_list tl d
	  else normalize_term_list tl d false 
	in
	  if (fst (Transitions.add (Trans(f,res_state_list,dest)) d))
	  then Trans(f,res_state_list,dest)::translist
	  else translist
    | State(q) -> [Epsilon(q,dest)]
    | Var(x) -> failwith "Automaton:oldnormalize_term: cannot be a variable!"








let term_to_automaton (tl: term list) (vl: variable list) (stt: s2t): (automaton * statevar * state list)=
  let rec aux (tl: term list) (a: automaton) (var_assoc: statevar) (fs: state list): 
    (automaton * statevar * state list)=
    match tl with 
	[] -> (a,var_assoc,List.rev fs)
      | t::rem ->
	  let (final,trans)= normalize_term t a.trans true in
	  let _= Hashtbl.add stt final t in
	  let _= List.iter (function t ->
			      match t with 
				  (Trans(f,_,dest)) -> 
				    if List.mem f vl 
				    then statevar_add var_assoc dest f
				    else ()
				| _ -> failwith "Automaton:term_to_automaton: transitions cannot be epsilon here!") trans 
	  in
	    aux rem {trans= a.trans; fstates= Stateset.add final a.fstates} var_assoc (final::fs)
  in
    aux tl {trans=Transitions.new_delta(); fstates= Stateset.empty} (new_statevar ()) []
      
   let state_list_to_string l= 
     List.fold_right (function e ->function s -> ((State.to_string e)^" "^s)) l ""
       
   let to_string (a: automaton): string=
     ("States "^(state_list_to_string (Stateset.elements (Transitions.get_all_states a.trans)))^"\nFinal States "^
      (state_list_to_string (Stateset.elements a.fstates))^"\nTransitions\n"^(Transitions.to_string a.trans))


   (* Merging on a tree automaton, does not produce a new one! *)
(* modifies mt *)

let merge (a: automaton) (mt: state_merge_table) (merge_list: (state * state) list):
  automaton * ((state list) * state) list=
  let merging_list= Transitions.merge a.trans mt merge_list in
  let new_aut= 
    {trans= a.trans;
     fstates= Transitions.stateset_map (function state -> Transitions.merged_value mt state) a.fstates} in
    (new_aut,merging_list)

      
let unepsilon (a: automaton): automaton=
  let new_trans= Transitions.propagate_epsilon 
		   (List.map (function t -> match t with Trans(t2) -> t2
			       | _ -> failwith "Automaton:unepsilon:Cannot be an Epsilon trans") (Transitions.trans_to_list a.trans.Transitions.trans))
		   a.trans []
  in
  let new_delta= Transitions.new_delta() in
    begin
      List.iter (function t -> ignore(Transitions.add_regular_trans t new_delta)) new_trans;
      {trans= new_delta; 
       fstates= a.fstates}
    end

    (* Merges all symtables of l into acc *)

let rec merge_symtable (l:(symbol*table)list) (acc: symtable): unit=
  match l with 
      [] -> ()
    | (f,t)::rem -> 
	let _= 
	  try (let table= Hashtbl.find acc f in
		 Hashtbl.replace acc f (Statetable.union t table)
	      )
	  with Not_found -> Hashtbl.add acc f t 
	in
	  merge_symtable rem acc
		   
      
(* Checks if there is at least one common symbol between list l1 and l2 *)

let rec matchone_symbol (l1: (symbol*table)list) (t2: symtable)=
  match l1 with 
      [] -> false
    | (f1,_)::rem ->
	(Hashtbl.mem t2 f1)||(matchone_symbol rem t2)


(* Filters a list of (symbol->table) s.t. the first list contains only 
   constants and the second one only non constant transitions *)

let rec get_const (l: (symbol * table) list) (resconst: (symbol * table) list) (resnotconst: (symbol * table) list): ((symbol * table) list)*((symbol * table) list)=
  match l with 
      [] -> (resconst,resnotconst)
    |(f,t)::rem ->
       if (Transitions.constant_table t) 
       then get_const rem ((f,t)::resconst) resnotconst
       else get_const rem resconst ((f,t)::resnotconst) 


(* Check if there is at least one common constant symbol between l1 and t2 *)
(* Symbols of l1 are only constants *)

let rec fast_const_test (l1: (symbol*table)list)(t2: symtable): bool=
  match l1 with
      [] -> false
    | (s,t)::rem -> 
	(Hashtbl.mem t2 s)||(fast_const_test rem t2)


(* Ordering couples so that the smaller becomes the first one *)

let couple_order s1 s2=
  if compare s1 s2 >0 then (s2,s1)
  else (s1,s2)
  
let mem_couple s1 s2 t=
  let (c1,c2)= couple_order s1 s2 in
    try (Stateset.mem c2 (Hashtbl.find t c1))
    with Not_found -> false

let add_couple s1 s2 t=
  let (c1,c2)= couple_order s1 s2 in
    try (let setc1= Hashtbl.find t c1 in
	   Hashtbl.replace t c1 (Stateset.add c2 setc1))
    with Not_found -> Hashtbl.add t c1 (Stateset.singleton c2)


(* test of non emptyness of intersection of a list of couple of states
   of a single tree automaton  ** modulo epsilon ** *)
(* input list [(q1,q2);(q3,q4)] checks if q1/\q2 is not empty AND
  q3/\q4 not empty *)
(* Terminates also on cycling empty tree automata *)

let not_empty_intersection (l: (state * state) list)(delta: Transitions.delta): bool=
    (* table for inductive couple of states (for intersection of 
       recursive transitions *)
  let inductive= Hashtbl.create Common.max_state_symb in
  let ok= Hashtbl.create Common.max_state_symb in 
  let deltatable= delta.Transitions.trans in
  let rec check_state_list (lq1: state list) (lq2:state list)=
    match lq1,lq2 with 
	[],[] -> true
      | q1::rem1, q2::rem2 -> (check_state q1 q2)&&(check_state_list rem1 rem2)
      | _,_ -> failwith "Automaton:not_empty_intersection: state lists should have the same length"
  and check_state q1 q2=
    if mem_couple q1 q2 ok then true
    else
    if mem_couple q1 q2 inductive 
      (* Couple has already been considered we are now dealing with
	 recursive transitions which are of no help for checking emptyness of an
	 intersection*)
    then false
      (* Otherwise, it has to be investigated more deeply *)
    else
      if (q1=q2)||(is_pred q1 q2 delta)||(is_pred q2 q1 delta) 
      then let _= add_couple q1 q2 ok in true
      else 
	(* We are considering all predecessors of q1 and q2 in order to 
	   cover epsilon closure *)
	let allq1= (Transitions.get_pred q1 delta) in
	let allq2= (Transitions.get_pred q2 delta) in
	let tableq1= Hashtbl.copy (Hashtbl.find deltatable q1) in
	let tableq2= Hashtbl.copy (Hashtbl.find deltatable q2) in
	let ltables1= (List.flatten (List.map 
				       (function q -> 
					  (hashtbl_to_list (Hashtbl.find deltatable q))) allq1)) in
	let ltables2= (List.flatten (List.map 
				       (function q -> 
					  (hashtbl_to_list (Hashtbl.find deltatable q))) allq2)) in
	let _= add_couple q1 q2 inductive in
	let _= merge_symtable ltables1 tableq1 in
	let _= merge_symtable ltables2 tableq2 in
	let listallq1= hashtbl_to_list tableq1 in

	  (* Checks if there is at least one common symbol between list l1 and
	     l2 for efficiency reasons *)

	  if matchone_symbol listallq1 tableq2
	  then (* there is at least one common symbol *)
	    let (const,nonconst)= get_const listallq1 [] [] in
	      if ((fast_const_test const tableq2) || (check_transition_list nonconst tableq2))
	      then let _= add_couple q1 q2 ok in true
	      else false
	  else (* there is no common symbol *)
	    false
	      
  and check_transition_list (l1: (symbol*table)list)(t2: symtable): bool=
    match l1 with
	[] -> false
      | (f,t)::rem  ->
	  try (let tab2= Hashtbl.find t2 f in
		 (* Afraid of what? :-) 
		    Check that there EXISTS a statelist1 of t s.t.
		    the intersection with statelist2 is not empty.
		    To check if intersection between statelist1=(q1,q2)
		    and statelist2=(q3,q4) is not empty we need that
		    intersection between q1 and q3 and q2 and q4 is not
		    empty *)
		 
		 (Statetable.exists 
		    (function statelist1 -> 
		       Statetable.exists 
		       (function statelist2 ->
			  check_state_list statelist1 statelist2)
		       tab2)
		    t) || (check_transition_list rem t2))
	  with Not_found -> (check_transition_list rem t2)
  in
  let (lq1,lq2)= List.split l in
    check_state_list lq1 lq2
      


(* ------------------------------------------------ Tests ----------------------------------------------*)


IFDEF TEST THEN print_string "Automaton.ml\n\n" END

IFDEF TEST THEN
  let stt= new_s2t() 
  let stt2= new_s2t()
  let t1= Fsym("f", [Fsym("a",[]);Fsym("g", [Var("x"); Var("y")])])
  let (aut1,sv,_)= term_to_automaton [t1] ["x";"y"] stt
  let t2= Fsym("f", [Fsym("a",[]);Fsym("g", [Fsym("f",[Fsym("a",[]);Fsym("b",[])]);Fsym("b",[])])])
  let t3= Fsym("f", [Fsym("g",[Fsym("b",[])]);Fsym("g",[Fsym("b",[])])])
  let t4= Fsym("f", [Var("_");Fsym("g",[Fsym("b",[])])])
  let (aut1,sv,_)= term_to_automaton [t1] ["x";"y"] stt
  let (aut2,_,_)= term_to_automaton [t2] [] stt
  let sv2= Hashtbl.create Common.max_var
  let _= (Hashtbl.add sv2 1 "x") 
  let _= (Hashtbl.add sv2 2 "y") 
  let (aut4,_,_)= term_to_automaton [t2] [] stt
  let mt= Transitions.new_merge_table ()
  let (aut5,_,ls)= term_to_automaton [t1;t2] ["x";"y"] stt2
END

IFDEF TEST THEN
begin
  print_string ("Testing state to term table and printing\n");
  print_string (s2t_to_string stt2);
  asserteq (s2t_to_string stt2) "q8 -> f(a,g(f(a,b),b))\nq4 -> f(a,g(x,y))\n" "";
  print_string ("d="^Transitions.to_string Transitions.d^"\n\n");
  print_string "f(a, g(a,a)) -> q1\n";
  asserteq (oriented_run Term.tn Transitions.s1 Transitions.d) true "";
  print_string "f(g(a,a),g(a,a)) -/-> q1\n";
  asserteq (oriented_run Term.tm Transitions.s1 Transitions.d) false "";
  print_string "det_run f(a,g(a,a)) --> f(q0,g(q0,q0))\n";
  asserteq (det_run Term.tn Transitions.d) Term.tq0 "";
  print_string "Add f(g(a,a),g(a,a)) to q1 in d\n";
  let (sl,_)= normalize_term Term.tm d false in
  let d2= Transitions.new_delta () in
    print_string ("New top state="^(Transitions.state_list_to_string [sl])^"\n\n");
    print_string ("New d=\n"^(Transitions.to_string d)^"\n");
    print_string "now, det_run f(g(a,a),g(a,a)) --> ok\n";
    asserteq (det_run Term.tm Transitions.d) (State sl) "";
    print_string ("We create a tree automaton for "^(Term.to_string Term.tmx)^"\n");
    let _= normalize_term Term.tmx d2 true in
      print_string (Transitions.to_string d2);
      print_string ((to_string aut1)^"\nshould be equal to\n"^"States q0 q1 q2 q3 q4 \nFinal States q4 \nTransitions\ny -> q2\nx -> q1\nf(q0,q3) -> q4\na -> q0\ng(q1,q2) -> q3\n\n");
      asserteq (to_string aut1) "States q0 q1 q2 q3 q4 \nFinal States q4 \nTransitions\ny -> q2\nx -> q1\nf(q0,q3) -> q4\na -> q0\ng(q1,q2) -> q3\n\n" "";
      asserteq (Common.hashtbl_equal_level1 sv sv2 (=)) true "statevar tables";
      print_string ("Before merge aut4 ="^(to_string aut4)^"\n");
      let (new_aut4,_)= merge aut4 mt [1,0;0,4;3,1;3,2] in
	print_string ("After merge 1=0 0=4 3=1 3=2, aut4 becomes= "^(to_string new_aut4)^"\n");
	asserteq (to_string new_aut4) "States q4 \nFinal States q4 \nTransitions\nb -> q4\ng(q4,q4) -> q4\na -> q4\nf(q4,q4) -> q4\n\n" "";
	print_string ("Checking order of states in state list of: "^(to_string aut5)^"\n");
	asserteq ls [4;8] "Order of final states should be the same as the order of terms in the list of input"
end
END

IFDEF TEST THEN
  let aut10={trans= Transitions.new_delta (); fstates=Stateset.empty}
  let s0= new_state aut10.trans 
  let s1= new_state aut10.trans 
  let s2= new_state aut10.trans 
  let s3= new_state aut10.trans 
  let s4= new_state aut10.trans 
  let s5= new_state aut10.trans 
  let s6= new_state aut10.trans
  let s7= new_state aut10.trans
	    (* In the following automaton f(g(a),g(a)) is in the intersection 
	       between states q0 and q4
	       
	       f(q0,q0) -> q0
	       g(q1) -> q0 
	       b -> q1
	       a -> q3
	       q3 -> q1
	       g(q6) -> q6

	       f(q6, q5) -> q4
	       g(q4) -> q6
	       g(q4) -> q5
	       g(q7) -> q7
	       a -> q4
	       b -> q4 *)

  let _= ignore (Transitions.add (Trans ("f",[s0;s0],s0)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("g",[s1],s0)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("g",[s6],s6)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("b",[],s1)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("a",[],s3)) aut10.trans)
  let _= ignore (Transitions.add (Epsilon (s3,s1)) aut10.trans)

  let _= ignore (Transitions.add (Trans ("f",[s6;s5],s4)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("g",[s4],s6)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("g",[s4],s5)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("g",[s7],s7)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("a",[],s4)) aut10.trans)
  let _= ignore (Transitions.add (Trans ("b",[],s4)) aut10.trans)
  let ind= Hashtbl.create Common.max_state_symb 
END

IFDEF TEST THEN
begin
  print_string "all_run\n";
  print_string ("On aut10="^(Transitions.to_string aut10.trans)^"\n");
  asserteq (all_run t3 aut10.trans) [0;4] "f(g(b),g(b)) is recognized by two states q0 and q4";
  asserteq (all_run t4 aut10.trans) [0;4] "f(_,g(b)) is recognized by two states q0 and q4";
  print_string "not_empty_intersection\n";
  asserteq (mem_couple 1 2 ind) false "1,2 not present in empty table";
  add_couple 1 2 ind;
  add_couple 3 1 ind;
  asserteq (mem_couple 2 1 ind) true "1,2 with order";
  asserteq (mem_couple 1 2 ind) true "1,2 without order";
  asserteq (mem_couple 1 3 ind) true "1,3 without order";
  asserteq (mem_couple 3 1 ind) true "1,3 with order";
  asserteq (not_empty_intersection [(s0,s4)] aut10.trans) true "intersection not empty";
  asserteq (not_empty_intersection [(s4,s0)] aut10.trans) true "should work in both directions";
  asserteq (not_empty_intersection [(s1,s5)] aut10.trans) false "intersection empty";
  asserteq (not_empty_intersection [(s5,s1)] aut10.trans) false "other direction";
  asserteq (not_empty_intersection [(s6,s7)] aut10.trans) false "check termination of intersection on cycling empty langages"
end
END

