IFDEF TEST THEN open Dumper 
END
open Common
open Set2
open Alphabet 
open Transitions
open Term
open Substitution
open Automaton
open Rewriting
open Equation

type substitution=Substitution.t
type substitutionset= SubstitutionSet.t
type substable= SubstitutionTable.t

(* condition_tables map states of lhs of conditional rule to sets of qredex states
   we reconsider every conditional rule at every completion step *)

type condition_table= (state, stateset) Hashtbl.t


(* in exteq_table each qlhs of a pattern is associated to either Unique (the 
pattern is alone [f(x)] -> ... or Eq like [f(x)=g(y)] or Any like [f(x),g(x)]
in that case the state gives the qrhs. The term*term list is the list of 
merging equations *)

type exteq= Unique of ((term * term) list )
		| Eq of (state * ((term * term) list))
		| Any of (state *((term * term) list))


let condition_table_to_string t=
  List.fold_right (function (k,e) -> function s -> 
		     ((State.to_string k)^" -> "^
		      Stateset.to_string e)^"\n"^s)
    (hashtbl_to_list t) ""  


(* Propagation of new substitutions should apply modulo epsilon *)
(* Propagation of epsilon should apply on substitutions! and propagate upwards
   to final states of rules (if possible) *)
(* For instance if (qg,q2) {{x->q1}} and q2 -> q3 by epsilon then
   (qg, q3) should contain at least {{x->q1}} *)


(* Match one returns the set of substitutions corresponding to the 
   matching of state q1 on q2 with substable s and variable list vl (states
   recognizing variables in lhs) *)


let match_one (q1: state) (q2: state) (s: substable) (vl: statevar): substitutionset=
  try( (* If q1 is a variable *)
    let varname= Automaton.statevar_find vl q1 in
      SubstitutionSet.singleton (Substitution.add varname (State q2) Substitution.empty) 
  ) (* q1 is a regular state... so we look for solutions of (q1,q2) in s*)
  with Not_found -> 
    try(
      SubstitutionTable.find (q1,q2) s 
    ) (* Otherwise the set of solution is empty *)
    with SubstitutionTable.Couple_not_found -> SubstitutionSet.empty



(* filter_set returns for a list of transitions of lhs and a new trans the list
   association (q1,q2)->new set of substitutions to be added to (q1,q2).
   vl is the variable (given as states) list *)

let rec filter_set (l: external_symtrans list) ((f,sl,dest): external_symtrans) (s: substable) (vl: statevar) (res:((state*state)*substitutionset)list): ((state*state)*substitutionset)list=
  match l with
      [] -> res
    | (_,sl2,dest2)::rem ->
	(let value_list=
	  List.map2 (function q1 -> function q2 -> match_one q1 q2 s vl) sl2 sl
	in
	let new_subst= ((dest2,dest), SubstitutionSet.combine value_list) in
	  filter_set rem (f,sl,dest) s vl (new_subst::res))

(* Filtering function for a term automaton lhs (gathering all left-hand sides of
   rules) and a tree automaton a representing a regular language of
   terms. Does not compute the product automaton. vl is
   the list of (state*variable) variables in lhs. nt is the list of new transitions. 
   ns is the list of new substitution by couple of states produced by nt. *)

(* Attention aux constantes ex: mb gauche g(a) ... subst vide *)

(* This function does not propagate new substitutions to fathers etc. *)
(* This function does not modify s!!! *)

(* The iseq flag is there only to use unoptimized new substitution production in case
   of equations: a transition may need to be reconsidered even if corresponding substitutions
   are already in the substable because they may be present there because of merging. 
   They do not need to be reconsidered for TRS because merging of substitutions also
   applies on rhs hence new solutions, but it is necessary for equation because it may lead
   to new equation applications. See TU-ex7.txt.
*)

let rec filter (lhs: automaton) (nt: external_symtrans list) (ns: ((state*state)*substitutionset)list)(s: substable) (vl: statevar)(iseq:bool): ((state*state)*substitutionset) list= 
  match nt with
      [] -> ns
    | (f,sl,dest)::rem ->
	if sl=[] 
	then (* Constant symbols *)
	  (* There should be at most one transition of the form a->qa in the lhs.
	     not modulo epsilon! *)
	  let destlhs= Transitions.get_trans_sym_constant f lhs.trans in
	    match destlhs with
		[] -> (* There is no transitions labelled by f in lhs *)
		  filter lhs rem ns s vl iseq
	      | [(_,[],dest2)] ->  
		  (* There is exactly one transition labelled by f in lhs *)
		  if SubstitutionTable.mem (dest2,dest) Substitution.empty s 
		  then 
		    (* Found substitution is not new *)
      		    filter lhs rem ns s vl iseq
		  else (* Found substitution is new *)
		    filter lhs rem (((dest2,dest),(SubstitutionSet.singleton Substitution.empty))::ns) s vl iseq
	      | _ -> (* Any other case is impossible *) 
		  failwith "Completion:filter:malformed lhs constant automaton"

	else (* General transitions *)
	  (* Get all the transitions of lhs with f as top symbol. Not modulo epsilon *)
	  let all_ftrans= Transitions.simple_get_trans_f f lhs.trans in
	    
	  (* Get all the substitutions producted by the new transitions *)

	  let new_substs= filter_set all_ftrans (f,sl,dest) s vl [] in
	    
	  (* Filter-out those who are already covered by s, for each es (set of 
	     substitution) retrieve in es all the substs that are already in s
	  (done by find and diff *)

(*	Necessary to reconsider all substitutions
       for filtering correctly on fixpoint_merging for equations
       hence we remove the optimization: 
*)	
	  let new_not_covered= if iseq then new_substs else
	    List.fold_left 
	      (function l -> function (couple,es) ->
		 let newset= 
		   try (
		     SubstitutionSet.diff es (SubstitutionTable.find couple s))
		   with SubstitutionTable.Couple_not_found -> es in
		   if SubstitutionSet.is_empty newset then l
		   else (couple, newset)::l)
	      []
	      new_substs in 
	    filter lhs rem (List.append new_not_covered ns) s vl iseq



(* Filter_epsi propagate the q1 -> q2 transitions. This is necessary when an epsilon
   transition q1 -> q2 appear some steps after that q1 has been filled. Thanks to
   this epsilon transition a NEW substitution may occur in q2 and may thus be declared.
   It will be propagated further into by propagation if q2 ->* q3 
*)

(* This function does not propagate new substitutions to fathers etc. *)
(* This function does not modify s!!! *)

let rec filter_epsi (ne: external_epsilon list) (ns: ((state*state)*substitutionset)list)(s: substable): ((state*state)*substitutionset) list= 
  match ne with
      [] -> ns
    | (q1,q2)::rem ->
	(* We add all substs of q1 to q2 *)
	let res= 
	  try (let all_new_sub= Common.hashtbl_to_list (Hashtbl.find s q1) in
		 List.fold_left (function l -> function (q,es) ->
			      let new_es= SubstitutionTable.add_set es (q,q2) s in
				if (SubstitutionSet.is_empty new_es) then l
				else ((q,q2),new_es)::l)
		   ns all_new_sub)
	  with Not_found -> ns in
	  filter_epsi rem res s


(* Pour Propagate: Hypothese faite ici l'ordre dans lequel les external trans arrivent n'importent pas 
pour la coherence des substitutions calculees par intersection locale puis propagation.
Ex: 

f(qg, qf2) -> qf1		f(q1, q3) -> q0
g(qvg1) -> qg 			g(q2) -> q1
f(qvf1, qvf2) -> qf2		f(q5, q6) -> q3

(qg,q1)={{qvg1->q2}}

On suppose que new1=f(q1,q3)->q0 et new2=f(q5,q6) -> q3

Si on ajoute new1 avant new2: (qf,q0) reste vide (pas de subst possible), et 
en ajoutant new2 + propage on obtient: (qf2,q3)={{qvf1->q5,qvf2->q6}} et en 
propageant au dessus: (qf,q0)={{qvg1->q2,qvf1->q5,qvf2->q6}}

Ce qui est identique a ce que l'on obtiendrait par l'operation inverse: new2 avant
new1.

par contre, il faut garder en tete que les nouvelles epsilon, ex: q1 -> q2
peuvent ajouter des substitutions a q2 (ET TOUS SES DESCENDANTS!) et il faut
repropager a partir de ces nouvelles substitutions. *)


let rec match_in_list l1 l2 e1 e2: bool=
  match (l1,l2) with
      ([],[]) -> false
    | (elt1::rem1, elt2::rem2) -> ((elt1=e1)&&(elt2=e2)) || (match_in_list rem1 rem2 e1 e2)
    | _,_ -> failwith "Completion:match_in_list: lists do not have the same size!"


(* For a given symbol f, destination states qlhsdest (lhs) and qdest (redex), 
   corresponding tables tlhs and t2 (redex), and parameter states qlhs and q (redex) 
   this function produces the list of external transitions such that at least one
   position of qlhs corresponds to one position of q *)


let intersec_by_table (f: symbol) (qlhsdest: state) (tlhs: table) (qlhs: state) 
    (q: state) (qdest: state) (t2: table) (res: (external_symtrans*external_symtrans)list): (external_symtrans*external_symtrans)list=

  Statetable.fold 
    (function elt1 -> function l1 -> 
       (Statetable.fold 
	  (function elt2 -> function l2 ->
	      if match_in_list elt1 elt2 qlhs q
	      then ((f,elt1,qlhsdest),(f,elt2,qdest))::l2
	      else l2)
	  t2
	  l1))
    tlhs
    res

let rec intersec_by_table_list f qlhsdest t qlhs q lred res=
  match lred with 
      [] -> res
    | (qdest,t2)::rem -> 
	intersec_by_table_list f qlhsdest t qlhs q rem (intersec_by_table f qlhsdest t qlhs q qdest t2 res)


(* intersec_by_dest computes the list of couples of transition lhs*redex where the
   top symbol is f and the position of qlhs in lhs trans is the
   same a q in redex trans*)

let rec intersec_by_dest (f: symbol) (qlhs: state) (q: state) (llhs: (state*table) list) (lred: (state*table) list) (res: (external_symtrans*external_symtrans)list): (external_symtrans*external_symtrans)list=
  match llhs with 
      [] -> res
    | (q2,t)::rem ->
	intersec_by_dest f qlhs q rem lred (intersec_by_table_list f q2 t qlhs q lred res)
	

(* intersec computes the list of couples of transition lhs*redex where the
   top symbol is the same and the position of qlhs in lhs trans is the
   same a q in redex trans*)

let rec intersec (qlhs: state) (q: state) (llhs: (symbol * (state , table) Hashtbl.t)list)  (sd: satable) (res: (external_symtrans * external_symtrans) list): (external_symtrans * external_symtrans) list=
  match llhs with
      [] -> res
    | (f,t)::rem ->
	try(
	  let lred_by_dest= (Common.hashtbl_to_list (Hashtbl.find (Hashtbl.find sd q) f)) in
	  let llhs_by_dest= Common.hashtbl_to_list t in
	    intersec qlhs q rem sd (intersec_by_dest f qlhs q llhs_by_dest lred_by_dest res)
	) with Not_found -> intersec qlhs q rem sd res


(* Propagation of new substitution w.r.t. list of couples of transitions *)

(* For propagation on father transitions, it is enough to propagate new
   substitutions of (qlhs,q) with old substitution of other states.
   It is possible that several occurences of (qlhs,q) occur in ls1 ls2 but
   since rules are linear qlhs cannot contain a variable and thus it it not
   necessry to combine old solutions of (qlhs,q) with new ones (it can only
   be the trivial substitution. *)

let rec propagate_one_layer (linter: (external_symtrans * external_symtrans) list)
  (qlhs: state) (q: state)(es: substitutionset) (s: substable) (vl: statevar)(res: ((state * state) * substitutionset) list):((state * state) * substitutionset) list=
  match linter with
      [] -> res
    | ((f,ls1,dest1),(_,ls2,dest2))::rem ->
	let newsub= 
	  SubstitutionSet.combine (List.map2 
			 (function qlhs2 -> function ql2 ->
			    if qlhs2=qlhs && ql2=q 
			    then es
			    else match_one qlhs2 ql2 s vl)
				     ls1 ls2)
	    
	in
	  if SubstitutionSet.is_empty newsub 
	  then (* One couple of states as no entry in s, and thus 
			    the list of subst is empty in that case*)
	    propagate_one_layer rem qlhs q es s vl res
	  else 
	    propagate_one_layer rem qlhs q es s vl (((dest1,dest2), newsub)::res)


(* propagation of a new substition w.r.t. a substable. New substitutions gives
   all the point where there are new values i.e. (q1,q2) -> {{x -> q1}} This has
   to be propagated for every Fathers of (q1,q2) w.r.t. the current values of
   brothers of (q1,q2). For instance if in a left-hand side of the rule we have:
   f(q1, q3) -> q0, there is a new value on couple (q1,q4) hence, one have to
   find the possible values for (q3,?) and propagate the combination of
   substitutions to (q0,?) etc... The result is the list of couples (q1,q2)
   where q1 is a final state of a lhs. Final is the set of final states of the
   lhs. Note that there may be a father to final states in automata for the lhs
   (but no cycles in the transition relation.

   We also propagate new substitution to successors modulo epsilon of
   states. However, if (qr,q1) is new and q1->epsilon*q2 then (qr,q2) is
   new but it is not added to the set of results as is (otherwise we 
   would not share the fact that successors w.r.t. R are also successors of
   q2. We only propagate this "new" information to fathers of q2... those 
   may also be added to the result if they provide a new substitution based
   on (qr,q2)... *)
	      
let rec bottomup_propagate (qlhs: state) (q: state) (es: substitutionset) (alhs: automaton) (a: automaton) (s: substable) (successor: bool) (vl: statevar) (iseq: bool) (newsub: ((state*state)*substitutionset) list): ((state*state)*substitutionset) list=
  try(
    let satablelhs= Hashtbl.find alhs.trans.satable qlhs in
    let llhs= Common.hashtbl_to_list satablelhs in
    let satablered= a.trans.satable in
      (* List of couples of transition lhs*redex where the top symbol is the
	 same and the position of qlhs in lhs trans is the same a q in redex
	 trans *)
    let (linter: (external_symtrans * external_symtrans) list)= intersec qlhs q llhs satablered [] in 
    let (new_combined_subst: ((state*state)*substitutionset) list)= 
      propagate_one_layer linter qlhs q es s vl [] in 

      (* Add the substitution corresponding to (qlhs,q) to s and get_back the
	 new elements *)
      
(* Necessary to reconsider all substitutions
   for filtering correctly on fixpoint_merging 
   hence we remove the optimization for equations *)
   
    let newes= if iseq then begin ignore(SubstitutionTable.add_set es (qlhs,q) s);es end
    else
      SubstitutionTable.add_set es (qlhs,q) s 
    in
    let newnewsub= 

      (* qlhs is a final state and q is not considered as a successor of another
	 q' then (qlhs,q)->es is a new set of subst for final states and if
	 it contains at least a new susbt (newes contains only new subst) *)

      if (Stateset.mem qlhs alhs.fstates) && 
	(not successor) &&
	(not (SubstitutionSet.is_empty newes))  
      then ((qlhs,q),newes)::newsub
      else newsub 
    in
      breadth_propagate new_combined_subst alhs a s vl iseq newnewsub 
  ) 

  with Not_found -> (* Cannot go up, qlhs is a final state *)
    (* Necessary to reconsider all substitutions
       for filtering correctly on fixpoint_merging 
       hence we remove the optimization for equations.
    *)

    let newes= 
      if iseq then begin ignore(SubstitutionTable.add_set es (qlhs,q) s);es end
      else
	SubstitutionTable.add_set es (qlhs,q) s 
    in
      if (Stateset.mem qlhs alhs.fstates) && 
	(not successor)
	(* qlhs is a final state then (qlhs,q)->es is a new set of subst for
	   final states and if it contains at least a new susbt (newes contains
	   only new subst) *)
      then ((qlhs,q),newes)::newsub
      else newsub 	      


(* propagation of a list of new substition w.r.t. a substable. New substitutions
   gives all the points in the intersection between alhs and aredex where there
   are new values i.e. (q1,q2) -> {{x -> q1}} *)

and breadth_propagate (ls: ((state * state) * substitutionset)list) (alhs: automaton) (a: automaton) (s: substable) (vl: statevar)(iseq: bool) (newsub: ((state*state)*substitutionset) list): ((state*state)*substitutionset) list=
  match ls with
      [] -> newsub
    | ((q1, q2), es)::rem ->
	(* if there is no father to q1 (q1 is the top of a rule) then we do not
	   propagate new substitutions to successors (the resulting values will
	   be implicitely propagated by epsilon closure)*)
	(* Find lhs transitions concerned by new values in q1 *)
	(* note that new in (q1,q2) => new in (q1,q3) if q1 ->epsilon* q3 *)

	let all_succ= Transitions.get_succ q2 a.trans in
	  (* We apply bottomup_propagate to q2, where q2 is not considered
	     as a successor of himself *)
	let first= bottomup_propagate q1 q2 es alhs a s false vl iseq newsub in
	  (* We apply bottomup_propagate to all successors of q2 (with
	     a shared list of new substitutions (fold). Only if we are not
	     in the case of an equation! But it is applied if we are
	     in noepsi mode! *)
	let bup= if iseq && not(!completion_type=Noepsi) then first else 
	  List.fold_left
		   (function sol -> function state ->
		      bottomup_propagate q1 state es alhs a s true vl iseq sol)
		   first
		   all_succ in
	  breadth_propagate rem alhs a s vl iseq bup  


(* returns the list of new substitutions where the first state corresponds to
   the lhs of the rule (resp. lhs or rhs of equation) and the second state
   corresponds to the redex tree automaton *)

let complete_filter (lhs: automaton)(aredex: automaton)(nt: external_symtrans list)
  (ne: external_epsilon list)(s: substable)
  (vl: statevar)(iseq: bool): ((state * state) * substitutionset) list= 

  (* filter and filter_epsi functions do not modify the substitution table s *)
  let new_sub= filter lhs nt [] s vl iseq in  (* new sub for new normal trans *)
    (* We do not propagate substs using epsilon for equations *)
    (* new sub for new epsi trans *)
  let new_sub2= if iseq && not(!completion_type=Noepsi) then new_sub else filter_epsi ne new_sub s in  
    (* modification is performed by breadth_propagate *)
    breadth_propagate new_sub2 lhs aredex s vl iseq []  



(* Applying merging on the state values of a substitution *)

let merge_subst_values (mt: state_merge_table) (sub: substitution): substitution=
  Substitution.map (function t -> 
		      (match t with 
			   Term.State(q) -> (Term.State (Transitions.merged_value mt q)) | _ -> failwith "Completion:merge_subst_values: should be applied to state substitution only!")) sub


(* Applying merging on the state values of a substitution set *)

let rec merge_substset (mt: state_merge_table) (substset: substitutionset): substitutionset=
  try(let e= SubstitutionSet.choose substset in
	    SubstitutionSet.add (merge_subst_values mt e) (merge_substset mt (SubstitutionSet.remove e substset))) 
      with SubstitutionSet.Empty_set -> SubstitutionSet.empty


(* merge one layer of states (from) into (towards) on a substable w.r.t.
   already known state merge values (that contains from -> towards) mt. 
   Produces a list of new substitutions. *)

let merge_one_layer (from: state list) (s: substable) (mt: state_merge_table)
  (res: ((state * state) * substitutionset) list): ((state * state) * substitutionset) list=
  (* construct the list of associations qr->{enssubst} associated with
     all the alhs states associated with all the states of the from list *)
  let rec filter_new l res=
    match l with 
	[] -> res
      | (towards,(qlhs,es))::rem -> 
	  let newsub= SubstitutionTable.add_set (merge_substset mt es) (qlhs,towards) s in
	    if SubstitutionSet.is_empty newsub then filter_new rem res
	    else filter_new rem (((qlhs,towards),newsub)::res)
  in      
  let all_new_sub= List.flatten
		     (List.map 
			(function qred -> 
			   let towards= Transitions.merged_value mt qred in
			     try
			       List.map (function e -> (towards,e))
				 (Common.hashtbl_to_list (Hashtbl.find s qred))
			     with 
				 Not_found -> [])
			from)
  in
  let result= filter_new all_new_sub res in
    begin
      (* Remove all merged states *)
      List.iter (Hashtbl.remove s) from;
      result
    end



(* Apply merge on the substitution table and produces new substitutions!
   without propagation to fathers etc. *)

let merge_subst_one_layer (merge_list: (state list * state) list)(s: substable)(mt:state_merge_table): ((state * state) * substitutionset) list=
  let rec aux l res=
    match l with 
	[] -> res
      | (sl,towards)::rem -> aux rem (merge_one_layer sl s mt res)
  in
    aux merge_list []


let merge_subst (merge_list: (state list * state) list)(alhs: automaton)(aredex: automaton)(s: substable)(mt: state_merge_table)(vl: statevar): ((state * state) * substitutionset) list=

  let newsub= merge_subst_one_layer merge_list s mt in
  let res= breadth_propagate newsub alhs aredex s vl true [] in
    res


(* Merge applied on condition_table *)

let merge_condition_table (ct: condition_table) (mt: state_merge_table): unit=
  let l= Common.hashtbl_to_list ct in
    List.iter (function (q,stateset) -> 
		 let new_state_set=
		   Transitions.stateset_map (Transitions.merged_value mt) stateset
		 in
		   Hashtbl.replace ct q new_state_set)
      l
      

(* The complete merge function: applies all merges on aredex, subs, mt, and
   produces a list of new substitution, used to produce new transitions. Also modifies mt *)

let complete_merge (merge_list: (state * state) list)(alhs: automaton)
  (aeq: automaton)(aredex: automaton)(slhs: substable)(seq: substable)(ct: condition_table)(mt: state_merge_table)(vllhs: statevar)(vleq: statevar): (automaton * ((state * state) * substitutionset) list *((state * state) * substitutionset) list)= 
  let (new_aut,refined_merge_list)= Automaton.merge aredex mt merge_list in
  let _= merge_condition_table ct mt in
    (new_aut,
     (merge_subst refined_merge_list alhs new_aut slhs mt vllhs),
     (merge_subst refined_merge_list aeq new_aut seq mt vleq))


(* An equation set maps an equation lhs state (resp. rhs) to the corresponding
   lhs state (resp. lhs). There can be more than one! *)

type equation_table= (state, state list) Hashtbl.t

type exteq_table= (state, exteq list) Hashtbl.t

let equation_table_to_string t=
  List.fold_right (function (k,e) -> function s -> ((State.to_string k)^" -> "^
  (String.concat "" ["{";
		     String.concat "," (List.map State.to_string e);
		       "}"]))^"\n"^s)
    (hashtbl_to_list t) ""

let exteq_to_string e=
  (match e with 
       Unique(_) -> "Unique"
     | Eq(q,_) -> "Eq("^(State.to_string q)^")"
     | Any(q,_) -> "Any("^(State.to_string q)^")")^
  (match e with
       Unique(rl) | Eq(_,rl) | Any(_,rl) ->
	 "["^(String.concat "," (List.map (function t1,t2 -> (Term.to_string t1)^"="^(Term.to_string t2)) rl))^"]")


let exteq_table_to_string t=
  List.fold_right 
    (function (k,l) -> function s -> 
       (State.to_string k)^" -> "^
       (String.concat "\n" (List.map exteq_to_string l))^"\n"^s)
    (hashtbl_to_list t) ""


(* This table gives for each lhs state the list of corresponding rhs (there
   could be more than one *) 

type rule_table= (state, (term * (condition list)) list) Hashtbl.t


let rule_table_to_string t=
  List.fold_right 
    (function (k,e) -> 
       function s -> 
	 ((State.to_string k)^" -> "^
	  "{"^(String.concat "," 
		 (List.map (function (t,ctl) ->
			      "("^(Term.to_string t)^" if "^
			      (String.concat " /\ "
				 (List.map 
				    (function c ->
				       match c with
					   Joinable(t1,t2) -> 
					     ((Term.to_string t1)^"<->"^(Term.to_string t2))
					 | _ -> failwith "Completion:rule_table_to_string: Not supported yet")
				    ctl))^")")
		    e))^"}"^"\n"^s))
    (hashtbl_to_list t) ""



(* This table gives for each lhs/rhs of an equation the set of states
   related (in redex automaton) ... used to find in susbstset the 
   corresponding new values... *)

type equation_subst= (state, stateset) Hashtbl.t

let equation_subst_to_string t=
  List.fold_right (function (k,e) -> function s -> ((State.to_string k)^" -> "^
  Stateset.to_string e)^"\n"^s)
    (hashtbl_to_list t) ""


let new_rule_table ()= Hashtbl.create Common.max_rule_node 

let new_equation_table ()= Hashtbl.create Common.max_equation

let new_exteq_table ()= Hashtbl.create Common.max_equation

let new_condition_table ()= Hashtbl.create Common.max_condition 

(* Fill pt and retrieves the corresponding final states *)

let rec fill_pattern (l: term list) (sl: state list)(pt: Automaton.s2t): state list=
  match l,sl with
      [],_ -> sl
    | (t::rem),(s::rem2) -> 
	let _= Hashtbl.add pt s t in
	  fill_pattern rem rem2 pt
    | _,_ -> failwith "Completion:fill_pattern: no other matching possible"


(* Fill the rtable and etable from the list of rules and list of equation
   + the list of corresponding states. There should be one state per rule
   and two per equation *)

let rec fill_tables (trs: trs) (et: equational_theory) (fs: state list)
  (rtable: rule_table) (etable: equation_table) (eetable: exteq_table)
  (ctable: condition_table): unit=
  match trs,et,fs with
      [],[],[] -> ()
    | _::_,_,[] -> failwith "Completion:fill_tables: bad number of states in fs!"
    | [],_::_,[] -> failwith "Completion:fill_tables: bad number of states in fs!"
    | [],_::_,[_] -> failwith "Completion:fill_tables: bad number of states in fs!"
    | [],[],_::_ -> failwith "Completion:fill_tables: bad number of states in fs!"
    | [],e::rem,s1::(s2::rem2) -> 
	(* For equations *)
	(* We add the corresponding links either in the etable or in the eetable
	   s1 -> s2 and s2 -> s1 *)
	(match e with
	     (* f(x)=(y)*)
	    Regular(_,_,_) -> 
	      begin
		(try (
		   let l= Hashtbl.find etable s1 in
		     Hashtbl.replace etable s1 (Common.add s2 l))
		 with Not_found -> Hashtbl.add etable s1 [s2]);
		(try (
		   let l= Hashtbl.find etable s2 in
		     Hashtbl.replace etable s2 (Common.add s1 l))
		 with Not_found -> Hashtbl.add etable s2 [s1]);
		fill_tables trs rem rem2 rtable etable eetable ctable
	      end
	      (* [f(x)] -> ... The case where there is only one term in pattern is modeled
		 by an [f(x)=_] *)
	   | Exteq(lhs,Var("_"),rl) ->
	       begin
		 (try let l= Hashtbl.find eetable s1 in
		    Hashtbl.replace eetable s1 (Common.add (Unique(rl)) l)
		  with Not_found -> Hashtbl.add eetable s1 [Unique(rl)]);
		 fill_tables trs rem rem2 rtable etable eetable ctable
	       end
		 (* [f(x)=g(y)]-> ...*)
	   | Exteq(lhs,rhs,rl) ->
	       begin
		 (try let l= Hashtbl.find eetable s1 in
		    Hashtbl.replace eetable s1 (Common.add (Eq(s2,rl)) l)
		  with Not_found -> Hashtbl.add eetable s1 [Eq(s2,rl)]);
		 (try let l= Hashtbl.find eetable s2 in
		    Hashtbl.replace eetable s2 (Common.add (Eq(s1,rl)) l)
		  with Not_found -> Hashtbl.add eetable s2 [Eq(s1,rl)]);
		 fill_tables trs rem rem2 rtable etable eetable ctable
	       end
		 (* [f(x),g(y)]-> ... *)
	   | Extany(lhs,rhs,rl) -> 
	       begin
		 (try let l= Hashtbl.find eetable s1 in
		    Hashtbl.replace eetable s1 (Common.add (Any(s2,rl)) l)
		  with Not_found -> Hashtbl.add eetable s1 [Any(s2,rl)]);
		 (try let l= Hashtbl.find eetable s2 in
		    Hashtbl.replace eetable s2 (Common.add (Any(s1,rl)) l)
		  with Not_found -> Hashtbl.add eetable s2 [Any(s1,rl)]);
		 fill_tables trs rem rem2 rtable etable eetable ctable
	       end)

    | (_,r,cond)::rem,_,(s::rem2) -> 
	begin
	  (try (
	     let l= Hashtbl.find rtable s in
	       Hashtbl.replace rtable s ((r,cond)::l))
	   with Not_found -> Hashtbl.add rtable s [(r,cond)]);
	  if not(cond=[]) then Hashtbl.add ctable s (Stateset.empty);
	  fill_tables rem et rem2 rtable etable eetable ctable
	end
	


(* Propagation of new subst on equation_subst table:
   for a solution: ((q1,q2),es), if q1 is the top symbol of an equation
   this function adds q2 to the set of states associated with q1 *)

let rec propagate_on_substeq (l: ((state * state) * substitutionset)list) (seq: equation_subst) (et: equation_table) (eet: exteq_table): unit=
  match l with
      [] -> ()
    | ((q1,q2),_)::rem ->
	(* We check if q1 is a member of an equation, if so we add the corresponding
	   state *)
	begin
	  if (Hashtbl.mem et q1) || (Hashtbl.mem eet q1)
	  then
	    let state_set= (try (Hashtbl.find seq q1) with Not_found -> Stateset.empty) in 
	      Hashtbl.replace seq q1 (Stateset.add q2 state_set)
	  else ();
	  propagate_on_substeq rem seq et eet
	end

(* completion_step type:
   - alhs is the automaton of all left-hand sides of rules and equations, 
   - aredex is the tree automaton being completed
   - svlhs is the list of all state names associated to variables in lhs of rules in alhs
   - sveq same for equations
   - et: equation table associating the state of an equation member to the state of the other
   - eet: extended (with patterns) equations
   - rt: rule_table
   - ct: condition_table
   - pattern: table associating state names to the pattern (term) for explicit pattern detection
   - subst_for_eq: found equivalent states w.r.t to a state
   - sublhs: tabled substitutions for lhs of rules
   - subeq: tabled substitution for equations left and right hand sidesame for lhs of rules
   - new_trans: new "regular" transitions that are part of aredex and not of the previous completed T.A
   - new_epsi: same for epsilon transitions
*)

type completion_step= {alhs: automaton;
		       aeq: automaton;
		       aredex: automaton;
		       svlhs: statevar;
		       sveq: statevar;
		       et: equation_table;
		       eet: exteq_table;
		       rt: rule_table;
		       ct: condition_table;
		       state2term: Automaton.s2t;
		       pattern: Automaton.s2t;
		       subst_for_eq: equation_subst;
		       sublhs: substable;
		       subeq: substable;
		       new_trans: external_symtrans list;
		       new_epsi: external_epsilon list}

let to_string (c: completion_step): string=
  "---------\nalhs= "^(Automaton.to_string c.alhs)^
  "\naeq= "^(Automaton.to_string c.aeq)^
  "\n\naredex= "^(Automaton.to_string c.aredex)^
  "\n\nsv= "^(Automaton.statevar_to_string c.svlhs)^
  "\n\neqsv= "^(Automaton.statevar_to_string c.sveq)^
  "\n\net= "^(equation_table_to_string c.et)^
  "\n\neet= "^(exteq_table_to_string c.eet)^
  "\n\nrt= "^(rule_table_to_string c.rt)^
  "\n\nct= "^(condition_table_to_string c.ct)^
  "\n\npattern= "^(Automaton.s2t_to_string c.pattern)^
  "\n\nstate2term= "^(Automaton.s2t_to_string c.state2term)^
  "\n\nsubst_for_eq= "^(equation_subst_to_string c.subst_for_eq)^
  "\n\nsublhs substable= "^(SubstitutionTable.to_string c.sublhs)^
  "\n\nsubeq substable= "^(SubstitutionTable.to_string c.subeq)^
  "\n\nmergetable= "^(Transitions.merge_table_to_string c.aredex.trans.stmt)^
  "\n\nnew_trans= "^(Transitions.symextlist_to_string c.new_trans)^
  "\n\nnew_epsi= "^(Transitions.epsiextlist_to_string c.new_epsi)^
  "\n\n--------\n\n"


(* Applies a substitution list to a list of equations el, i.e. the rhs of an extended
   equation l1=r1...ln=rn, to produce a list of state couples for merging *)

let rec apply_subst_equationlist (el: (term*term) list) (sl: substitution list) (d: delta) (mt: state_merge_table) (res: (state*state) list): (state*state) list=
  match sl with 
      [] -> res
    | s::rem -> 
	let newres= 
	  List.fold_right 
	    (function (lhs,rhs) -> function l ->
	       let qlhs= Automaton.all_run (Rewriting.subst_apply s lhs) d in
		 if qlhs=[] then l
		 else 
		   let qrhs= Automaton.all_run (Rewriting.subst_apply s rhs) d in
		     if qrhs=[] then l
		     else 
		       let all_combination= 
			 List.filter (function q1,q2 -> not(q1=q2)) 
			   (Common.combine qlhs qrhs) 
		       in
		       let _= if (!Common.verbose>1) then 
			 List.iter 
			   (function (q1,q2) -> 
			      print_string ("Merging states: "^(State.to_string q1)^" and "^(State.to_string q2)^"\n")) all_combination in
			 Common.union all_combination l)
	    el res
	in
	  apply_subst_equationlist el rem d mt newres 
	

(* Get the set of substitutions to apply for a given exteq and set of substitution.
   given seq1 the top state of lhs of equation, sredex1 the top state of matching equation 
   in aredex automaton. *)

let filter_subst_exteq (exteq: exteq)(seq1: state)(sredex1: state)
  (es: substitutionset)(seq: substable)(substeq: equation_subst)(mt: state_merge_table)(stt: s2t)
  : (substitutionset * ((term * term) list))=
  match exteq with 
      Unique(rl) -> 
	let _= 
	  if (!Common.verbose>1) then try (print_string
	    ("Equation ["^(Term.to_string (Hashtbl.find stt seq1))^"] => ["^
	     (String.concat " " (List.map (function (t1,t2) -> (Term.to_string t1)^"="^(Term.to_string t2)) rl))^"]"^" applied on state "^(State.to_string sredex1)^"\n"))
	  with Not_found -> failwith ("Term state "^(State.to_string seq1)^" does not have image in stt table!")
	in 
	  es,rl
    | Eq(seq2,rl) -> (* for 'Eq' the redex state must be the same *)
	(try (let othersub=Hashtbl.find (Hashtbl.find seq sredex1) seq2 in 
	      let _= 
		if (!Common.verbose>1) then 
		  try (print_string
			 ("Equation ["^(Term.to_string (Hashtbl.find stt seq1))^"="^
			  (Term.to_string (Hashtbl.find stt seq2))^
			  "] => ["^(String.concat " " (List.map (function (t1,t2) -> (Term.to_string t1)^"="^(Term.to_string t2)) rl))^"] applied on state "^(State.to_string sredex1)^"\n"))
		  with Not_found -> failwith ("Term state "^(State.to_string seq1)^"or "^(State.to_string seq2)^" does not have image in stt table!")
	      in 
		(SubstitutionSet.extended_intersection es othersub),rl) 
	 with Not_found -> (SubstitutionSet.empty,rl))
    | Any(seq2,rl) -> (* for 'Any' the redex state can be anyone *)
	(try 
	   (let redex_for_seq2= (Hashtbl.find substeq seq2) in
	    let new_redex_for_seq2= Transitions.stateset_map (Transitions.merged_value mt) redex_for_seq2 in
	      (* Look if substeq has to be updated, if so, do Subst_for_equation
		 is updated in a lazy way here. *)
	    let _ = if not (redex_for_seq2=new_redex_for_seq2) then Hashtbl.replace substeq seq2 new_redex_for_seq2 else ()  
	    in  
	    let _= 
	      if (!Common.verbose>1) then 
		try (print_string
		       ("Equation ["^(Term.to_string (Hashtbl.find stt seq1))^","^
			(Term.to_string (Hashtbl.find stt seq2))^
			"] => ["^(String.concat " " (List.map (function (t1,t2) -> (Term.to_string t1)^"="^(Term.to_string t2)) rl))^"] applied on state "^(State.to_string sredex1)^" for lhs and states "^
			(String.concat "," (List.map State.to_string (Stateset.elements new_redex_for_seq2)))^" for rhs\n"))
		with Not_found -> failwith ("Term state "^(State.to_string seq1)^"or "^(State.to_string seq2)^" does not have image in stt table!")
	    in 
	    let eslist= Stateset.fold  
			  (function sred -> function l -> 
			     try ((SubstitutionTable.find (seq2,sred) seq)::l)
			     with SubstitutionTable.Couple_not_found -> l)
			  new_redex_for_seq2 [] 
	    in
	      List.fold_right 
		(function es2 -> function esres ->
		   let inter= SubstitutionSet.extended_intersection es es2 in
		     if SubstitutionSet.is_empty inter then esres
		     else SubstitutionSet.union inter esres)
		eslist SubstitutionSet.empty),rl
	 with Not_found -> SubstitutionSet.empty,rl)


(* applies a list of labelled substitution (labelled by qlhs and qredex states) on
   an exteq_table eet, w.r.t state merging table mt  *)	

let rec apply_subst_extendedeq (ls: ((state * state) * substitutionset)list) 
  (eet: exteq_table)(seq: substable)(substeq: equation_subst)(d: delta)
  (mt: state_merge_table)(s2t: s2t)(res: (state * state) list): ((state * state) list)=
  (match ls with
      [] -> res
    | ((seq1,sredex1),es)::rem ->
	try(
	  let eelist= Hashtbl.find eet seq1 in (* get the list of exteq associated to seq1 *)
	  let subst_to_apply= List.map (function e -> filter_subst_exteq e seq1 sredex1 es seq substeq mt s2t) eelist in
	  let newres=
	    List.fold_right 
	      (function (newes,rl) -> function l ->
		 if SubstitutionSet.is_empty newes then l
		 else
		   apply_subst_equationlist rl (SubstitutionSet.elements newes) d mt l)
	      subst_to_apply res
	  in
	    apply_subst_extendedeq rem eet seq substeq d mt s2t newres
	)
	with Not_found -> (* The root state is not part of an exteq *)
	  apply_subst_extendedeq rem eet seq substeq d mt s2t res)


(* Apply subst applies substitution to lhs/rhs of equations and checks if the
   equation applies. This function is also supposed to update substeq and seq
   w.r.t. mt. This function does not produce trivial merge equation.
   If the lhs or rhs is just a variable, we simply merge with this state *)

let rec apply_subst_eq (ls: ((state * state) * substitutionset) list) (delta: delta)(seq: substable)(et: equation_table) (mt: state_merge_table) (substeq: equation_subst)(sv: statevar)(stt:Automaton.s2t)(res2: ((state * state) list)): ((state * state) list)=

  (* qrhs is the alhs state for the right hand side (or lhs) of equation 
     q2 is the state matched by the lhs (or rhs) in the aredex *)

  let rec aux (qlhs: state) (qrhs: state) (q2: state) (es: substitutionset) : (state * state) list =
    (* if rhs is a variable we merge the result of es applied to var of qhs, 
       with q2 *)
    try(
      let variable= Automaton.statevar_find sv qrhs in
	List.fold_left 
	  (function l -> function s ->
	     if s=q2 then l else 
	       (* equation applies *)
	       let _= if (!Common.verbose>1) then 
		 (try 
		   (print_string ("Equation "^(Term.to_string (Hashtbl.find stt qlhs))^" = "^
				  (* (Variable.to_string variable) *)
				  (Term.to_string (Hashtbl.find stt qrhs))
				     ^" applies on states: "^ 
				  (State.to_string q2)^" and "^(State.to_string s)^"\n"))
		 with Not_found -> failwith ("Term state "^(State.to_string qlhs)^" does not have image in stt table!")) in
	       (s,q2)::l)
	  [] (List.map Term.get_state (Rewriting.substset_apply es (Var variable))))
    with Not_found ->
    try(  
      (* Look for aredex states associated with substs associated with qrhs in aredex *)
      let state_redex= Hashtbl.find substeq qrhs in
      let new_state_redex= Transitions.stateset_map (Transitions.merged_value mt) state_redex in
	(* Look if substeq has to be updated, if so, do 
	Subst_for_equation is updated in a lazy way here. *)
      let _ = if not (state_redex=new_state_redex) then Hashtbl.replace substeq qrhs new_state_redex else () 
      in
      let new_state_list= Stateset.elements new_state_redex in
	(* List associating with states in aredex (for rhs match) the 
	   corresponding sets of substitution *)
	List.fold_left
	  (function l -> function qredex -> 
	     try 
	       (let es2= SubstitutionTable.find (qrhs,qredex) seq in
		  if not(qredex=q2) && 
		    SubstitutionSet.matching_substitutionset es2 es 
		  then 
		    (* a new merge position has been found! *)
		    let _= 
		      if (!Common.verbose>1) then
		      (try 
			 (print_string ("Equation "^(Term.to_string (Hashtbl.find stt qlhs))^" = "^
					(Term.to_string (Hashtbl.find stt qrhs))^" applies on states: "^
				  (State.to_string q2)^" and "^(State.to_string qredex)^"\n"))
		 with Not_found -> failwith ("Term state "^(State.to_string qlhs)^" or "^(State.to_string qrhs)^" does not have image in stt table!")) in
		    (q2,qredex)::l 
		  else l)
	     with SubstitutionTable.Couple_not_found -> l)
	  []
	  new_state_list )
    with Not_found -> []    
  in
    match ls with
	[] -> res2   
	  (* q1 is the qlhs, q2 is the qredex *)
      | ((q1,q2),es)::rem ->
	  let new_res2= 
	    (* memb is the list of rhs of corresponding equation *)
	    try (let memb= Hashtbl.find et q1 in 
		   List.append (List.flatten (List.map (function elt_memb -> aux q1 elt_memb q2 es) memb)) res2)
	    with Not_found -> res2 in
	    apply_subst_eq rem delta seq et mt substeq sv stt new_res2 
	


let new_completion_step (a: automaton) (trs: trs) (et: equational_theory) (pattern: term list)(trsvl: variable list)(eqvl:variable list): completion_step= 
  (* we only add lhs of rules in alhs *)
  let realet= if !approx_type=Equationnal then et else [] in
  let s2t= Automaton.new_s2t() in
  let all_terms_lhs= List.fold_right 
		       (function (lhs,_,_) -> function e -> 
			  lhs::e)
		       trs []
  in
    (* we add lhs and rhs for equations in alhs *)
  let all_terms_eq=  
    List.fold_right 
      (function e -> function l -> 
	 match e with Regular(lhs,rhs,_) | Exteq(lhs,rhs,_) | Extany(lhs,rhs,_) ->
	   lhs::(rhs::l))
      realet []
  in
  let (alhs,svlhs,fslhs)= Automaton.term_to_automaton (List.append pattern all_terms_lhs) trsvl s2t in
  let (aeq,sveq,fseq)= Automaton.term_to_automaton all_terms_eq eqvl s2t in
  let rtable= new_rule_table () in
  let etable= new_equation_table () in  
  let eetable= new_exteq_table () in  
  let ptable= Automaton.new_s2t () in
  let ctable= new_condition_table () in
  let new_fslhs= fill_pattern pattern fslhs ptable in	 	
  let _= fill_tables trs realet (List.append new_fslhs fseq) rtable etable eetable ctable in
  let (trans,epsi)= Transitions.partition (Transitions.to_list a.trans) in

  let subst_for_eq= Hashtbl.create Common.max_equation in
  let subeq= SubstitutionTable.new_substable () in
    {alhs=alhs;
     aeq=aeq;
     aredex= a;
     svlhs= svlhs;
     sveq= sveq;
     et= etable;
     eet= eetable;
     rt= rtable;
     ct= ctable;
     pattern= ptable;
     state2term= s2t;
     subst_for_eq=subst_for_eq;
     subeq= subeq;
     sublhs= SubstitutionTable.new_substable ();
     new_trans= trans;
     new_epsi= epsi}
      

(* For a given list of (right hand side + a list of conditions) and set of
   substitutions es, this function returns the list of couples (rhs, substituted
   right hand sides) for all cases where conditions are satisfied. *)

let rec check_conditions (l: (term * (condition list))list) (es: substitutionset) (delta: delta) (acc: (term*term) list): (term*term) list=
  let rec aux (es: substitutionset) ((rhs,cond):(term * (condition list))) (acc: (term*term)list): ((term*term) list)=
    if SubstitutionSet.is_empty es then acc
    else 
      let s= SubstitutionSet.choose es in
      let substcond= List.map 
		       (function c ->
			  match c with 
			      Joinable(c1,c2) -> 
				(Term.get_state (Rewriting.subst_apply s c1)),
				(Term.get_state (Rewriting.subst_apply s c2))
			    | _ -> failwith "Completion:check_conditions: Not supported yet")
		       cond
      in
      let new_es= (SubstitutionSet.remove s es) in
	if (Automaton.not_empty_intersection substcond delta) 
	then 
	  aux new_es (rhs,cond) ((rhs,(Rewriting.subst_apply s rhs))::acc)
	else
	  aux new_es (rhs,cond) acc
  in
    match l with 
	[] -> acc
      | (rhs,cond)::rem ->
	  check_conditions rem es delta (aux es (rhs,cond) acc)
	    


(* OPTIMISATION possible here 
   Checking if a term is already recognized, is performed in a non-deterministic way.
   If it is not precise enough comment the following code...
   Note that in that case normalisation may generate new transitions not be 
   deterministic. For instance in example c: 
   we have the following transition to normalize: f(s(q3),s(q4)) -> q8 but
   f(s(q3),s(q4)) is normalized to q5 (using deterministic normalization (first
   transition to come is ok :-)) and thus q5 -> q8 is added to delta though
   f(s(q3), s(q4)) ->* q8

   Code to comment:
   if Automaton.oriented_run t q2 delta 
   then 
   let _= print_string ("Transition covered!\n") in
   []
   else *)



(* Add an entry to qlhs if it is already in the table (it is a conditional rule) *)

let update_condition_table (qlhs: state) (qredex:state) (ct: condition_table)=
  try (let stateset= Hashtbl.find ct qlhs in
	 Hashtbl.replace ct qlhs (Stateset.union stateset (Stateset.singleton qredex)))
  with Not_found -> ()


(* Depending on the strategy we use new or existing states. Depending on the 
   completion type, we use either epsilon transitions for top rewriting or not. 
   Note that, specific normalization add the new rules to delta! *) 

let specific_normalization (t:term) (dest:state) (delta: delta) (cs: strategy) (ct: comptype): external_trans list=
  match cs with
      Prior -> 
	(* If the transition exists in Delta we reuse it *)
	if ct=Noepsi then Automaton.oldnormalize_term false t dest delta 
	else let (top_state, new_trans)= Automaton.normalize_term t delta false in
	  Epsilon(top_state,dest)::new_trans
    | New -> (* we use only new states *)
	if ct=Noepsi then Automaton.oldnormalize_term true t dest delta
	else let (top_state, new_trans)= Automaton.normalize_new_states t delta in
	  Epsilon(top_state,dest)::new_trans


(* Apply subst applies substitution to rhs of rules.
   This function is also supposed to update slhs w.r.t. mt.
   also add new transitions to delta *)

let rec apply_subst_lhs (ls: ((state * state) * substitutionset) list) (delta: delta)(rt: rule_table)(ct: condition_table)(mt: state_merge_table)(stt:s2t) (res1: external_trans list): (external_trans list)=
  match ls with
      [] -> res1
	(* q1 is the qlhs, q2 is the qredex *)
    | ((q1,q2),es)::rem ->
	(* We apply the renaming by mt before applying subst *)
	let new_es= merge_substset mt es in
	  (* If it is a conditional rule we signal the new hit on q2 *)
	let _= update_condition_table q1 q2 ct in
	let new_res1=
	  try (let erule= (Hashtbl.find rt q1) in
		 (* New set of substituted right-hand sides *)
	       let etcouple= check_conditions erule new_es delta [] in
	       let lhs= try (Term.to_string (Hashtbl.find stt q1)) 
	       with Not_found -> failwith ("Term state "^(State.to_string q1)^" does not have image in stt table!")
	       in
	       let _=  
		 if (!Common.verbose>1) then 
		   (print_string (String.concat "\n" (List.map 
			(function (r,rs) -> ("Rule "^lhs^" -> "^(Term.to_string r)^" applied on state "^(State.to_string q2)^" with substituted form: "^(Term.to_string rs))) etcouple)^"\n"))
	       in
	       let et=List.map snd etcouple in

		 (* We construct the list of new transitions for es *)
		 List.append 
		   (List.flatten
		      (List.map (function t -> 
				   match t with
				       Fsym(_,_) -> 
					 let _= if (!Common.verbose>1) then print_string ("Transition to normalize: "^(Term.to_string t)^" -> "^(State.to_string q2)^"\n") else () in
					   (* We normalize new term t w.r.t. delta
					      if it is covered for merge_loop mode
					      it is enough *)
					   if !(Common.merge_loop) && (Automaton.oriented_run t q2 delta) 
					   then 
					     let _=  if (!Common.verbose>1) then print_string ("Transition covered!\n") else () in
					       []
					   else
					     (* Depending on the strategy we use new or existing states. Depending on the 
						completion type, we use either epsilon transitions for top rewriting or not. 
						Note that, specific normalization add the new rules to delta! *) 

					     let normedtrans=
					       specific_normalization t q2 delta !completion_strategy !completion_type in
					     let _= if (!Common.verbose>1) then print_string ("Adding new_trans="^(Transitions.list_to_string normedtrans)^"\n") else () in
					       normedtrans
						       
				     | State(q3) -> 
(*
					 if Automaton.oriented_run t q2 delta 
					 then let _= print_string ("Transition covered!\n") in
					   []
					 else *)
					 [Epsilon(q3,q2)]
				     | _ -> failwith "Completion:apply_subst:cannot be anything else") et))
		   
		   res1)
	  with Not_found -> res1 in
	  apply_subst_lhs rem delta rt ct mt stt new_res1
	  

(* Apply subst applies substitution to patterns and check if it applies.
   If so, we return the pattern and the fond substitution set*)

let rec apply_subst_pattern (ls: ((state * state) * substitutionset) list) (delta: delta)(pat: Automaton.s2t)(mt: state_merge_table)(res1: (term * state * substitutionset) list): (term * state * substitutionset) list=
  let _= if (!Common.verbose>3) then print_string ("ls= "^(SubstitutionTable.list_to_string ls)^"\n") in
  match ls with
      [] -> res1
	(* q1 is the qlhs, q2 is the qredex *)
    | ((q1,q2),es)::rem ->
	if (SubstitutionSet.is_empty es) 
	then (apply_subst_pattern rem delta pat mt res1)
	else 
	  let new_res1= 
	    try (
	      let term= (Hashtbl.find pat q1) in
		(term,q2,es)::res1)
	    with Not_found -> res1 in
	    apply_subst_pattern rem delta pat mt new_res1



(* Multiple adding of epsilon transition. Result is the list of uncovered 
   new epsilon transitions and list of states merging due to epsilon adding *)

let rec multiple_add_epsilon (le: external_trans list) (d: delta) (uncovered_epsi: external_epsilon list) (sm: ((state list * state) list))=
  match le with 
      [] -> uncovered_epsi,sm
    | Epsilon((q1,q2))::rem -> 
	(* Since every adding of epsilon transition may add merging
	   we apply systematically the merging *)
	let new_q1= Transitions.merged_value d.stmt q1 in
	let new_q2= Transitions.merged_value d.stmt q2 in
(*	  if new_q1=new_q2 then multiple_add_epsilon rem d uncovered_epsi sm
	  else *)
	let (b,res)= Transitions.add (Epsilon (new_q1,new_q2)) d in
	  if b then 
	    multiple_add_epsilon rem d ((new_q1,new_q2)::uncovered_epsi)
	      (match res with
		   [] -> sm 
		 | [(sl,s)] -> (sl,s)::sm
		 | _ -> failwith "Completion:multiple_add_epsilon: at least 2 states"
	      )
	  else
	    multiple_add_epsilon rem d uncovered_epsi sm
    | _ -> failwith "Completion:multiple_add_epsilon: cannot be another kind of transition"


(* Returns all top substitution of conditional rules *)

let rec add_conditional_sub (ct: (state * stateset) list) (slhs: substable) (acc: ((state * state) * substitutionset) list): ((state * state) * substitutionset) list=
  match ct with
      [] -> acc
    | (s,stateset)::rem ->
	let new_acc=
	Stateset.fold 
	  (function elt -> function l ->
	     try (((s,elt),(SubstitutionTable.find (s,elt) slhs))::l)
	     with SubstitutionTable.Couple_not_found -> failwith "Completion:add_conditional_sub:Subsitution not found in table for conditional rule!")
	  stateset
	  acc
	in add_conditional_sub rem slhs new_acc


(* Merging fixpoint: applies successively merging equation up to a normal
   form. The result is the normalized automaton+ list of transitions modified by
   renaming... used to be added to the new transitions obtained by completion
   steps *)

let fixpoint_merging (init_sym: external_symtrans list) (init_epsi: external_epsilon list) (aut: automaton) (stmt: state_merge_table) (eq: equation_table) (eet: exteq_table) (aeq: automaton) (eqvar: statevar) (es: equation_subst) (subeq: substable)(stt:Automaton.s2t): (automaton * external_symtrans list * external_epsilon list)=
  let rec fpmerging_aux (a: automaton) (sym_to_consider: external_symtrans list) (epsi_to_consider: external_epsilon list) (merged_states: state list): (automaton * state list)=
    match sym_to_consider, epsi_to_consider with
	[],[] -> a,merged_states
      | _,_ -> 
	  let soleq= complete_filter aeq a sym_to_consider epsi_to_consider subeq eqvar true in 
	  let _= propagate_on_substeq soleq es eq eet in

	  let states_to_merge1= apply_subst_eq soleq a.trans subeq eq stmt es eqvar stt [] in 
	  let states_to_merge2= apply_subst_extendedeq soleq eet subeq es a.trans stmt stt states_to_merge1 in
	  let _= if (!Common.verbose>2) then 
	    (if states_to_merge2=[] then print_string "\nNo more states to merge!\n" else print_string ("\nStates to merge="^(String.concat ", " (List.map (function (q1,q2) -> "("^(State.to_string q1)^","^(State.to_string q2)^")") states_to_merge2))^"\n"))
	  else ()
	  in
					       
	  let (new_aut,refined_state_list)= Automaton.merge a stmt states_to_merge2 in
	    if refined_state_list=[] then new_aut,merged_states
	    else 
	      let new_merged_states=  (* At this level we only keep trace of states receiving new 
					 transitions, to find the new sets of sym and epsilon trans *)
		(List.fold_right (function (l,a) -> function rem -> 
				    a::rem) refined_state_list [])
	      in
		(* When renaming q1->q2, Transitions to RE-consider are:
		   transitions of the form c->q2 and ALSO transitions of the
		   form f(...q2...) ->q *)
	      let new_sym= 
		List.append (Transitions.simple_get_all_trans new_merged_states new_aut.trans)
		  (Transitions.get_trans_parameter_list new_merged_states new_aut.trans) in
	      let new_epsi= Transitions.dest_epsilon_to_list new_merged_states new_aut.trans in
		fpmerging_aux new_aut new_sym new_epsi (List.append new_merged_states merged_states)

  in
  let (new_aut,merged_states) = fpmerging_aux aut init_sym init_epsi [] in
  let remaining_states= 
    List.fold_right 
      (function st -> function sl -> 
	 let new_name= Transitions.merged_value stmt st in
	   if List.mem new_name sl 
	   then sl
	   else new_name::sl)
      merged_states [] in					   
    (* When renaming q1->q2, Transitions to RE-consider are: transitions
       of the form c->q2 and ALSO transitions of the form f(...q2...) ->q *)

  let final_sym= 
    List.append (Transitions.simple_get_all_trans remaining_states new_aut.trans)
      (Transitions.get_trans_parameter_list remaining_states new_aut.trans)
  in
  let final_epsi= Transitions.dest_epsilon_to_list remaining_states new_aut.trans in
  let _ = if (!Common.verbose>3) then print_string ("Transitions due to merging:\n"^(Transitions.symextlist_to_string final_sym)^"\n"^(Transitions.epsiextlist_to_string final_epsi)^"\n\n") in 
    new_aut,final_sym,final_epsi



(* filter substitutions associated to final states *)

let filter_final (ls: (term * state * substitutionset) list) (fs: Stateset.t): (term *state * substitutionset)list=
  List.filter (function (_,s,_) -> Stateset.mem s fs) ls 


(* One step of completion, performs a full copy of current completion step 
   and returns the set of substitution matching the pattern terms (if any)
   last_delta is there only to ckeck if new_trans once merge are covered... *)

let completion_one_step (c: completion_step)(lastdelta: delta): (completion_step * ((term * state * substitutionset) list))= 
  (* 1) We make copies of mutable structures for the new completion step *)
  let new_subst_for_eq= Hashtbl.copy c.subst_for_eq in
  let temp_aredex= Automaton.copy c.aredex in
  let new_slhs= SubstitutionTable.copy c.sublhs in
  let new_seq= SubstitutionTable.copy c.subeq in
  let new_ct= Hashtbl.copy c.ct in

  (*  let _= print_string ("\nMerging with "^
      (String.concat ", " 
      (List.map (function (a,b) -> 
      "("^(State.to_string a)^"="^(State.to_string b)^")") c.ml))^"\n") in
  *)

  (* 2) We apply merging *)

  let (new_aredex, new_sym_from_merge, new_epsi_from_merge)= fixpoint_merging c.new_trans c.new_epsi temp_aredex temp_aredex.trans.stmt c.et c.eet c.aeq c.sveq new_subst_for_eq new_seq c.state2term in

  (* We merge the set of transition (added by last completion step) to add
     w.r.t. stmt, and suppress empty epsilon transitions *)
  (* and we add the new transitions produced by merging to the set of transitions to reconsider.
     they do not need to be merged so they are added as is, as the  accumulator of fold_rights *)

  let tmp_merged_new_trans= clean
    (List.fold_right 
      (function t -> function l -> (Transitions.merge_symext new_aredex.trans.stmt t)::l)
      c.new_trans new_sym_from_merge) in

  let tmp_merged_new_epsi= clean
     (List.fold_right (function (q1,q2) -> function l->
		       let new_q1= Transitions.merged_value new_aredex.trans.stmt q1 in
		       let new_q2= Transitions.merged_value new_aredex.trans.stmt q2 in
			 if new_q1=new_q2 then l
			 else (new_q1,new_q2)::l) c.new_epsi new_epsi_from_merge) in
    

  (* After the merge step, some transitions may be covered by old delta, they
     are removed *)

  let (merged_new_trans: external_symtrans list)= List.filter (function (t: external_symtrans) -> not (Transitions.mem t lastdelta.Transitions.trans)) tmp_merged_new_trans in
    
  (* If states q1 or q2 does not even occur in lastdelta, they are definitely
     new and so is the transition. Otherwise we check if the epsilon is already
     covered or not *)

  let merged_new_epsi= List.filter 
			 (function (q1,q2) -> 
			    if (Transitions.mem_state lastdelta q1)&&(Transitions.mem_state lastdelta q2) 
			    then 
			      not (Transitions.is_pred q1 q2 lastdelta)
			    else true) tmp_merged_new_epsi in
  let _= if (!Common.verbose>2) 
  then print_string ("After merging, the new transitions to consider are: \n"^
		     (symextlist_to_string merged_new_trans)^"\nand Epsilon:\n"^
		     (epsiextlist_to_string merged_new_epsi)) in
    if merged_new_trans=[] && merged_new_epsi=[] 
    then 
      ({subst_for_eq= new_subst_for_eq;
	aredex= new_aredex;                  
	sublhs= new_slhs;
	subeq= new_seq;
	new_trans= [];
	new_epsi= [];
	et= c.et;
	eet= c.eet;
	rt= c.rt;
	ct= new_ct;
	pattern= c.pattern;
	state2term= c.state2term;
	alhs= c.alhs;
	aeq= c.aeq;
	svlhs= c.svlhs;
	sveq= c.sveq},
       [])  (* If there are no new transition at all (even with merging operations)
	       then patterns cannot be matched now if it was not the case before *)

    else 
      (* 3) We search and produce new transitions *)
      (* We use transitions of the LAST completion step to obtain new matching 
	 with lhs *)
      
      let _= if (!Common.verbose>3) then
	print_string ("Before complete filter:\nnewaredex="^(Automaton.to_string new_aredex)^"\nmerged_new_trans="^(Transitions.symextlist_to_string merged_new_trans)^"\nmerged_new_epsi="^(Transitions.epsiextlist_to_string merged_new_epsi)^"\nnew_slhs="^(SubstitutionTable.to_string new_slhs)^"\nsvlhs="^(Dumper.dump c.svlhs)^"\n") in
      let sublhs= complete_filter c.alhs new_aredex merged_new_trans merged_new_epsi new_slhs c.svlhs false in
      let _= if (!Common.verbose>3) then print_string ("sublhs= "^(SubstitutionTable.list_to_string sublhs)^"\n") in
      let all_sublhs= 
	add_conditional_sub (Common.hashtbl_to_list new_ct) new_slhs sublhs in
	
      (* We check if a pattern is filled (pattern or subpattern depending on what was declared in 
	 the specification file *)    
      let patsub= 
	if ((Hashtbl.length c.pattern)=0) then []
	else 
	  let temp= apply_subst_pattern all_sublhs new_aredex.trans c.pattern new_aredex.trans.stmt [] in
	    if !Common.subpattern then temp 
	      (* for patterns we check in final states and predecessors of final states w.r.t epsilon
	         i.e. terms that comes from final states of the initial TA *)
	    else filter_final temp (Transitions.get_all_pred new_aredex.fstates new_aredex.trans)
      in
	
      (* We apply the substitution to obtain the new transitions. Note that
         they are added to delta! *)
      let obtained_trans= apply_subst_lhs all_sublhs new_aredex.trans c.rt new_ct new_aredex.trans.stmt c.state2term [] in
	
      (* We split transitions between trans and epsilon trans *)
      let (trans,epsi)= List.partition Transitions.not_epsi obtained_trans in
	
      (* Epsi transitions are added and states to merge are gathered *) 
      let (epsi_not_covered, epsi_merge_states)=
	multiple_add_epsilon epsi new_aredex.trans [] [] in
	
      (* Each new transition (not epsi) is added as is (after renaming w.r.t. 
	 merge of epsi *)

      (* New transitions of 'trans' have already been added but not those that may be 
	 a renaming of 'trans', so we add them *)

      let trans_not_covered= 
	List.append
	  (List.filter (function t -> fst(Transitions.add t new_aredex.trans))
	     (List.map (Transitions.merge_external new_aredex.trans.stmt) trans))
	  trans
      in  

      (* adding an epsi producing a cycle apply merging on trans but not on final state set. It
	 has to be done now *)
      let full_aredex= {trans= new_aredex.trans;
			fstates= Transitions.stateset_map (function state -> Transitions.merged_value new_aredex.trans.stmt state) new_aredex.fstates}  in
	
	
      (* 5) Since states used for epsi-renaming can be considered as *new* their
	 transitions are considered as new also... *)

      let states_to_reconsider= List.map (function (_,s) -> s) epsi_merge_states in
      let sym_to_reconsider= Transitions.simple_get_all_trans states_to_reconsider full_aredex.trans in
      let epsi_to_reconsider= Transitions.dest_epsilon_to_list states_to_reconsider full_aredex.trans in
	
	({subst_for_eq= new_subst_for_eq;
	  aredex= full_aredex;                  
	  sublhs= new_slhs;
	  subeq= new_seq;
	  new_trans= List.append sym_to_reconsider (List.map Transitions.ext2trans trans_not_covered);
	  new_epsi= List.append epsi_to_reconsider epsi_not_covered;
	  et= c.et;
	  eet= c.eet;
	  rt= c.rt;
	  ct= new_ct;
	  pattern= c.pattern;
	  state2term=c.state2term;
	  alhs= c.alhs;
	  aeq= c.aeq;
	  svlhs= c.svlhs;
	  sveq= c.sveq},
	 patsub)


(* Interactive pattern checking *)

let check_intersection (a: automaton) (pattern: term list) (vl: variable list):
  (term * state * substitutionset) list=
  let state2term_table= Automaton.new_s2t () in
  let (autpat,svpat,fspat)= Automaton.term_to_automaton pattern vl state2term_table in
(*   let _= fill_pattern pattern fspat state2term_table in *)
  let (trans,epsi)= List.partition Transitions.not_epsi (Transitions.to_list a.trans) in
  let subpat= SubstitutionTable.new_substable () in
  let fil= complete_filter autpat a (List.map Transitions.ext2trans trans)
	     (List.map Transitions.ext2epsi epsi) subpat svpat false in
  apply_subst_pattern fil a.trans state2term_table a.trans.stmt []


let patternres_to_string (sol: (term * state * substitutionset) list): string=
  (String.concat "\n" (List.map (function (t,s,es) -> (Term.to_string t)^": "^(State.to_string s)^" -> "^(SubstitutionSet.to_string es)) sol))^"\n"


(* ------------------------------TESTS ----------------------------------------*)

IFDEF TEST THEN print_string "Completion.ml\n\n" END

IFDEF TEST THEN
let s= Substitution.add "x" (Term.State 1) Substitution.empty
let set1= SubstitutionSet.add s SubstitutionSet.empty
let t= SubstitutionTable.new_substable ()
let _= SubstitutionTable.add s (1,1) t
let sv1= new_statevar ()
let _= Automaton.statevar_add sv1 1 "x"
let sv2= new_statevar ()
let tl= List.map (function (Trans t) -> t | (Epsilon _) -> failwith "Completion:unitary testing") (Transitions.to_list (Automaton.aut2.trans))
let substsol= SubstitutionTable.new_substable ()
let substsolres= SubstitutionTable.new_substable ()
let substsol4= SubstitutionTable.new_substable ()
let substsol7= SubstitutionTable.new_substable ()
let substsol9= SubstitutionTable.new_substable ()
let (substsol2: substable)= SubstitutionTable.new_substable ()
let (subst1: substitution)= (Substitution.add "x" (State 2) (Substitution.add "y" (State 1) Substitution.empty)) 
let substens= SubstitutionSet.add subst1 SubstitutionSet.empty
let mt= Transitions.new_merge_table()
let mt2= Transitions.new_merge_table()
let mt3= Transitions.new_merge_table()
END

IFDEF TEST THEN
begin
  print_string "match_one\n";
  asserteq (match_one 1 1 (SubstitutionTable.new_substable()) sv1) set1 "match_one_1";
  asserteq (match_one 1 1 (SubstitutionTable.new_substable()) sv2) (SubstitutionSet.empty) "match_one_2";
  asserteq (match_one 1 1 t sv2) set1 "match_one_3";
  print_string "filter\n\n";
  print_string ("Before filter substsol=\n"^(SubstitutionTable.to_string substsol)^"\n")
end
END


IFDEF TEST THEN
(* aut1= f(a, g(x, y)) and tl contains f(a, g(f(a,b), b)) *)
let stt= Automaton.new_s2t ()
let sol= filter Automaton.aut1 tl [] substsol Automaton.sv false
let sol2= complete_filter Automaton.aut1 Automaton.aut2 tl [] substsol2 Automaton.sv false
let t3= Fsym("f", [Fsym("c",[])]) 
let (aut4b,_,_)= Automaton.term_to_automaton [Fsym("f",[Fsym("a",[])])] [] stt
let tl2= List.map (function (Trans t) -> t | (Epsilon _) -> failwith "Completion:unitary testing") (Transitions.to_list (aut4b.trans))
let t4= Fsym("a",[])
let t5= Fsym("b",[])
let t6= Fsym("c",[])

let t7= Fsym("f", [Fsym("g",[Var("x")]);Fsym("g",[Var("y")])])
let t8= Fsym("f", [Fsym("g",[Fsym("a",[])]);Fsym("b",[])])

let (aut7,sv7,_)= Automaton.term_to_automaton [t7] ["x";"y"] stt
(* aut8 and aut9 are identical but used for distinct tests *)
let (aut8,sv8,_)= Automaton.term_to_automaton [t8] [] stt
let (aut9,sv9,_)= Automaton.term_to_automaton [t8] [] stt

let (aut3,sv3,_)= Automaton.term_to_automaton [t3] [] stt
let (aut4,_,_)= Automaton.term_to_automaton [t4;t5;t6] [] stt
let tl4= List.map (function (Trans t) -> t | (Epsilon _) -> failwith "Completion:unitary testing") (Transitions.to_list (aut4.trans))

let s7= Transitions.new_state aut8.trans
let s8= Transitions.new_state aut8.trans
let _= Transitions.add (Trans ("g", [s7], s8)) aut8.trans
let _= Transitions.add (Trans ("b", [], s7)) aut8.trans

let s7b= Transitions.new_state aut9.trans
let s8b= Transitions.new_state aut9.trans
let _= Transitions.add (Trans ("g", [s7b], s8b)) aut9.trans
let _= Transitions.add (Trans ("b", [], s7b)) aut9.trans

let tl8= List.map (function (Trans t) -> t | (Epsilon _) -> failwith "Completion:unitary testing") (Transitions.to_list (aut8.trans))
let _= Transitions.add (Epsilon (2,1)) aut4.trans
let _= Transitions.add (Epsilon (1,0)) aut4.trans
let sol4= complete_filter aut3 aut4 tl4 [] substsol4 sv3 false
let sol3= complete_filter aut3 aut4 tl2 [] substsol4 sv3 false
let sol7= complete_filter aut7 aut8 tl8 [] substsol7 sv7 false
let _= complete_filter aut7 aut9 tl8 [] substsol9 sv7 false

(* for completion *)

let t1= Fsym("f", [Fsym("a",[]);Fsym("b",[])]) (* f(a,b) *)
	  (* f(x, y) -> f(s(x), s(y)) *)
let (autb,_,_)= Automaton.term_to_automaton [t1] [] stt
let rule1= (Fsym("f", [Var("x");Var("y")]), Fsym("f", [Fsym("s", [Var("x")]);Fsym("s", [Var("y")])]), [])
	  (* s(s(x)) = s(x) *)
let eq= Regular(Fsym("s", [Fsym("s", [Var("x")])]), Fsym("s", [Var("x")]), [])
let vl= ["x";"y"]


(* for completion *)

let t2= Fsym("a",[]) (* a *)
	  (* a -> f(a)  *)
let (autc,_,_)= Automaton.term_to_automaton [t2] [] stt
let rule2= Fsym("a", []),Fsym("f", [Fsym("a", [])]),[]
let vl2= []

(* for completion *)

let t3= Fsym("f",[Fsym("a",[])]) (* f(a) *)
	  (* f(x) -> f(s(x))  *)
let (autd,_,_)= Automaton.term_to_automaton [t3] [] stt
let rule3= (Fsym("f", [Var("x")]),Fsym("f", [Fsym("s", [Var("x")])]),[])
let vl3= ["x"]


(* for completion *)

let t4= Fsym("f",[Fsym("a",[])]) (* f(a) *)
let t5= Fsym("b",[]) 
let t6= Fsym("s", [Fsym("s", [Fsym("o",[])])])

	  (* f(b) -> f(c), a -> b  *)
let (aute,_,_)= Automaton.term_to_automaton [t4;t5] [] stt
let rule4= (Fsym("f", [Fsym("b",[])]),Fsym("f", [Fsym("c", [])]),[])
let rule5= (Fsym("a",[]),Fsym("b",[]),[])
let vl4= []

(* for completion *)

let rule3= (Fsym("f", [Var("x")]),Fsym("f", [Fsym("s", [Var("x")])]),[])
let eq2= Regular(Fsym("s", [Var("x")]), Var("x"), [])

(* for completion *)

let eq3= Regular(Fsym("f", [Var("_");Var("_")]),Fsym("f",[Var("_");Var("_")]),[])

let eq4= Regular(Fsym("s", [Fsym("o",[])]),Fsym("o",[]),[])

let eq5= Regular(Fsym("s", [Fsym("s", [Fsym("s", [Fsym("o",[])])])]),Fsym("s", [Fsym("s", [Fsym("o",[])])]),[])


let emptydelta= Transitions.new_delta () 
END

IFDEF TEST THEN
begin
  print_string ("filter aut1="^(Automaton.to_string Automaton.aut1)^"\n on aut2="^(Automaton.to_string Automaton.aut2)^"\n is equal to "^(SubstitutionTable.list_to_string sol)^"\n");
  asserteq (SubstitutionTable.list_to_string sol) "(q3,q3) -> {{y -> q1, x -> q2}}\n(q0,q0) -> {{}}\n" "";
  asserteq substsol substsolres "NEW substitutions should NOT be added to
     substable"; 
  print_string "complete_filter\n";
  asserteq (match_in_list [1;2;3;4;3;5] [2;2;2;2;5;4] 3 5) true "";
  asserteq (match_in_list [1;2;3;4;5;5] [2;2;2;2;5;4] 3 5) false "";
  print_string ("Complete filter on aut1 + aut2 should fill substsol2="^(SubstitutionTable.to_string substsol2)^"\n");
  asserteq sol2 [((4,4),substens)] "Set of solution w.r.t. final set of automata";
  print_string ("aut3 (lhs)="^(Automaton.to_string aut3)^"\n");
  print_string ("aut4 (redex)="^(Automaton.to_string aut4)^"\n");	  
  print_string ("substsol4= "^(SubstitutionTable.to_string substsol4)^"\n");

  (* Substsol4 est incomplet(?) il manque les substitutions sur q1 et q0*)

  print_string "Complete_filter with new transition f(q0) -> q3 should find the subst\n";
  print_string ("sol3="^(SubstitutionTable.list_to_string sol3)^"\n");
  asserteq sol4 [] "No solutions without rules of tl2";
  (* Test unitaire pour la construction des substitutions modulo epsilon *)
  asserteq sol3 [((1,1),(SubstitutionSet.add Substitution.empty SubstitutionSet.empty))] "Solution found with transitions of tl4 (taking into account epsilon)";
  print_string "merge on substution table\n";
  (* aut7= f(g(x), g(y)) and aut8= f(g(a), b(q2)) -> q3 et g(b) -> q5 *)
  print_string ("on aut7(lhs)="^(Automaton.to_string aut7)^
		"\n\nand on aut8(red)="^(Automaton.to_string aut8)^"\n\n"^
		"we obtain substsol7="^(SubstitutionTable.to_string substsol7)^"\n");
  asserteq sol7 [] "Before merging no substitution can be found";
  let (aut8merged,ml)= Automaton.merge aut8 mt [5,2] in 
  let newsub= merge_subst_one_layer ml substsol7 mt in
    begin
      print_string ("After merging 2 and 5 in aut8, aut8 becomes: "^(Automaton.to_string aut8merged)^"\n and substsol7 becomes: "^(SubstitutionTable.to_string substsol7));
      print_string ("\nAnd the resulting substs are: "^(SubstitutionTable.list_to_string newsub)^"\n");
      asserteq (SubstitutionTable.list_to_string newsub) "(q1,q2) -> {{x -> q4}}\n(q3,q2) -> {{y -> q4}}\n"  "Merging of subst produces only subst";
      print_string ("Merge on substitution solution + merging of state values in subst");
      print_string ("on aut7(lhs)="^(Automaton.to_string aut7)^
		    "\n\nand on aut9(red)="^(Automaton.to_string aut9)^"\n+ Merging of 2=5 and 0=4, we obtain\n");
      let (aut9merged,ml2)= Automaton.merge aut9 mt2 [5,2;4,0] in
      let newcompletesub= merge_subst ml2 aut7 aut9 substsol9 mt2 sv9 in
	begin
	  print_string ("\nResulting full matching new substitution is: "^(SubstitutionTable.list_to_string newcompletesub));
	  asserteq (SubstitutionTable.list_to_string newcompletesub) "(q4,q3) -> {{y -> q0, x -> q0}}\n" "New complete subsitution contains only subst on top, with additional propagation of 0=4";
	  print_string "\nDone!\n"
	end
    end
end
END

IFDEF TEST THEN
let d10=Automaton.aut10.trans
let t10=Fsym("f",[Fsym("g",[Var("_")]);Fsym("g",[Fsym("b",[])])])
let t11=Fsym("g",[Fsym("g",[Var("X")])])
let sl= [(Substitution.add "X" (Fsym("a",[])) Substitution.empty)]

END

IFDEF TEST THEN
begin		
 asserteq (apply_subst_equationlist [t10,t11] sl d10 mt3 [0,6;4,6]) [0,6;4,6] ""
end
END

IFDEF TEST THEN
let dc= new_completion_step aute [rule4;rule5] [] [] vl4 vl4 in
  print_string ("dc= "^(to_string dc)^"\n");
  let (dc1,_)= completion_one_step dc emptydelta in
    print_string ("dc1= "^(to_string dc1)^"\n");
    let (dc2,_)= completion_one_step dc1 emptydelta in
    print_string ("dc2= "^(to_string dc2)^"\n");
      let (dc3,_)= completion_one_step dc2 emptydelta in
	print_string ("dc3= "^(to_string dc3)^"\n");
	asserteq (Transitions.symextlist_to_string dc2.new_trans) "f(q3) -> q4\nc -> q3\n" "adding an epsilon trans should raise a new computation of substitution tables";
	asserteq (dc3.new_trans=[] && dc3.new_epsi=[]) true "The automaton is complete"
END


IFDEF TEST THEN
    let cb= new_completion_step autc [rule2] [] [] vl2 vl2 in
      print_string ("\n\n other one_completion_step \n\n");
      print_string ("cb= "^(to_string cb)^"\n");
    let (c2b,_)= completion_one_step cb emptydelta in
      print_string ("c2b= "^(to_string c2b)^"\n");
    let (c3b,_)= completion_one_step c2b emptydelta in
      print_string ("c3b= "^(to_string c3b)^"\n");
      asserteq c3b.new_trans [] "After 2 completion steps the t.a. is complete";
END

IFDEF TEST THEN
  print_string ("\n\n other one_completion_step \n\n");  
  let cc= new_completion_step autd [rule3] [eq] [] vl3 vl3 in
    print_string ("cc before completion= "^(to_string cc)^"\n");
    let (c2c,_)= completion_one_step cc emptydelta in
      print_string ("cc after 1 completion= "^(to_string c2c)^"\n");
    let (c3c,_)= completion_one_step c2c emptydelta in
      print_string ("cc after 2 completion= "^(to_string c3c)^"\n");
    let (c4c,_)= completion_one_step c3c emptydelta in
      print_string ("cc after 3 completion= "^(to_string c4c)^"\n");      
    let (c5c,_)= completion_one_step c4c emptydelta in
      print_string ("cc after 4 completion= "^(to_string c5c)^"\n");
    let (c6c,_)= completion_one_step c5c emptydelta in
      print_string ("cc after 5 completion= "^(to_string c6c)^"\n")
END


IFDEF TEST THEN
    let c= new_completion_step autb [rule1] [eq] [] vl vl in
    let cstring= (to_string c) in      
    let (c2,_)= completion_one_step c emptydelta in
      print_string "\n\n one_completion_step\n\n";
      asserteq cstring (to_string c) "completio_one_step should not change c";
      print_string ("c before completion="^(to_string c)^"\n");
      print_string ("c after completion="^(to_string c2)^"\n");
    let (c3,_)= completion_one_step c2 emptydelta in
      print_string ("c after 2 completion="^(to_string c3)^"\n");	
    let (c4,_)= completion_one_step c3 emptydelta in
      print_string ("c after 3 completion="^(to_string c4)^"\n");
    let (c5,_)= completion_one_step c4 emptydelta in
      print_string ("c after 4 completion="^(to_string c5)^"\n");
      print_string "merge_one_layer\n\n";
      Transitions.merge_states 2 5 c3.aredex.trans.epsilon c3.aredex.trans.stmt;
      asserteq (SubstitutionTable.list_to_string
		  (merge_one_layer [2] c3.sublhs c3.aredex.trans.stmt [])) "(q2,q5) -> {{y -> q1, x -> q0}}\n" "Merge should produce this unique new subst";
END

IFDEF TEST THEN 
let rule2= (Fsym("a", []),Fsym("f", [Fsym("a", [])]),[])
let rule10= (Fsym("f", [Fsym("a", [])]),Fsym("a", []),[])
let vl2= []
END

(* terme: a et regles a -> f(a), f(a) -> a  *)

IFDEF TEST THEN
let ec= new_completion_step autc [rule2;rule10] [] [] [] [] in
  print_string ("Test merging epsilon cycle\n");
  print_string ("c = "^(to_string ec)^"\n");
  let (ec2,_)= completion_one_step ec emptydelta in
    print_string ("c2= "^(to_string ec2)^"\n");
  let (ec3,_)= completion_one_step ec2 emptydelta in
    print_string ("c3= "^(to_string ec3)^"\n");
  let (ec4,_)= completion_one_step ec3 emptydelta in
    print_string ("c4= "^(to_string ec4)^"\n");
    asserteq (Automaton.to_string ec4.aredex) "States q1 \nFinal States q1 \nTransitions\na -> q1\nf(q1) -> q1\n\nq1 -> q1\n" "adding epsilon rules has performed a complete merge of q0 and q1";
    asserteq ec4.new_epsi [] "";
    asserteq ec4.new_trans [] "Automaton is complete";
    print_string "Verification pour equation s(x)=x\n"
END

(* autd={f(a), rule3=f(x) -> f(s(x)), eq2= s(x)=x *)
IFDEF TEST THEN
    let ed= new_completion_step autd [rule3] [eq2] [] vl3 vl3 
    let (ed2,_)= completion_one_step ed emptydelta
    let (ed3,_)= completion_one_step ed2 emptydelta
END

IFDEF TEST THEN
begin
  print_string ("ed = "^(to_string ed)^"\n");
  print_string ("ed2= "^(to_string ed2)^"\n");
  print_string ("ed3= "^(to_string ed3)^"\n");
  assertneq ed2.new_epsi [] "Automaton is not complete";
  asserteq ed3.new_epsi [] "Automaton is complete";
  asserteq ed3.new_trans [] "Automaton is complete"
end
END

IFDEF TEST THEN
let pat= Fsym("f",[Fsym("s",[Fsym("s",[Fsym("s",[Var "z"])])])])
let pat2= Fsym("f",[Fsym("s",[Fsym("f",[Fsym("s",[Var "z"])])])])
let res= check_intersection ed3.aredex [pat] ["z"]
let res2= check_intersection ed3.aredex [pat2] ["z"]
END

IFDEF TEST THEN
begin
  print_string "Check intersection\n";
  print_string (patternres_to_string res);
  print_string (patternres_to_string res2);
  asserteq res2 [] "No solution for this pattern"; 
  asserteq (patternres_to_string res) "f(s(s(s(z)))): q1 -> {{z -> q2}}\nf(s(s(s(z)))): q3 -> {{z -> q2}}\n" "only two solutions";
  print_string "Completion step: pattern detection";
 (* autd={f(a), rule3=f(x) -> f(s(x)), eq2= s(x)=x *)
  let ee= new_completion_step autd [rule3] [eq2] [pat;pat2] ["x";"z"] ["x";"z"] in
  let (ee2,res1)= completion_one_step ee emptydelta in
  let (ee3,res2)= completion_one_step ee2 emptydelta in
    print_string ("ee = "^(to_string ee)^"\n");
    print_string ("ee2= "^(to_string ee2)^"\n");
    print_string ("ee3= "^(to_string ee3)^"\n");
    print_string "Check intersection during completion\n";
    print_string ("Pattern res1= "^(patternres_to_string res1)^"\n");
    asserteq res1 [] "No pattern found";
    print_string ("Pattern res2= "^(patternres_to_string res2)^"\n");
    asserteq (patternres_to_string res2) "f(s(s(s(z)))): q1 -> {{z -> q2}}\nf(s(s(s(z)))): q3 -> {{z -> q2}}\n" "only two solutions"
end
END

IFDEF TEST THEN
begin
  print_string "Check anonym variables in equations\n";
  let ef= new_completion_step autb [rule1] [eq3] [] ["x";"y";"_"] ["x";"y";"_"] in
    print_string (to_string ef);
    let (ef2,_)= completion_one_step ef emptydelta in
      print_string (to_string ef2);
      let (ef3,_)= completion_one_step ef2 emptydelta in
	print_string (to_string ef3);
	print_string "\nTest regles conditionnelles\n"

end
END



IFDEF TEST THEN 
let t1= Fsym("f", [Fsym("a",[]);Fsym("b",[])]) (* f(a,b) *)
	  (* f(x, y) -> f(s(x), s(y)) *)
let (autb,_,_)= Automaton.term_to_automaton [t1] [] stt
let rule11= (Fsym("f", [Var("X");Var("Y")]),Fsym("g",[Var("X")]),[Joinable(Var("X"),Var("Y"))])
let rule12= (Fsym("a", []),Fsym("b",[]),[])
let pat3= Fsym("g",[Var("_")])
let vl2= ["X";"Y"]
let ef= new_completion_step autb [rule11] [] [pat3] vl2 vl2
let eg= new_completion_step autb [rule11;rule12] [] [pat3] ("_"::vl2) ("_"::vl2)
END

IFDEF TEST THEN
begin
  let (ef1,res1)= completion_one_step ef emptydelta in
    print_string (to_string ef1);
    let (ef2,res2)= completion_one_step ef1 emptydelta in
      print_string (to_string ef2);
      asserteq res1 [] "g(_) is not found, rule does not apply";
      asserteq res2 [] "g(_) is not found, rule does not apply, even after 2 steps";
  let (eg1,res1)= completion_one_step eg emptydelta in
  let (eg2,res2)= completion_one_step eg1 emptydelta in
  let (eg3,res3)= completion_one_step eg2 emptydelta in
  let (eg4,res4)= completion_one_step eg3 emptydelta in
    asserteq res1 [] "After first step conditional rule does not apply";
    asserteq res2 [] "After second step conditional rule applies but pattern is still not found";
    assertneq res3 [] "g(_) is found, condition rule has been applied";
    asserteq eg4.new_trans [] "Automaton is complete after 3 steps";
    asserteq eg4.new_epsi []  "Automaton is complete for epsi after 3 steps"
end
END      


IFDEF TEST THEN
let (autf,_,_)= Automaton.term_to_automaton [t6] [] stt
let fc= new_completion_step autf [rule12] [eq4;eq5] [] [] []
let (init_sym,init_epsi)= Transitions.partition (Transitions.to_list autf.trans) 
let (new_autf,esl,eel)= fixpoint_merging init_sym init_epsi autf autf.trans.stmt fc.et fc.eet fc.aeq fc.sveq fc.subst_for_eq fc.subeq fc.state2term
END

IFDEF TEST THEN
begin
  print_string "Fixpoint merging test\n";
  print_string (Automaton.to_string new_autf)
end
END
