IFDEF TEST THEN open Dumper
  END

(*
The goal of this file is to extract 
data from a parsed specification. 
The parser ensures several properties: 
  - arity of functionnal symbols (in automata, sets, 
  trs, approximations) fits with the signature;
  - an initial set of terms is ground over the set of 
  functionnal symbols;
  - There is no state in the TRS. 
  - There is no variable in transitions of a tree automaton. 
  - Each functionnal symbol or variable has been declared in 
  the Ops or Variables section. 
 

Main function: extraction. 

  It returns (trs,automatalist,approximationlist) where: 
  - trs if the term rewriting system parsed and checked;
  - automatalist is a list of tree automata defined from 
  a sequence of both automata specifications or a terms sets 
  specifications.
  - approximationlist is a list of approximations.

*)



open Global
open Common
open ErrorHandler
open Alphabet 
open Set2
open Transitions
open Term
open Equation
open Rewriting
open Automaton



exception BadAutomaton of string
exception ApproxFormatNotHandled of string
exception Automaton_not_found

(* Counter for relabelling of transitions for the --exact normalization *)
let counter= ref 0


(* Initialisation of correspondance table for "new" states used for --exact
   completion.  Those states should be of smaller index... thus we begin the
   numbering by those ones... *)

let rec init_corres (sl: string list) (d: Transitions.delta) (corres: (string * state) list): (string * state) list=
  match sl with
      [] -> corres
    | s1::rem -> 
	let ns= Transitions.new_state d in
	  init_corres rem d ((s1,ns)::corres)

(*
  The following function must be called before the 
  update_correspondence_table. 
  It returns a set of final states and a correspondence table which must be 
  updated with the list of states of the parsed automaton.

*)

let rec initialize_final_states_set (fsl: string list) (sl:string list) (d:Transitions.delta) (res: stateset) (corres: (string * state) list): (stateset * ((string * state) list)) = 
  match fsl with 
      [] -> (res,corres)
    | cs::rem -> 
      if (List.mem cs sl) then 
	if (List.mem cs rem) then 
	   initialize_final_states_set rem sl d res corres 
	else
	  let ns = Transitions.new_state d in 
	  let nres = Transitions.Stateset.add ns res in 
	  let ncorres = (cs,ns)::corres in 
	    initialize_final_states_set rem sl d nres ncorres 
      else
	raise (BadAutomaton (("A final state is not declared in the states section.")))


(*
  The following function updates the correspondence table. 
The arguments corres is the result of the function 
initialize_final_states_set called with the following parameters: 
fls sl d Stateset.empty []
*)

let rec update_correspondence_table (sl: string list) (fsl: string list) (corres: (string * state) list) (d:Transitions.delta):  (string * state) list= 
  match sl with 
      [] -> corres 
    | cs::rem -> 
	if (List.mem cs rem) then 
	  update_correspondence_table rem fsl corres d 
	else
	  if (List.mem cs fsl) then 
	    update_correspondence_table rem fsl corres d 
	else
	  let ns = Transitions.new_state d in 
	  let ncorres = (cs,ns)::corres in 
	    update_correspondence_table rem fsl ncorres d 
	      

(* 

The following function builds a set of transitions 
from the one parsed. 

*)


let rec transitions_set_building (transitions: (pTerm * pTerm) list) (lst:pTerm list) (corres: (string * state) list) (d: Transitions.delta) : unit = 
  match transitions with 
      [] -> ()
    | (lhs,rhs)::rem -> 
	if (List.mem rhs lst) then 
	  begin
	    match lhs with 
		PFsym(f,l) -> 
		  if (List.for_all (fun x -> List.mem x lst) l) then 
		    begin
		      let _ = Transitions.add (Trans (f,(List.map (fun x -> List.assoc (Global.PTerm.to_string x) corres) l), (List.assoc (Global.PTerm.to_string rhs) corres))) d  in 
			transitions_set_building rem lst corres d;
		    end
		  else
		    raise (BadAutomaton ("Some of states used in transitions are not declared."))
	      | PState(q) -> 
		  if (List.mem lhs lst) then 
		    begin
		      let _  = Transitions.add (Epsilon ((List.assoc q corres), ((List.assoc (Global.PTerm.to_string rhs) corres)))) d in 
		      transitions_set_building rem lst corres d;
		    end
		  else
		    raise (BadAutomaton ("The state "^(Global.PTerm.to_string lhs) ^ " is not declared in the states section."))
		      
	      | _ -> raise (BadAutomaton ("There is a variable in the lhs of a transition."))
	  end
	else
	  raise (BadAutomaton ("The state "^(Global.PTerm.to_string rhs) ^ " is not declared in the states section."))
		  
	




let rec build_a_set_of_transitions_for_each_term (id: string) (l: pTerm list) (d:Transitions.delta) (finalstates: stateset): Automaton.automaton =
match l with 
    [] -> {trans=d;fstates=finalstates}
  | t::rem -> 
      begin
	try(
	  let (qf,_) =  Automaton.normalize_term (Term.pTerm_to_term t) d false in 
	  let nfinalstates = Transitions.Stateset.add qf  finalstates in 
	    build_a_set_of_transitions_for_each_term id rem d nfinalstates
	)
	with
	  _ -> raise (BadAutomaton ("Check the initial term set "^id^": states or variables may occur within."))
      end
	  
	



(* Those function produce a copy of the initial transition function so that
   for each transition t -> q we add t -> ###iq where i is incremented each time
   so that the corresponding transitions fulfill the --exact requirements.
   Since states of the form ###iq will also be the first to be numbered, those
   renamed transitions will be the first to be used for --prior normalization...
   This encode the --exact principle.
*)

let dupplicate_transition (l,r: pTerm * pTerm): (string * (pTerm*pTerm))=
  match r with 
      PState(x) -> 
	let news= "###"^(string_of_int !counter)^x in
	  begin 
	    counter:= !counter +1;
	    (news,(l,PState(news)))
	  end
    | _ -> failwith "Dataextracting:dupplicate_transitions: unexpected term!"


let rec dupplicate_trans_list (l: (pTerm * pTerm) list): (string list *((pTerm*pTerm)list))=
  match l with
      [] -> ([],[])
    | t1::rem ->
	let (news,newt)= dupplicate_transition t1 in
	let (newsl, newtl)= dupplicate_trans_list rem in
	  news::newsl,newt::newtl



(* This function builds a 3-uple (name,automaton,l) 
where name is the identifier of the parsed automaton, 
automaton is the result after filling 
*)



let automaton_construction (aut:pAutomaton): (string * Automaton.automaton * ((string * state ) list))=
  match aut with 
      PAut(n,templst,lfst,prior,temptrans) -> 
(*	let _= print_string "temptrans="; List.iter (function l,r -> print_string ((PTerm.to_string l)^" -> "^(PTerm.to_string r)^"\n")) temptrans in *)
	let d = Transitions.new_delta() in 
 	let empty_states = Transitions.Stateset.empty in 
	let st = (List.map Global.PTerm.to_string templst) in 
	let fst = (List.map Global.PTerm.to_string lfst) in 
	let (corres,lst,trans)= 
	  if !Common.exact_completion then
	    let (news,newt)= dupplicate_trans_list temptrans in
	      (init_corres news d [], 
	       (List.map (function s-> PState s) news)@templst,
		newt@temptrans)
	  else ([],templst,temptrans) 
	in
(*	let _= print_string "trans="; List.iter (function l,r -> print_string ((PTerm.to_string l)^" -> "^(PTerm.to_string r)^"\n")) trans in *)
	let (finalstates,corres2) = initialize_final_states_set fst st d empty_states corres in 
	let correspondence_table =  update_correspondence_table st fst corres2 d in 
(*	let _= print_string "corres=" ; List.iter (function (s,s1)-> print_string (s^"="^(string_of_int s1)^", ")) correspondence_table in *)
	  begin
	    transitions_set_building (prior @ trans) lst correspondence_table d;
	    (n,{trans=d;fstates=finalstates},correspondence_table)
	  end
    | PSet(n,termslist) -> 
	let d = Transitions.new_delta() in 
	let empty_states = Transitions.Stateset.empty in 
	let baut =  build_a_set_of_transitions_for_each_term n termslist d empty_states in 
	  (n,baut,[])
	    
	
let rec autassoc s l: automaton=
  match l with 
      [] -> raise Automaton_not_found
    | (n,a,_)::rem -> if s=n then a else autassoc s rem


let rec approximations_from_parser_format approxs = 
  match approxs with 
      [] -> [] 
    | first::rem -> 
	let app = 
	  match first with 
	      PApp(id,dec,rules) -> 
		(*Norm_rules(id,(List.map Term.pTerm_to_term dec), 
		  (List.map (fun (lhs,rhs,norm) -> 
		  ((Term.pTerm_to_term lhs), (Term.pTerm_to_term rhs), 
		  (List.map (fun (x,y) -> ((Term.pTerm_to_term x),(Term.pTerm_to_term y))) norm)))
		  rules))*)
		raise ((ApproxFormatNotHandled "Normalization rules are not handled for now. "))
	    | PSEquation (id,eql) -> 
		(id, (List.map Equation.pEquation_to_equation eql))
	in
	  app::(approximations_from_parser_format rem)



(* Timbuk Specification Files parsing *)


let extraction = function
    nomFichier -> 
      try (
	let channel = if nomFichier="" then stdin else open_in nomFichier in
	let (trss,automata,patterns,dotpatterns,approxs) =  Parser.timbukspec Lexer.token (Lexing.from_channel (channel))  in             
	  ((List.map (fun (x,y) -> (x,Rewriting.trs_from_parser_format y)) trss), (List.map automaton_construction automata),(List.map Term.pTerm_to_term patterns),
	   (List.map Term.pTerm_to_term dotpatterns),
	   (approximations_from_parser_format approxs))
      )
      with 
	  Rewriting.InclusionTestFailed(s) -> 
	    begin
	      print_string (s^"\n");
	      exit(-1);
	    end
	| BadAutomaton(s) -> 	  
	    begin 
	      print_string (s^"\n");
	      exit(-1);
	    end

	| NotHandledYet(s) -> 
	    begin 
	      print_string (s^"\n");
	      exit(-1);
	    end

	| Equation.ContextualEquationNotHandledYet ->
	    begin
	      print_string ("Sorry, contextual equations are coming next.\n");
	      exit(-1);
	    end

	| Equation.EquationMember(s) -> 
	    begin
	      print_string (s^"\n");
	      exit(-1);
	    end

	| Equation.ProblemWithConditions(s) -> 
	    begin
	      print_string (s^"\n");
	      exit(-1);
	    end
	| Rewriting.RecapError(s)-> 
	    begin
	      print_string (s^"\n");
	      exit(-1);
	    end

	    
	| _ -> ErrorHandler.handleParseError();
;;
(*
	| _ -> ErrorHandler.handleParseError()
*)



IFDEF TEST THEN
(*let resultat_extraction =  extraction "TU-ex0.txt"*)

let rec pretty_string_for_parsed_automata l = 
  match l with 
      [] -> ""
    | (name,aut,_)::rem -> 
	"Automaton "^name^":\n"^(Transitions.to_string aut.trans) ^"\n"^(pretty_string_for_parsed_automata rem)



let rec pretty_string_for_parsed_trss l = 
  match l with 
      [] -> "" 
    | (name,rewr)::rem -> 
	"TRS "^name^":\n"^(Rewriting.to_string rewr) ^"\n"^(pretty_string_for_parsed_trss rem)



let rec pretty_string_for_parsed_data input = 
  match input with 
      (trss,automata,_) -> (pretty_string_for_parsed_trss trss)^"\n"^ (pretty_string_for_parsed_automata automata)

ENDIF




(*--------------------------TESTS---- *)

	  

IFDEF TEST THEN
let d_yoh = Transitions.new_delta ()
let state_spec = ["qa";"qb";"qc";"qd";"qf"]
let final_state_spec = ["qf"]
let set_yoh = Transitions.Stateset.empty

let (set_yoh1,corres) = initialize_final_states_set final_state_spec state_spec d_yoh set_yoh [] 

ENDIF
	
IFDEF TEST THEN 
  print_string "dataextracting.ml\n\n" 
END

IFDEF TEST THEN 
  print_string "Test  initialize_final_states_set \n\n" 
END



IFDEF TEST THEN
asserteq corres [("qf",0)] "";
END      



IFDEF TEST THEN 
  print_string "Test  update_correspondence_table \n\n"
END


IFDEF TEST THEN 

let final_corres =  update_correspondence_table state_spec final_state_spec corres d_yoh
END


IFDEF TEST THEN 

asserteq final_corres [("qd",4);("qc",3);("qb",2);("qa",1);("qf",0)] "";
END





IFDEF TEST THEN 
  print_string "Test  transitions_set_building \n\n";
END



IFDEF TEST THEN 
let parsedStates = List.map (fun x -> PState x) state_spec

let trans_1_yoh = (PFsym("a",[]), PState("qa"))

let trans_2_yoh = (PFsym("f",[PState("qa");PState("qb");PState("qc")]), PState("qa"))

let trans_3_yoh = (PState("qa"), PState("qf"))

let trans_4_yoh = (PFsym("b",[]), PState("qb"))

let trans_5_yoh = (PState("qb"), PState("qc"))

let trans_6_yoh = (PState("qp"), PState("qc"))

let trans_7_yoh = (PFsym("f",[PState("qa");PState("qp");PState("qc")]), PState("qa"))

let trans_set = [trans_1_yoh;trans_2_yoh;trans_3_yoh;trans_4_yoh;trans_5_yoh]

let bad_trans_set_1 = [trans_1_yoh;trans_2_yoh;trans_3_yoh;trans_4_yoh;trans_5_yoh;trans_6_yoh]

let bad_trans_set_2 = [trans_1_yoh;trans_2_yoh;trans_3_yoh;trans_4_yoh;trans_5_yoh;trans_7_yoh]


let test_errors transitions = 
  try(
    transitions_set_building transitions parsedStates final_corres d_yoh;
    "ok";
  )
  with 
      BadAutomaton(res) -> res
	




END



IFDEF TEST THEN 

transitions_set_building trans_set parsedStates final_corres d_yoh;

asserteq (Transitions.to_string d_yoh) "b -> q2\na -> q1\nf(q1,q2,q3) -> q1\n\nq2 -> q3\nq1 -> q0\n" "";


asserteq (test_errors trans_set) "ok" "";

asserteq (test_errors bad_trans_set_2)  "Some of states used in transitions are not declared." "";

asserteq (test_errors bad_trans_set_1) "The state qp is not declared in the states section." "";

asserteq (test_errors ((PVar "X",(PState "qa"))::trans_set)) "There is a variable in the lhs of a transition." "";


END




IFDEF TEST THEN 
  print_string "Test  automaton_construction \n\n";
END




IFDEF TEST THEN 

let terme1_yoh = PFsym("f",[ PFsym("h",[ PFsym("b",[]); PFsym("g",[ PFsym("a",[])])]); PFsym("b",[])])
let terme2_yoh = PFsym("f",[ PFsym("b",[]); PFsym("b",[])])
let terme3_yoh =  PFsym("f",[ PFsym("g",[ PFsym("a",[])]); PFsym("g",[ PFsym("a",[])])])
let spec_Set = PSet("Init",[terme1_yoh;terme2_yoh;terme3_yoh])

let (nom,aut,correspondence) = automaton_construction spec_Set


END


IFDEF TEST THEN 

asserteq (Transitions.to_string aut.trans) "f(q2,q2) -> q6\nf(q0,q0) -> q5\nf(q3,q0) -> q4\nh(q0,q2) -> q3\ng(q1) -> q2\na -> q1\nb -> q0\n\n" "" ;
asserteq (Transitions.Stateset.to_string aut.fstates) "{q4,q5,q6}" "" ;

END

IFDEF TEST THEN

let rec do_for_each_trs l f =
  match l with 
      [] -> "\n"
    | (x,trs)::tail -> 
	x^"\n"^ (f trs) ^ (do_for_each_trs tail f)
	

let rec do_for_each_autom l f =
  match l with 
      [] -> "\n"
    | (x,trs,_)::tail -> 
	x^"\n"^ (f trs) ^ (do_for_each_autom tail f)


let rec do_for_each_equation l = 
  match l with 
      [] -> "\n"
    | (id,eqs)::rem -> 
	id^"\n"^ (Equation.to_string eqs) ^ (do_for_each_equation rem)
	
let rec string_list f l s = 
  match l with 
      [] -> "\n"
    | x::[] -> 
	(f x) ^"\n"
    | x::rem -> 
	(f x) ^ s ^ (string_list f rem s)

let _= print_string "TU-ex0.txt\n" in
let (trs,autom,patterns,dotpatterns,approx) =   extraction "TU-ex0.txt" in 
  begin 
    print_string (do_for_each_trs trs Rewriting.to_string );
    print_string (do_for_each_autom autom Automaton.to_string );
    print_string ("Patterns\n"^(string_list Term.to_string patterns ","));
    print_string ("DotPatterns\n"^(string_list Term.to_string dotpatterns ","));
    print_string (do_for_each_equation approx);
  end
    
ENDIF
