IFDEF TEST THEN open Dumper 
END
open Global
open Common
open Alphabet 
open Transitions
open Term
open Variable 
open Substitution

(* Conditionnal equation system  *)

(* Regular: (f(x)=g(x,_) if x=y 
   Exteq: [f(x,y)=g(_)] -> [g(x)=toto, h(y)=h(x)] 
   Extany: [f(x,y),g(x)] -> [g(x)=toto, h(y)=h(x)] 

   Note that an equation of the form f(x)=g(x) will be encoded by f(x)=g(x) if true and
   [f(x,y)] =>[x=y] will be encoded by [f(x,y)=_] => [x=y]

*)

type equation =  
    Regular of (term * term * (term * term) list) 
  | Exteq of (term * term * (term * term) list)
  | Extany of (term * term * (term * term) list)

type equational_theory = equation list


exception InclusionTestFailed of string;;
exception NotHandledCondition of string;;
exception EquationMember of string;;
exception ConditionsException of (term * term);;
exception ContextualEquationNotHandledYet;;

exception ProblemWithConditions of string ;;


(* Variables in conditions of the equation have to appear in the equation.
   hence, no need to look at them *)

let vars (e: equation): variable list=
  match e with
      Regular(l,r,_) | Exteq(l,r,_) | Extany(l,r,_) -> Common.union (Term.vars l) (Term.vars r)

let free_vars (e: equation): variable list=
  match e with
      Regular(l,r,c) | Exteq(l,r,c) | Extany(l,r,c) ->
	let lvars= Term.vars l in
	let rvars= Term.vars r in
	let condvars= List.fold_right 
			(function (cl,cr) -> function rem -> 
			   Common.union (Term.vars cr) (Common.union (Term.vars cl) rem))
			c [] in
	let lvars2= Common.diff lvars ("_"::(rvars@condvars)) in  (* "_" is not free *)
	let rvars2= Common.diff rvars ("_"::(lvars@condvars)) in
	let condvars2= Common.diff condvars ("_"::(lvars@rvars)) in
	  Common.clean (lvars2@rvars2@condvars2)

  
let rec vars_eq (eq: equational_theory): variable list=
  match eq with 
      [] -> []
    | e::rem -> Common.union (vars e) (vars_eq rem)


let rec conditions_checking  conds lvars= 
  match conds with 
      [] -> () 
    | (t1,t2)::rem -> 
	if ((List.for_all (fun x -> (List.mem x lvars)) ((Term.vars t1) @ (Term.vars t2)))) 
	then 
	  conditions_checking  rem lvars
	else 
	  raise (ConditionsException ((t1,t2)))
  


let equation_to_string (lhs:term) (rhs:term):string = 
  (Term.to_string lhs)^" = "^(Term.to_string rhs)

let rec to_string (rew:equational_theory):string = 
  match rew with 
      [] -> "" 
    | Regular(lhs,rhs,cond)::rem -> 
	(equation_to_string lhs rhs)^
	(if cond=[] then "" else " if ")^
	(String.concat " /\\ " 
	   (List.map (function (t1,t2) -> 
			((Term.to_string t1)^"!="^(Term.to_string t2)))
	      cond))^"\n"^(to_string rem)
    (* Special treatment of Exteq of the form [s] => ... encoded by an equation
       [s=_] => ... *)
    | Exteq(lhs,Var("_"),cond)::rem ->
        "["^(Term.to_string lhs)^"]"^
          (if cond=[] then failwith "Equation:to_string:invalid Exteq equation"
	   else " => ["^(String.concat ", "
			   (List.map (function (t1,t2)->
			     equation_to_string t1 t2)
			      cond))^"]\n"^(to_string rem))
    | Exteq(lhs,rhs,cond)::rem ->
	"["^(equation_to_string lhs rhs)^"]"^
	(if cond=[] then failwith "Equation:to_string:invalid Exteq equation"
	 else " => ["^(String.concat ", "
			 (List.map (function (t1,t2)->
				      equation_to_string t1 t2)
			    cond))^"]\n"^(to_string rem))
    | Extany(lhs,rhs,cond)::rem ->
	"["^(Term.to_string lhs)^","^(Term.to_string rhs)^"]"^
	(if cond=[] then failwith "Equation:to_string:invalid Exteq equation"
	 else " => ["^(String.concat " "
			 (List.map (function (t1,t2)->
				      equation_to_string t1 t2)
			    cond))^"]\n"^(to_string rem))



let apply (s: substitution) (e: equation): equation=
  match e with
      Regular(l,r,cond) -> Regular(Rewriting.subst_apply s l, 
				   Rewriting.subst_apply s r, 
				   List.map (function c1,c2 ->
					       Rewriting.subst_apply s c1,
					       Rewriting.subst_apply s c2) cond)
    | Exteq(l,r,cond) -> Exteq(Rewriting.subst_apply s l, 
				   Rewriting.subst_apply s r, 
				   List.map (function c1,c2 ->
					       Rewriting.subst_apply s c1,
					       Rewriting.subst_apply s c2) cond)
    | Extany(l,r,cond) -> Extany(Rewriting.subst_apply s l, 
				   Rewriting.subst_apply s r, 
				   List.map (function c1,c2 ->
					       Rewriting.subst_apply s c1,
					       Rewriting.subst_apply s c2) cond)


(* remove unused variables and replace them by anonymous ones *)
let rec simplify (e: equation):equation= 
  let freev= free_vars e in
  let renaming= List.fold_right 
		  (function v -> function s ->
		     Substitution.add v (Var("_")) s)
		  freev Substitution.empty in
    apply renaming e


(* Check syntax and simplify an equation *)

let check (e: equation): equation=
  let simpe= simplify e in
    match e with
	Regular(lhs,rhs,cond) ->
	  (if ((lhs = Var("_")) or (rhs = Var("_"))) then 
	    raise (EquationMember (("Left and right hand side of an equation must be different of '_'.")))
	  else
	    try(
	      begin
		conditions_checking  cond ((Term.vars lhs) @ (Term.vars rhs) @ ["_"]);
(*		Term.non_linearity_checking lhs;
		Term.non_linearity_checking rhs; *)
		simpe
	      end
	    )
	    with 
		ConditionsException(t1,t2) -> raise ((ProblemWithConditions ("Some variables of the condition "^(Term.to_string t1) ^" != "^(Term.to_string t2) ^ " are not included in the set iof variables extracted from the equation "^ (Term.to_string lhs)^" = "^ (Term.to_string rhs))))
	      | Term.Non_linear -> raise (EquationMember ("Left and right hand sides of the equation "^(Term.to_string lhs)^" = "^(Term.to_string rhs)^" have to be linear. ")))

      | Exteq(Var(_),Var(_),_) | Extany(Var(_),Var(_),_) -> raise (EquationMember "Both left and right hand side of an equation cannot be reduced to a variable")
      | Exteq(lhs,rhs,tl) | Extany(lhs,rhs,tl) -> 
	  try(
	    begin
	      Term.non_linearity_checking lhs;
	      Term.non_linearity_checking rhs;
	      List.iter (function (t1,t2) -> Term.non_linearity_checking t1;Term.non_linearity_checking t2) tl;
	      if (List.exists (function t1,t2 -> t1= (Var "_") || t2=(Var "_")) tl)
	      then raise (EquationMember ("Having anonymous or free variables in merging equation lists of contextual equations is forbidden: "^(to_string [e])))
	      else simpe
	    end)
	  with 
	    Term.Non_linear -> raise (EquationMember ("Left and right hand sides of the equation "^(Term.to_string lhs)^" = "^(Term.to_string rhs)^" have to be linear. "))

	  

(* Conversion of parser format to equation type of TImbuk *)

let pEquation_to_equation c = 
  match c with (* Regular equations *)
      PEquation(lhs,rhs,cond) -> 
	let new_lhs = (Term.pTerm_to_term lhs) in 
	let new_rhs = (Term.pTerm_to_term rhs) in 
	let new_conds = 
	  (List.map (fun x -> 
		       begin
			 match x with 
			     PNotEqual(t1,t2) -> ((Term.pTerm_to_term t1),(Term.pTerm_to_term t2
									  ))
			   | _ -> raise ((NotHandledCondition "Only disequalities are expected in conditions of equations."))
			       
		       end) 
	     cond) 
	in 
	  check (Regular(new_lhs, new_rhs, new_conds))


    (* Contextual or Extended equations *)
    | PExteq(lhs,rhs,tl) ->
	let new_lhs = (Term.pTerm_to_term lhs) in 
	let new_rhs = (Term.pTerm_to_term rhs) in 
	let new_tl= (List.map (function (t1,t2) -> ((Term.pTerm_to_term t1),(Term.pTerm_to_term t2))) tl) in
	  if tl=[] then raise (EquationMember ("List of merging equations of contextual equation ["^(Term.to_string new_lhs)^"="^(Term.to_string new_rhs)^"] -> [] cannot be empty!"))
	  else 
	    check (Exteq(new_lhs, new_rhs, new_tl))

    | PUnique(lhs,tl) ->
	let new_lhs = (Term.pTerm_to_term lhs) in 
	let new_tl= (List.map (function t1,t2 -> (Term.pTerm_to_term t1),(Term.pTerm_to_term t2)) tl) in
	  if tl=[] then raise (EquationMember ("List of merging equations of contextual equation ["^(Term.to_string new_lhs)^"] -> [] cannot be empty!"))
	  else 
	    check (Exteq(new_lhs, (Var "_"), new_tl))
    | PAny(lhs,rhs,tl) ->
	let new_lhs = (Term.pTerm_to_term lhs) in 
	let new_rhs = (Term.pTerm_to_term rhs) in 
	let new_tl= (List.map (function t1,t2 -> (Term.pTerm_to_term t1),(Term.pTerm_to_term t2)) tl) in
	  if tl=[] then raise (EquationMember ("List of merging equations of contextual equation ["^(Term.to_string new_lhs)^"="^(Term.to_string new_rhs)^"] -> [] cannot be empty!"))
	  else 
	    check (Extany(new_lhs, new_rhs, new_tl))



	  
(* ----------------------------------- Tests -------------------------------------------*)


IFDEF TEST THEN print_string "Equation.ml\n\n" END

IFDEF TEST THEN
let t1= Fsym("f", [Var("X");Var("Y")])
let t2= Fsym("g",[Var("X")])
let t3= Fsym("f", [Var("X");Var("Z")])
let t4= Fsym("f", [Var("X");Var("_")])
let e1 = Regular(t1,t2,[t1,t2])
let e2= Regular(t1,t2,[])
let e3= Regular(t3,t2,[])
let e4= Regular(t4,t2,[])
END


IFDEF TEST THEN
print_string "Testing free_vars and simplify functions\n";
asserteq (free_vars e1) [] "";
asserteq (free_vars e2) ["Y"] "";
asserteq (simplify e3) e4 "";
END
