(* ocamlbuild -cflags -dtypes -pp "camlp4o pa_macro.cmo -DTEST" main.byte *)
(* ocamlbuild -pp "camlp4o pa_macro.cmo -UTEST" main.native *)

open Common
open Global
open Term
open Equation
open Completion
open Dataextracting
open Transitions
open Automaton


(* Verbose level: 
   - 0 number of completion steps and number of transitions + epsi to add
   - 1: 0+ the value of transitions to add
   - 2: 1+ the name and position of applied rules and equations + result of normalisation + result of merging
   - 3: 2+ many internal structures are printed

*)


let start_time=ref 0.0

(* Simple dot output function for completion *)

(* producing dot edges *)
let reltostring q es=
  Stateset.fold 
    (function q2 -> function s2 ->
       (State.to_string q)^" -> "^(State.to_string q2)^";\n"^s2)
    es ""
  
let sonrel (sts: stateset) (d: delta): (state * stateset) list=
  Stateset.fold
    (function s -> function l ->
       let sons= Transitions.all_sons s d in	 
	 (s,sons)::l)
    sts []


let write_dot (filename: string) (acomp: Automaton.automaton) (select: term list): unit=
  let fullepsilist= (Transitions.nonclosed_epsilon_to_list acomp.trans.epsilon) in
    (* States corresponding to states matched by select pattern *)
  let statelist= 
    List.flatten (List.map (function t -> (Automaton.all_run t acomp.trans)) select) 
  in
  let _= if (!Common.verbose>3) then print_string 
    ("List of initial states to write to dot file: "^(String.concat "," (List.map State.to_string statelist))^"\n") in
    (* We get all the predecessors of states w.r.t. epsilon transitions,
       i.e. descendants w.r.t. rewriting *)
  let stateset= 
    List.fold_right 
      (function s -> function set ->
	 Stateset.add_list (s::(Transitions.get_pred s acomp.trans)) set)
      statelist Stateset.empty 
  in
  let _= if (!Common.verbose>3) then print_string 
    ("List of all states to write to dot file: "^(Stateset.to_string stateset)^"\n") 
  in 
  let _= if (!Common.verbose>3) then print_string ("Epsilist before filtering: "^(epsiextlist_to_string fullepsilist)^"\n")
  in
  let epsilist=
    (List.filter (function q1,q2 -> Stateset.mem q1 stateset || Stateset.mem q2 stateset) fullepsilist) 
  in
  let _= if (!Common.verbose>3) then print_string ("Epsilist after filtering: "^(epsiextlist_to_string epsilist)^"\n")
  in
  let sonrel= if !dotson then sonrel stateset acomp.trans else [] in
  let _= if (!Common.verbose>3) then
    print_string (List.fold_right (function (s,es) -> function str -> "sonrel="^(State.to_string s)^","^(Stateset.to_string es)^"\n"^str) sonrel "" )
  in
  let resstring= "digraph compdot {\n"^
		 (String.concat ";\n" 
		    (List.map (function q1,q2 -> 
				 (State.to_string q2)^" -> "^(State.to_string q1))
		       epsilist))^
		 ";\nedge [color=brown3];\n"^
		 (List.fold_right
		    (function (q1,es) -> function s1 ->
		       (reltostring q1 es)^s1)
		    sonrel "")^"\n}" 
  in
  let outchan= open_out filename in
    (try(
       let _= Printf.fprintf outchan "%s" resstring in
	 close_out outchan)
     with x -> let _= close_out outchan in raise x)
					    

let write_file (filename: string) (alpha: (string*int)list) (varset: string list)
  (trs: Rewriting.trs) (ainit: Automaton.automaton) (pl: term list)
  (el: equational_theory) (acomp: Automaton.automaton): unit=
  let resstring= 
    "% Completion time: "^(string_of_float (Sys.time() -. !start_time))^" seconds\n\n"^
    "Ops "^
    (List.fold_right (function s,i -> function rem ->
			s^":"^(string_of_int i)^" "^rem)
       alpha
       ("\n\nVars "^
	(List.fold_right (function x -> function rem ->x^" "^rem)
	   varset
	   ("\n\nTRS R1\n"^(Rewriting.to_string trs)^
	    "\n\nAutomaton Ainit\n"^(Automaton.to_string ainit)^
	    "\nAutomaton Acomp\n"^(Automaton.to_string acomp)^
	    (if (List.length pl)=0 then "" else
	       (if !Common.subpattern then "\nSubPatterns\n" else "\nPatterns\n")^(Term.list_to_string pl)^"\n")^
	   (if (List.length el)=0 then "" else 
	      "Equations Approx\nRules\n"^(Equation.to_string el)^"\n")))))
  in	     				    
  let outchan= open_out filename in
    try(
      let _= Printf.fprintf outchan "%s" resstring in
	close_out outchan)
    with x -> let _= close_out outchan in raise x


(* Main function call for completion: filename and maximal number of completion steps. Results into a bool saying if the automaton is complete, the completed automaton, the number of the last performed completion step and a list of found patterns 
*)

let first_glance (filename: string) (autname: string) (nbstep: int): (bool * Automaton.automaton * int * (Term.term * Common.state * Substitution.substitutionset) list)=
  let (trsl,al,pl,dpl,el)= Dataextracting.extraction filename in
  let alpha= List.map (function a,b -> (a,int_of_string b)) !(Global.alpha) in
  let varset= !(Global.varset) in
    match trsl,al with
	[],_ -> failwith "Main:first_glance: Missing Term Rewriting System\n"
      |_,[] -> failwith "Main:first_glance: Missing Tree Automaton\n"
      |((_,trs)::_), ((_,tmpta,_)::_) -> 
	 let ta= if autname="" then tmpta else try (Dataextracting.autassoc autname al) 
	 with Dataextracting.Automaton_not_found -> failwith "Main:first_glance: No automaton of this name"
	 in
	 let equ= try (let (_,eql)=List.hd el in eql)
	 with _ -> [] in  
	   (* get all variables in patterns+trs on one side and equations *)
	 let trs_and_pl_vars= List.fold_right 
		       (function t -> function l ->
			  Common.union (Term.vars t) l)
		       pl (Rewriting.vars_trs trs) in 
	 let eqvars= (Equation.vars_eq equ) in
	   if !Common.check_only then 
	     let allres= (check_intersection ta pl trs_and_pl_vars) in
	       (* If not SubPatterns are considered we focus on final states and
		  predecessors w.r.t. epsilon trans of final states *)
	     let allfiltered= if !Common.subpattern then allres else
	       let finalandpred= Transitions.get_all_pred ta.fstates ta.trans in
		 List.filter (function (_,s,_) -> Stateset.mem s finalandpred) allres
	     in
	       begin (* no completion, only pattern check *)
	       print_string ("\n\nPattern check:\n--------------\n\n"^
			     (patternres_to_string allfiltered)^"\n");
	       (false,ta,0,[])  (* don't care of the result in that case *)
	     end
	   else (* full completion *)
	     (let c= ref (Completion.new_completion_step ta trs equ pl trs_and_pl_vars eqvars) in
	       let i= ref 1 in
	       let pat= ref [] in  
	       let olddelta= ref (Transitions.new_delta()) in
		 begin
		   if (!Common.verbose>3) then print_string ("init cs="^(Completion.to_string !c)^"\n\n");
		   while not(!c.new_trans=[] && !c.new_epsi=[]) && !i<=nbstep && !pat=[] do
		     let (c2,pat2)= completion_one_step !c !olddelta in
		       print_string ("\n\nCompletion step: "^(string_of_int !i)^
				     "\n---------------------------------------------------------\n");
		       if (!Common.verbose>3) then print_string ("Completion step at: "^(string_of_int !i)^
								 ":\n"^(Completion.to_string !c)^"\n\n");
		       if (!Common.verbose>2) then print_string ("Automaton at step: "^(string_of_int !i)^"\n"^(Automaton.to_string !c.aredex)^"\n");
		       print_string ("New transitions: "^(string_of_int (List.length c2.new_trans))^"\n");
		       print_string ("New epsilon: "^(string_of_int (List.length c2.new_epsi))^"\n\n");
		       flush stdout;
		       if (!Common.verbose>0) then print_string 
			 ("Transitions to add: "^(Transitions.symextlist_to_string c2.new_trans)^"\n\n");
		       if (!Common.verbose>0) then print_string 
			 ((Transitions.epsiextlist_to_string c2.new_epsi)^"\n");
		       (* 		 print_string ("cs="^(Completion.to_string c2));    *)
		       pat:= pat2;
		       olddelta:= !c.aredex.Automaton.trans;
		       c:= c2;
		       i:= !i+1;
		   done;
		   (* Merge a last time if the TA is not complete *)
		   let finalmerged= 
		     if (* (!c.new_trans=[] && !c.new_epsi=[]) || not(!pat=[]) *) false then !c.aredex
		     else
		       let (temp,_,_)=
			 fixpoint_merging !c.new_trans !c.new_epsi !c.aredex !c.aredex.trans.stmt !c.et !c.eet !c.aeq !c.sveq !c.subst_for_eq !c.subeq !c.state2term in
			 temp
		   in
		     (* Write result file *)
		     print_string "Export comp.res file...";
		     write_file "comp.res" alpha varset trs ta pl equ (Automaton.unepsilon finalmerged); 
		     print_string "Done\nExport comp.res.epsi file...";
		     write_file "comp.res.epsi" alpha varset trs ta pl equ finalmerged; 
		     print_string "Done\nExport comp.dot file...";
		     write_dot "comp.dot" finalmerged dpl;
		     print_string "Done\n";
		     (!c.new_trans=[] && !c.new_epsi=[],
		      finalmerged,
		      !i-1,
		      !pat)
		 end)


(* Main program *)	     
(* Main program can be found after unit testing for clarity reasons (outputs of
   unit testing must appear before normal outputs of the program *)




	     (* for debug with -g

		first_glance "nspk2.txt" 5
	     *)


(* ------------------------------TESTS ----------------------------------------*)

IFDEF TEST THEN
let _= print_string "TU-ex1.txt\n" in
let b,_,i,p= first_glance "TU-ex1.txt" "" 6 in
  asserteq b true "ex1 Automaton completed";
  asserteq i 5 "With exactly 5 steps";
  asserteq p [] "No pattern is found";

let _= print_string "TU-ex2.txt\n" in
let b,a,i,_= first_glance "TU-ex2.txt" "" 6 in
  asserteq b true "ex2 Automaton completed";
  asserteq (Stateset.cardinal a.fstates) 1 "The 2 final states have been merged";
  asserteq i 2 "With exactly 2 steps";

let _= print_string "TU-ex3.txt\n" in
let b,_,i,p= first_glance "TU-ex3.txt" "" 6 in
  asserteq b false "Not completed";
  asserteq i 4 "Pattern found on 4th step";
  asserteq (List.length p) 2 "The two patterns are found";

let _= print_string "TU-ex4.txt\n" in
let b,_,i,p= first_glance "TU-ex4.txt" "" 10 in
  asserteq b true "Completed";
  asserteq i 6 "Pattern not found on 6th step";

let _= print_string "TU-ex5.txt\n" in
let b,_,i,p= first_glance "TU-ex5.txt" "" 10 in
  asserteq b false "Not completed";
  asserteq i 2 "Pattern found on 2nd step";
  asserteq (List.length p) 3 "The solutions are found";

let _= print_string "TU-ex6.txt\n" in
let b,_,i,_= first_glance "TU-ex6.txt" "" 4 in
  asserteq b true "Completed!";
  asserteq i 2 "Completed on the 2nd step";

let _= print_string "TU-ex7.txt\n" in
let b,_,i,_= first_glance "TU-ex7.txt" "" 4 in
  asserteq b true "Completed!";
  asserteq i 2 "Completed on the 2nd step"

let _= print_string "TU-ex8.txt\n" in
let b,_,i,_= first_glance "TU-ex8.txt" "" 6 in
  asserteq b false "Not completed!";
  asserteq i 6 "Still not completed on the 6th step"

let _= print_string "TU-ex9.txt\n" in
let b,_,i,_= first_glance "TU-ex9.txt" "" 6 in
  asserteq b true "Completed! and Pattern not Found (look only for final states)";
  asserteq i 4 "Completed on the 4th step"

let _= print_string "TU-ex10.txt\n" in
let b,_,i,p= first_glance "TU-ex10.txt" "" 6 in
  asserteq b true "Completed! and SubPattern found (look in all states)";
  asserteq i 4 "Completed on the 4th step";
  asserteq (List.length p) 1 "One solution found"

let _= print_string "TU-ex11.txt\n" in
let b,_,i,p= first_glance "TU-ex11.txt" "" 2 in
  asserteq b true "Completed! and Pattern found (look predecessors of final states)";
  asserteq i 2 "Completed on the 2nd step";
  asserteq (List.length p) 1 "One solution found"

let _= print_string "TU-ex12.txt\n" in
let b,_,i,p= first_glance "TU-ex12.txt" "" 3 in
  asserteq b true "Completed! and Pattern not found!";
  asserteq i 2 "Completed on the 2nd step";
  asserteq (List.length p) 0 "Zero solution found"
END


(* ------------------------------ Main program ------------------------------ *)



let rec apply_option_list (l: string list): unit=
  match l with
      [] -> ()
    | "--new"::rem -> Common.completion_strategy:=Common.New;
	apply_option_list rem
    | "--mergeloop"::rem -> Common.merge_loop:=true;
	apply_option_list rem
    | "--prior"::rem -> Common.completion_strategy:=Common.Prior;
	apply_option_list rem
    | "--epsi"::rem -> Common.completion_type:=Common.Epsi;
	apply_option_list rem
    | "--noepsi"::rem -> Common.completion_type:=Common.Noepsi;
	(* In noepsi mode we also merge loops *)
	Common.merge_loop:=true;
	apply_option_list rem
    | "--equation"::rem -> Common.approx_type:=Common.Equationnal;
	apply_option_list rem
    | "--noapprox"::rem -> Common.approx_type:=Common.None;
	apply_option_list rem
    | "--check"::rem -> Common.check_only:= true;
	apply_option_list rem
    | "--autname"::s::rem -> Common.autname:=s; 
	apply_option_list rem
    | "--verbose"::s::rem -> 
	(try(
	  let vals=int_of_string s in 
	    if (vals>=0) && (vals<=4) then 
	      begin
		Common.verbose:=vals;
		apply_option_list rem
	      end
	    else
	      failwith ("--verbose option expects 0, 1, 2, 3 or 4 as parameter, given: "^s)
	) with Failure _ -> failwith ("--verbose option expects 0, 1, 2, 3 or 4 as parameter, given: "^s))
    | "--dotson"::rem -> Common.dotson:=true;
	apply_option_list rem
    | "--nodotson"::rem -> Common.dotson:=false;
	apply_option_list rem
    | "--exact"::rem -> Common.exact_completion:=true;
	apply_option_list rem
    | ["--autname"] -> failwith "--autname option should be followed by an automaton name"
    | ["--verbose"] -> failwith "--verbose option should be followed by an integer value between 0 and 4"
    | s::_ -> failwith ("Unknown timbuk option: "^s)



let arg_number= Array.length Sys.argv in
  if (arg_number=1) 
  then print_string "Usage= timbuk [option] <max number of completion steps> <timbuk spec file>
with options=
             --exact                       : uses exact normalization strategy
             --noapprox                    : does not use approximations defined
                                             in the file spec
             --noepsi                      : uses old style completion and also 
                                             sets --mergeloop to true
             --mergeloop                   : merge all the states on any 
                                             epsilon cycle
             --autname <name in spec file> : name of automaton to use as initial
             --check                       : does not perform completion but 
                                             only search for patterns in initial
                                             tree automaton
             --new                         : uses only new states to normalize
             --verbose <value>             : verbose mode 0 (default),1,2,3,4
             --dotson                      : direct subterm edges are exported
                                             the comp.dot file in red

             --epsi(default)               : uses new style completion with
                                             epsilon transitions
             --equation(default)           : uses approximation equations
             --prior(default)              : uses existing states/transitions 
                                             when possible
             --nodotson(default)           : subterm edges are not exported
"
  else 
    let l=ref [] in
      begin
	start_time:= Sys.time();
	if arg_number > 2 
	then
	  begin
	    for i=arg_number-3 downto 1 do
	      l:= Sys.argv.(i)::!l
	    done;
	    apply_option_list !l
	  end;
	try(
	  let nb= int_of_string Sys.argv.(arg_number-2) in
	  let file= Sys.argv.(arg_number-1) in
	    if Sys.file_exists file then
	      let (b,a,i,patres)=first_glance file !Common.autname nb in
		(if not(!Common.check_only) then 
		   begin
		     print_string (Automaton.to_string (Automaton.unepsilon a));
		     print_string ("\nStep: "^(string_of_int i)^"\n");
		     if b then print_string 
		       "\n\nAutomaton complete!
------------------\n";
		     if not (patres=[]) then 
		       begin
			 print_string 
			   "\n\nPattern found!
--------------\n";
			 print_string (patternres_to_string patres)
		       end
		     else if b then print_string "\nProof done!\n-----------\n\n";
		     print_string "\nExported comp.res, comp.res.epsi and comp.dot files\n";
		     print_string ("Completion time: "^(string_of_float (Sys.time() -. !start_time))^" seconds\n")
		   end
		 else ())
	    else print_string "File does not exist!\n")
	with Failure "int_of_string" -> print_string ("Bad number of completion steps: "^(Sys.argv.(arg_number-2)))
      end
	  
