IFDEF TEST THEN open Dumper 
END
open Common
open Alphabet 
open Transitions
open Global
open Term
open Substitution

(* Conditionnal term rewriting system *)


type condition = 
    Equal of term * term 
  | Joinable of term * term




exception IncriminatedCondition of condition;;


let pcondition_to_condition c = 
  match c with 
      PEqual(t1,t2) -> 
	Equal((Term.pTerm_to_term t1),(Term.pTerm_to_term t2))
    | PJoinable(t1,t2) -> Joinable((Term.pTerm_to_term t1),(Term.pTerm_to_term t2))
    | _ -> failwith "Rewriting:pcondition_to_condition: not covered yet"
	
	



type rule = (term * term * condition list)
type trs = rule list



exception InclusionTestFailed of string;;
exception NotHandledYet of string;;
exception RecapError of string;;

let rule_to_string ((lhs, rhs, cond):rule):string = 
  (Term.to_string lhs)^" -> "^(Term.to_string rhs)^
  (if cond=[] then "" else " if ")^
  (String.concat " /\\ " 
     (List.map (function c ->
		  match c with
		      Joinable(t1,t2) -> 
			((Term.to_string t1)^"<->"^(Term.to_string t2))
		    | _ -> failwith "Rewriting:rule_to_string: Not supported yet") cond))

let rec to_string (rew:trs):string = 
  String.concat "\n" (List.map rule_to_string rew)


(*
  The following function tests for a rule lhs -> rhs whether 
  var(rhs) is a subset of var(lhs) or not.
*)


let rec variables_inclusion_test (lhs:term) (rhs:term) : bool = 
  let (x_lhs,x_rhs) = ((Term.vars lhs),(Term.vars rhs)) in
    (List.for_all (fun x -> (List.mem x x_lhs)) x_rhs)


let vars ((l,r,c): rule): variable list= Term.vars l

let rec vars_trs (trs: rule list): variable list=
  match trs with
      [] -> []
    | r::rem -> Common.union (vars r) (vars_trs rem)


let rec variables_inclusion_test_for_cond new_conds vars_l = 
  match new_conds with 
      [] -> ()
    | (Equal(t1,t2))::rem -> 
	if ((List.for_all (fun x -> (List.mem x vars_l)) ((Term.vars t1) @ (Term.vars t2)))) then 
	  variables_inclusion_test_for_cond rem vars_l
	else
	  raise (IncriminatedCondition ((Equal(t1,t2))))

    | (Joinable(t1,t2))::rem -> 
	if ((List.for_all (fun x -> (List.mem x vars_l)) ((Term.vars t1) @ (Term.vars t2)))) then
	  variables_inclusion_test_for_cond rem vars_l
	else
	  raise (IncriminatedCondition ((Joinable(t1,t2))))


(* Is a TRS shallow *)

let rule_shallow ((_,r,_): rule): bool=
  Term.height r <= 1

let rec shallow (r: trs): bool=
  match r with
      [] -> true
    | rule::rem -> (rule_shallow rule)&&(shallow rem)

(*
The following function returns a trs from the parsed data 
only if the parsed rules are of the expected form.

 *)


let rec trs_from_parser_format (rulel: pRule list): trs = 
  match rulel with 
      [] -> [] 
    | Usual(lhs,rhs)::rem -> 
	let (nlhs,nrhs) = ((Term.pTerm_to_term lhs),(Term.pTerm_to_term rhs)) in 
	  if (variables_inclusion_test nlhs nrhs) then 
	    begin
	      try ( 
		Term.non_linearity_checking nlhs;
		(nlhs,nrhs,[])::(trs_from_parser_format rem);
	      )
	      with 
		  Term.Non_linear -> raise (RecapError ("The rule "^(Term.to_string nlhs)^" -> "^(Term.to_string nrhs)^" is not left-linear."))
	    end
	  else
	    raise (InclusionTestFailed ("Vars("^(PTerm.to_string rhs)^") is not a subset of Vars("^(PTerm.to_string lhs)^")"))
    | Conditionnal(lhs,rhs,conds)::rem -> 
	let (nlhs,nrhs) = ((Term.pTerm_to_term lhs),(Term.pTerm_to_term rhs)) in 
	  if (variables_inclusion_test nlhs nrhs) then 
	    let new_conds = (List.map pcondition_to_condition conds) in
	      try(
		begin
		  variables_inclusion_test_for_cond new_conds (Term.vars nlhs);
		  Term.non_linearity_checking nlhs;
		  (nlhs,nrhs,new_conds)::(trs_from_parser_format rem);
		end)
	      with 
		  IncriminatedCondition(Equal(t1,t2)) -> raise (InclusionTestFailed ("Some variables of the condition equal("^(Term.to_string t1)^","^(Term.to_string t2)^") of the rule "^(PTerm.to_string rhs)^" -> "^(PTerm.to_string lhs)^" are not included in vars("^(PTerm.to_string rhs)^")"))
		| IncriminatedCondition(Joinable(t1,t2)) ->  raise (InclusionTestFailed ("Some variables of the condition "^(Term.to_string t1)^"<->"^(Term.to_string t2)^" of the rule "^(PTerm.to_string rhs)^" -> "^(PTerm.to_string lhs)^" are not included in vars("^(PTerm.to_string rhs)^")"))
		| Term.Non_linear -> raise (RecapError ("The rule "^(Term.to_string nlhs)^" -> "^(Term.to_string nrhs)^" is not left-linear."))

	  else
	    raise (InclusionTestFailed ("Vars("^(PTerm.to_string rhs)^") is not a subset of Vars("^(PTerm.to_string lhs)^")"))
   
(*raise (NotHandledYet "Condionnal rules are not yet handled.")*)







(* Application of a substitution *)

let rec subst_apply (s: substitution) (t: term): term=
  match t with
      Fsym(f,largs) -> Fsym(f,List.map (subst_apply s) largs)
    | Var(x) -> 
	(try(
	  Substitution.find x s
	) 
	with Not_found -> Var(x))
    | State(q) -> State(q)

let rec substset_apply (es: substitutionset) (t: term): term list=
  if SubstitutionSet.is_empty es
  then []
  else 
    let s= SubstitutionSet.choose es in
      (subst_apply s t)::(substset_apply (SubstitutionSet.remove s es) t)

      
(* ----------------------------------- Tests -------------------------------------------*)


IFDEF TEST THEN print_string "Rewriting.ml\n\n" END

IFDEF TEST THEN
  let s= Substitution.add "x" Term.tn (Substitution.add "y" Term.tm Substitution.empty)
  let r1= (Term.tmx,Term.t1,[])
  let r2= (Term.tmx,Var("x"),[])
  let r3= (Term.tmx,Fsym("f",[Var("x")]),[])
END

IFDEF TEST THEN 
begin
  print_string ((Term.to_string Term.tx)^"\n");
  print_string ((Substitution.to_string s)^"\n");
  print_string ((Term.to_string (subst_apply s Term.tx))^"\n");
  asserteq (subst_apply s Term.tx) Term.ty "";
  asserteq (rule_shallow r1) false "f(...) -> f(a)";
  asserteq (rule_shallow r2) true "f(...) -> x";
  asserteq (rule_shallow r3) true "f(...) -> f(x)";
end
END
