IFDEF TEST THEN open Dumper
END

open Common
open Alphabet 
open State
open Term
open Map
open Set2


module Subtemp=Map.Make(Variable)
module Substitution= 
struct
  type substitution= term Subtemp.t
  type t=substitution
  exception Inclusion

  let empty= Subtemp.empty
  let add= Subtemp.add
  let equal= Subtemp.equal
  let find= Subtemp.find
  let fold= Subtemp.fold
  let map= Subtemp.map
  let compare (s1: substitution) (s2: substitution)= 
    (Subtemp.compare Variable.compare s1 s2)


  (* Cardinal of a substitution *)

  let cardinal s=
    Subtemp.fold (function _ -> function _ -> function i -> i+1) s 0


  (* s1 and s2 are NO LONGER supposed to be two substitutions with disjoint domains *)
      
  let union s1 s2=
    Subtemp.fold 
      (function x -> function t -> function subst -> 
	 if Subtemp.mem x subst then subst
	 else Subtemp.add x t subst)
      s1 s2
      
  let to_string (s: substitution)=
    let rec aux l=
      match l with
	  [] -> ""
	| [(x,v)] -> (Variable.to_string x)^" -> "^(Term.to_string v)
	| (x,v)::rem -> (Variable.to_string x)^" -> "^(Term.to_string v)^", "^(aux rem)
    in
    let res= ref [] in
    let _= Subtemp.iter (function x -> function v -> res:= (x,v)::!res) s in
      "{"^(aux !res)^"}"


  (* Substitution inclusion *)

  (* fold with exception, in the case where an association is not included
     in s2, in order not to follow all s1... due to unfriendly interface
     of maps: no for_all for instance... *)
  (* Note that variable "_" plays here a specific role. 
     if _ is associated to v1 on one side it may be associated (or not)
     on the other side *)

  let is_included (s1: substitution) (s2: substitution): bool=
    try(Subtemp.fold (function key -> function t -> function b ->
		    try (
		      if key="_" then b
		      else 
			let value= Subtemp.find key s2 in
			  if t=value then b else raise Inclusion)
		    with Not_found -> raise Inclusion)
	  s1 true)
    with Inclusion -> false

  (* Extended substitution equivalence such that {X -> q1, Z -> q3} is equivalent in
     {Y -> q2, Z-> q3}, i.e. substitution are always equivalent on variables they do not
     share.  *)

  let extended_equivalence (s1: substitution) (s2: substitution): bool=
    try(Subtemp.fold (function key -> function t -> function b ->
			try (
			  if key="_" then b
			  else 
			    let value= Subtemp.find key s2 in
			      if t=value then b else raise Inclusion)
			with Not_found -> b)
	  s1 true)
    with Inclusion -> false
end


type substitution=Substitution.t


module Substsettemp= Set2.Make(Substitution)
module SubstitutionSet= 
struct
  exception Empty_set= Substsettemp.Empty_set
  type substitutionset= Substsettemp.t
  type t= substitutionset
  let empty = Substsettemp.empty
  let mem= Substsettemp.mem
  let add= Substsettemp.add
  let choose= Substsettemp.choose
  let remove= Substsettemp.remove
  let is_empty= Substsettemp.is_empty
  let diff= Substsettemp.diff
  let filter= Substsettemp.filter
  let exists= Substsettemp.exists
  let singleton= Substsettemp.singleton
  let elements= Substsettemp.elements
  let union= Substsettemp.union

  (* Combination of substitionset list *)
  (* combine of [{{x -> q1},{x -> q2}};{{y->q3,z->q4}}] = 
     {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}}
     if one set of susbt is empty, so is the combination *)

  let rec combine (sl: substitutionset list): substitutionset=
    let rec one_subst_combine (s: substitution) (s2: substitutionset): substitutionset=
      if Substsettemp.is_empty s2 then Substsettemp.empty
      else 
	let first_sub= Substsettemp.choose s2 in
	  Substsettemp.add (Substitution.union s first_sub) 
	    (one_subst_combine s (Substsettemp.remove first_sub s2) )
    in
    let rec substens_combine (s1: substitutionset) (s2: substitutionset)=
      if Substsettemp.is_empty s1 then Substsettemp.empty
      else 
	let first_elt = Substsettemp.choose s1 in
	let combine_first= (one_subst_combine first_elt s2) in
	  Substsettemp.union 
	    combine_first 
	    (substens_combine (Substsettemp.remove first_elt s1) s2) 
    in
      match sl with
	  [] -> Substsettemp.empty
	| [s] -> s
	| s::rem ->  
	    if Substsettemp.is_empty s then Substsettemp.empty
	    else 
	      let res_rem= combine rem in
		if Substsettemp.is_empty res_rem 
		then Substsettemp.empty
		else substens_combine s res_rem
	      
  let to_string= Substsettemp.to_string

(* Matching substitution checks if at least one substitution of es1 is included in
   one substitution of es2 (or one of es2 is included in one of es1) 
   Note also that substitutions are not compared on variables not occurring in
   both es1 and es2. For instance, {{X -> q1}} and {{Y -> q2}} is matching. This
   is due to the fact that it is necessary for equations of the form s(X)=s(Y) where
   any substitution on X and Y is valid.
*)

(* Matching substitution w.r.t. extended_equivalence *)

let matching_substitutionset (es1: substitutionset) (es2: substitutionset): bool= 
  if is_empty es1 then false
  else 
    if is_empty es2 then false
    else 
      exists 
	(function elt -> exists 
	   (function elt2 -> Substitution.extended_equivalence elt elt2) es2) es1

(* Intersection of substitution sets where identity of substitutions is replaced by
   extended_equivalence *)

  let extended_intersection (es1: substitutionset) (es2: substitutionset): substitutionset=
    let rec aux sl1 sl2 sres=
      match sl1 with 
	  [] -> sres
	| s1::rem -> (* Solutions are union of substitutions of es1 and es2 such that 
			they are identitcal w.r.t. extended_equivalence *)
	    let newsres= List.fold_right
			   (function s2-> function es ->
			      if Substitution.extended_equivalence s1 s2
			      then (add (Substitution.union s1 s2) es)
			      else es)
			   sl2 sres 
	    in
	      aux rem sl2 newsres
    in
      aux (elements es1) (elements es2) empty

end

type substitutionset= SubstitutionSet.t


module SubstitutionTable=
struct 
  (* in implementation of substable set of substitutions are ordered by redex
     state then by rule state. But presentation (with to_string for instance) is
     just the opposite. The reason for ordering table by redex state firt is
     that merge applied on redex states merges substitutions for rule states. It
     is thus important to be abble to find all subtables for 2 distinct redex
     states in order to merge them *)
  
  type substable= ((state, (state, substitutionset) Hashtbl.t)Hashtbl.t)
  type t= substable
  let new_substable()= Hashtbl.create Common.max_state_symb
  exception Couple_not_found
    
  (* Finding a substitution s in a table t for a given couple of states cq.
     Throws Couple_not_found if cq is not in the table  *)
	  
  let find ((qlhs,qred): state * state) (t: substable): substitutionset=
    try Hashtbl.find (Hashtbl.find t qred) qlhs
    with Not_found -> raise Couple_not_found

  let mem (cq: state * state) (s: substitution) (t: substable): bool=
    try (SubstitutionSet.mem s (find cq t))
    with Couple_not_found -> false

(* Adding of a substitution s in a table t for a given couple of states cq.
   Returns true if the substitution was not already in the table *)

  let add (s: substitution) ((qlhs,qred): state * state) (t: substable): bool=
    try 
      (let sub_by_qlhs= Hashtbl.find t qred in
	 try
	   (let ens_subst= Hashtbl.find sub_by_qlhs qlhs in
	      if SubstitutionSet.mem s ens_subst then false
	      else 
		begin
		  Hashtbl.replace sub_by_qlhs qlhs (SubstitutionSet.add s ens_subst);
		  true
		end)
	 with Not_found -> 
	   begin
	     Hashtbl.add sub_by_qlhs qlhs (SubstitutionSet.singleton s);
	     true
	   end)
    with Not_found ->
      let newsub_by_qlhs= Hashtbl.create Common.max_rule_node in
	begin
	  Hashtbl.add newsub_by_qlhs qlhs (SubstitutionSet.singleton s);
	  Hashtbl.add t qred newsub_by_qlhs;
	true
      end

  (* Adding a set of substitution es in a table t for a given couple of
     states cq. Returns the set of substitution that were not already in 
     the table, and the empty set otherwise (in the case where they are
     all in the table. The first state is for lhs, the second one for redex *)

  let add_set (es: substitutionset) (cq: state * state) (t: substable): substitutionset=
    SubstitutionSet.filter (function s -> add s cq t) es 
	  

let to_string (t: substable): string=
  let all_sub1= Common.hashtbl_to_list t in
  let all_sub= List.map (function (q,symtable) -> (q, (Common.hashtbl_to_list symtable))) all_sub1 in
  let rec to_string_aux1 l q=
    match l with
	[] -> ""
      | (q2,es)::rem -> "("^(State.to_string q2)^","^(State.to_string q)^") -> "^
			(SubstitutionSet.to_string es)^"\n"^(to_string_aux1 rem q) in
  let rec to_string_aux l=
    match l with 
	[] -> ""
      | (q,l)::rem -> (to_string_aux1 l q)^(to_string_aux rem)
  in
    to_string_aux all_sub


  let list_to_string (l: ((state * state) * substitutionset) list): string=
        List.fold_right (function ((q1,q2),es) -> function s ->
		       ("("^(State.to_string q1)^","^(State.to_string q2)^") -> "^
			(SubstitutionSet.to_string es)^"\n"^s))
      l ""


  let copy (t: substable): substable=
    Common.hashtbl_copy_level2 t

end

type substable= SubstitutionTable.t

(* ------------------------------TESTS ----------------------------------------*)

IFDEF TEST THEN print_string "Substitution.ml\n\n" END

IFDEF TEST THEN 
open Substitution 
open SubstitutionSet
open SubstitutionTable
END

IFDEF TEST THEN
  let sub1= (Substitution.add "y" (Term.State 3) 
		 (Substitution.add "z" (Term.State 4)
		    Substitution.empty))
  let sub2= (Substitution.add "y" (Term.State 3) Substitution.empty)
  let sub3= (Substitution.add "y" (Term.State 3) 
		 (Substitution.add "z" (Term.State 5)
		    Substitution.empty))
  let sub4= Substitution.add "z" (Term.State 1) Substitution.empty
  let sub5= Substitution.add "_" (Term.State 2) Substitution.empty
  let sub6= Substitution.add "_" (Term.State 1) sub1
  let sub8= Substitution.add "_" (Term.State 2) sub3

  let s=Substitution.empty
  let setsubstempty= SubstitutionSet.add Substitution.empty SubstitutionSet.empty
  let s4=Substitution.empty
  let stest= (SubstitutionSet.add 
		 (Substitution.add "x" (Term.State 1) Substitution.empty)
		 SubstitutionSet.empty)

	   (* set2b= {{x -> q2}} *)
  let set2b= (SubstitutionSet.add 
		 (Substitution.add "x" (Term.State 2) Substitution.empty)
		 SubstitutionSet.empty)

	   (* set2= {{x -> q1},{x -> q2}} *)
  let set2= SubstitutionSet.add 
	      (Substitution.add "x" (Term.State 1) Substitution.empty)
	      (SubstitutionSet.add 
		 (Substitution.add "x" (Term.State 2) Substitution.empty)
		 SubstitutionSet.empty) 
	      (* set3= {{y->q3,z->q4}} *)
  let set3= SubstitutionSet.add
	      (Substitution.add "y" (Term.State 3) 
		 (Substitution.add "z" (Term.State 4)
		    Substitution.empty))
	      SubstitutionSet.empty
	      (* set4= {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}} *)
  let set4= SubstitutionSet.add
	      (Substitution.add "x" (Term.State 1) 
		 (Substitution.add "y" (Term.State 3) 
		    (Substitution.add "z" (Term.State 4)
		       Substitution.empty)))
	      (SubstitutionSet.add
		 (Substitution.add "x" (Term.State 2) 
		    (Substitution.add "y" (Term.State 3) 
		       (Substitution.add "z" (Term.State 4)
			  Substitution.empty)))
		 SubstitutionSet.empty)
END

IFDEF TEST THEN
  let _= print_string ((Substitution.to_string s)^"\n") in
  let s2= Substitution.add "x" Term.tn s in
  let s3= Substitution.add "y" Term.tm s2 in
  let s5= Substitution.add "y" Term.tm s4 in
  let s6= Substitution.add "x" Term.tn s5 in
  let set1= SubstitutionSet.add s6 SubstitutionSet.empty in
  let setemptysub= SubstitutionSet.add Substitution.empty SubstitutionSet.empty in
  let t1= SubstitutionTable.new_substable () in
    begin
      print_string ((Substitution.to_string s3)^"\n" );
      (* order of associations have no effect on subst values *)
      asserteq (compare s3 s6) 0 "order of assocications have no effect on subst values";
      asserteq (SubstitutionSet.add s3 set1) set1 "adding equivalent substitutions to a set has no effect";
      (* Combination of substitionset list *)
      (* combine of [{{x -> q1},{x -> q2}};{{y->q3,z->q4}}] = 
	 {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}} *)
      print_string ("\n\nCombine set2="^(SubstitutionSet.to_string set2)^" with set3="^(SubstitutionSet.to_string set3)^" is "^(SubstitutionSet.to_string (SubstitutionSet.combine [set2;set3]))^" and should be equal to "^(SubstitutionSet.to_string set4)^"\n\n");
      asserteq (SubstitutionSet.combine [set2;set3]) set4 "combine of [{{x -> q1},{x -> q2}};{{y->q3,z->q4}}] = {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}}";
      asserteq (SubstitutionSet.combine [set2;setsubstempty]) set2 "combine of [{{x -> q1},{x -> q2}};{{}}] should be equal to [{{x -> q1},{x -> q2}}]";
      let _= SubstitutionTable.add_set setemptysub (1,2) t1 in
      let _= SubstitutionTable.add_set setemptysub (2,3) t1 in
      let _= SubstitutionTable.add_set set2b (4,5) t1 in
	begin
	  asserteq (SubstitutionTable.find (1,2) t1) setemptysub "addition in substitution table";
	  asserteq (SubstitutionTable.find (2,3) t1) setemptysub "addition in substitution table";
	  asserteq (try (let _ = SubstitutionTable.find (5,5) t1 in false) with Couple_not_found -> true) true "(5,5) should not be already in the table";
	  asserteq (SubstitutionTable.add_set set2 (4,5) t1) stest "Adding a new set of substitution covering the first one should give the set of new substitutions";
  	  asserteq (SubstitutionTable.add (Substitution.add "x" (Term.State 1) Substitution.empty) (4,5) t1) false "Adding a subst already present returns false";
    	  asserteq 
	    ((SubstitutionTable.add (Substitution.add "x" (Term.State 1) Substitution.empty) (6,6) t1) &&
	     ((SubstitutionTable.find (6,6) t1)=stest)) true "Adding a subst not present returns true and adds it in the table";
	  print_string "cardinal\n";
	  asserteq (cardinal s4) 0 "";
	  asserteq (cardinal sub1) 2 "";
	  print_string "is_included\n";
	  asserteq (is_included Substitution.empty sub1) true "";
	  asserteq (is_included sub1 sub1) true "";
	  asserteq (is_included sub2 sub1) true "";
	  asserteq (is_included sub1 sub2) false "";
	  asserteq (is_included sub1 sub3) false "";
	  asserteq (is_included sub2 sub4) false "";
	  asserteq (is_included sub4 sub2) false "";
	  asserteq (is_included sub5 sub5) true "";
	  asserteq (is_included sub5 sub6) true "";
	  asserteq (is_included sub5 sub6) true "";
	  asserteq (is_included sub6 sub8) false "";
	  print_string "extended_equivalence\n";
	  asserteq (extended_equivalence sub5 sub6) true "";
	  asserteq (extended_equivalence sub1 sub2) true "on both directions";
	  asserteq (extended_equivalence sub2 sub1) true "on both directions";
	  asserteq (extended_equivalence sub1 sub3) false "";
	  asserteq (extended_equivalence sub4 sub2) true "No variables in common";
	  print_string "extended_intersection\n";
	  asserteq (extended_intersection set3 set4) set4 ""
	end
    end
END
