(* We assume that all hashtbl have unique occurrences of keys *)
(* Otherwise, table copies for transtable, satable etc. is not correct. 
   (older keys may be copied such that they became the more recent keys *)

IFDEF TEST THEN 
open Dumper 
ENDIF

open Common
open Alphabet 
open Set2

type alphabet= Alphabet.t

exception NotCovered
 
let average_nb_symb_by_state = 3 (* basic size of hashtbl for a given symbol *)
let average_nb_trans_by_state = 3 (* basic size of hashtbl for a given symbol *)


let rec state_list_to_string (ls: state list): string=
  match ls with
      [] -> ""
    | s::[]-> State.to_string s
    | s::rem -> (State.to_string s)^","^(state_list_to_string rem)

	       
module Statelist=
struct
  type t= state list
  let compare= compare
  let to_string = state_list_to_string
  let equal= (=)
end

(* we have one table by destination state (unique symbol), by symbol that gives
   a set of state list. *)

module Statetable= Set2.Make(Statelist)
type table= Statetable.t

(* the epsilon graph (closed by transitivity) *)

type epsilonmatrix= bool array array

(* table.(q2).(q1) is true if q1->q2. This is encoded in this way
   in order to get with table.(q2) all the predecessors of q2... *)

(* edispo contains the name of the next free state *)
(* A state is 'free' when it occurs in no transition/substitution
   solution... otherwise it is not free. *)
(* Merge: ne pas reprendre de nom d'etats anciens.
   Ainsi on peut garder l'info q0=10 et elle reste valable.
   Utile pour les substitutions solution (on les renomme une fois
   qu'on les parcourt/utilise. 
*)

(* a symboltable maps a symbol to its transition table *) 

type symtable = (symbol,table) Hashtbl.t

(* a transtable maps a final state (symbol) to its symbol table *) 

type transtable= (state,symtable) Hashtbl.t


(* a satable (state access table) maps a state qi, to a symbol f, to a dest
   state q to a statetable such that f(...qi...) -> q belongs to this table.
   Constant transitions do not occur in this table *)

type satable= (state, (symbol, (state , table) Hashtbl.t) Hashtbl.t) Hashtbl.t

(* an ftable gives for a symbol, all destination states where there are 
   transition for this symbol *)

module Stateset= Set2.Make(State)
type stateset= Stateset.t

let state_set_replace (s1: state) (s2: state) (set: stateset): stateset=
  Stateset.fold (function elt -> function set1 ->
		   let newelt= if elt=s1 then s2 else elt in
		     Stateset.add newelt set1)
    set
    Stateset.empty

type ftable= (symbol, stateset) Hashtbl.t

(* epsilontable type:
   edispo= list of integer numbers to be used for numbering states
   efilled= list of integer numbers already used for numbering states in the 
            current automaton
   closuretable= matrix containing the permanently maintained transitive closure
                 of epsilon transitions
   dpred= table associating to a state its DIRECT predecessor (no transitive closure)
          w.r.t. epsilon transitions
   dsucc= same for successors.
*)

type epsilontable= {edispo: int list ref; 
		    efilled: int list ref; closuretable: epsilonmatrix ref;
		    dpred: (state,stateset)Hashtbl.t;
		    dsucc: (state,stateset)Hashtbl.t}



let epsilontable_copy (e: epsilontable): epsilontable=
  {edispo= ref e.!edispo;
   efilled= ref e.!efilled;
   closuretable= ref  (Array.map Array.copy e.!closuretable);
   dpred= Hashtbl.copy e.dpred;
   dsucc= Hashtbl.copy e.dsucc}

let epsilontable_equal (e1: epsilontable) (e2: epsilontable): bool=
  (Common.subeq_list e1.!edispo e2.!edispo) &&
  (Common.subeq_list e2.!edispo e1.!edispo) &&
  (Common.subeq_list e1.!efilled e2.!efilled) &&
  (Common.subeq_list e2.!efilled e1.!efilled) &&
  (List.for_all (function i -> 
		   List.for_all (function j ->
				   e1.!closuretable.(i).(j)=e2.!closuretable.(i).(j))
		   e1.!efilled)
     e1.!efilled)
    
(* external trans where the first state is the destination state *)

type external_symtrans= (symbol * state list * state)

type external_epsilon= (state * state)

type external_trans= Trans of external_symtrans | Epsilon of external_epsilon


let not_epsi (t: external_trans): bool= match t with Trans(_) -> true | _ -> false

let ext2trans t= match t with Trans(trans) -> trans | _ -> failwith "Transitions:ext2trans:cannot be an epsilon"

let ext2epsi t= match t with Epsilon(trans) -> trans | _ -> failwith "Transitions:ext2epsi:cannot be an regular trans"

(* Partition list of external_trans into a list of external_symtrams and external_epsilon *)

let partition lt= 
  let rec aux lt ressym resepsi=
    match lt with
	[] -> ressym,resepsi
      | Trans(t)::rem  -> aux rem (t::ressym) resepsi
      | Epsilon(t)::rem -> aux rem ressym (t::resepsi)
  in
    aux lt [] []


(* This table gives for each state its cannonical state representative *)
(* for instance q1 -> q2 means that q1 and q2 can be merged and that
   q1 has to be renamed in q2 *)

type state_merge_table= (state, state) Hashtbl.t

let merge_table_to_string t=
  List.fold_right (function (k,e) -> function s -> ((State.to_string k)^" -> "^
  (State.to_string e))^"\n"^s)
    (hashtbl_to_list t) ""

type delta= {trans: transtable; epsilon: epsilontable; satable: satable; ftable: ftable; stmt: state_merge_table}


(* delta to string *)

(* Epsilon transition added by completion *)
(* We display only states occuring in efilled *)    

let epsilon_to_string (e: epsilontable): string=
  let temp= hashtbl_to_list e.dsucc in
  let aux s1 es acc=
    (Stateset.fold 
       (function s2 -> function l2 ->
	  if List.mem s2 e.!efilled 
	  then (State.to_string s1)^" -> "^(State.to_string s2)^"\n"^l2
	  else l2) es acc)
  in
    List.fold_right (function s1,es -> function l1 ->
		       if List.mem s1 e.!efilled then
			 aux s1 es l1 else l1) temp ""


(* Display the permanently transitively closed epsilon matrix *)
(* We display only states occuring in efilled *)    
let epsilonmatrix_to_string (e: epsilontable): string=
  let res= ref "\n" in
    begin
      List.iter (function i -> 
		   List.iter 
		   (function j ->
(*			if !(e.closuretable).(j).(i) 
			then res:= !res^(State.to_string i)^" -> "^(State.to_string j)^"\n"
			else ()) *)
		      try (
			if !(e.closuretable).(j).(i) 
			then res:= !res^(State.to_string i)^" -> "^(State.to_string j)^"\n"
			else ())
		      with Invalid_argument _ -> failwith ("Out of bounds:"^(string_of_int j)^","^(string_of_int i)))
		   e.!efilled)
	e.!efilled;
      !res
    end


let symext_to_string ((symb,l,dest): external_symtrans): string=
    (Symb.to_string symb)^(if l=[] then "" else "("^(state_list_to_string l)^")")^" -> "^(State.to_string dest)

let epsiext_to_string ((q1,q2): external_epsilon): string=
    (State.to_string q1)^" -> "^(State.to_string q2)

let symextlist_to_string (l: external_symtrans list): string=
  (String.concat "\n" (List.map symext_to_string l))^"\n"

let epsiextlist_to_string (l: external_epsilon list): string=
  (String.concat "\n" (List.map epsiext_to_string l))^"\n"

let epsihash_to_string (t: (state,stateset) Hashtbl.t): string=
  (String.concat "\n" (List.map (function s,es ->
				   "("^(State.to_string s)^","^(Stateset.to_string es)^")") (hashtbl_to_list t)))^"\n"


(* We display only states occuring in efilled *)    

let epsilon_to_list (e: epsilontable): external_trans list=
  let res= ref [] in
    begin
      List.iter (function i -> 
		   List.iter 
		   (function j -> 
		      if !(e.closuretable).(j).(i) 
		      then res:= (Epsilon (i,j))::!res
		      else ())
		   e.!efilled)
	e.!efilled;
      !res
    end


(* adding of an epsilon transition q1->q2 in the symbolic table of epsilon no
   transitive reflexive closure performed on this one *)

let direct_add (e: epsilontable) (q1: state) (q2: state): unit=
  let _= if (!Common.verbose>3) then 
    print_string ("Adding symbolic epsilon transition "^(State.to_string q1)^" -> "^(State.to_string q2)^"\n") in
  let newsucc= try (Hashtbl.find e.dsucc q1) with Not_found -> Stateset.empty in
  let newpred= try (Hashtbl.find e.dpred q2) with Not_found -> Stateset.empty in
    begin
      Hashtbl.replace e.dsucc q1 (Stateset.add q2 newsucc);
      Hashtbl.replace e.dpred q2 (Stateset.add q1 newpred);
      if (!Common.verbose>3) then 
	begin
	  print_string ("Dsucc="^(epsihash_to_string e.dsucc)^"\n");
	  print_string ("Dpred="^(epsihash_to_string e.dpred)^"\n")
	end
    end
          

(* Give epsilon transitions but without closure *)    

let nonclosed_epsilon_to_list (e: epsilontable): external_epsilon list=
  let _= if (!Common.verbose>3) then 
    begin
      print_string ("Epsilon table before nonclosed operation: "^(epsilon_to_string e)^"\n");
      print_string ("States present in the !filled list: "^
		    (String.concat "," (List.map string_of_int e.!efilled))^"\n")
    end
  in
  let temp= hashtbl_to_list e.dsucc in
  let aux s1 es acc=
    (Stateset.fold 
       (function s2 -> function l2 ->
	  if List.mem s2 e.!efilled then (s1,s2)::l2
	  else l2) es acc)
  in
    List.fold_right (function s1,es -> function l1 ->
		       if List.mem s1 e.!efilled then
			 aux s1 es l1 else l1) temp []


(* Gives the list of epsilon transitions having a destination state in dest (no closure) *)

let dest_epsilon_to_list (dest: state list) (d: delta): external_epsilon list=
  let rec aux (dest: state list)(acc: external_epsilon list)=
    match dest with 
	[] -> acc
      | st::rem -> 
	  try
	    (let pred_for_st= Stateset.elements (Hashtbl.find d.epsilon.dpred st) in
	       aux rem 
		 (List.append (List.map (function st2 -> (st2,st)) pred_for_st) acc)
	    )
	  with Not_found -> aux rem acc
  in
    aux dest []


(* Is a table representing a transition whose left-hand side is constant? *)

let constant_table (t: table): bool=
  Statetable.mem [] t


let table_to_string (s: symbol) (dest: state) (t: table): string=
  let rec tts_aux (l: state list list): string=
    match l with
	[] -> ""
      | sl::rem -> 
	  (Symb.to_string s)^"("^(state_list_to_string sl)^") -> "^(State.to_string dest)^"\n"^(tts_aux rem)
  in
    if constant_table t then (Symb.to_string s)^" -> "^(State.to_string dest)^"\n"
    else tts_aux (Statetable.elements t)


(* build a symbolic transition list from a given destination state, symbol and
   table (without modulo epsilon) *)

let table_to_list (s: symbol) (dest: state) (t: table): external_trans list=
  let rec tts_aux (l: state list list): external_trans list=
    match l with
	[] -> []
      | sl::rem -> 
	  (Trans (s, sl,dest))::(tts_aux rem)
  in
    if constant_table t then [(Trans (s,[],dest))]
    else tts_aux (Statetable.elements t)


let table_to_symlist (s: symbol) (dest: state) (t: table): external_symtrans list=
  let rec tts_aux (l: state list list): external_symtrans list=
    match l with
	[] -> []
      | sl::rem -> 
	  (s, sl,dest)::(tts_aux rem)
  in
    if constant_table t then [(s,[],dest)]
    else tts_aux (Statetable.elements t)



let trans_to_string (t: transtable): string=
  let all_trans1= Common.hashtbl_to_list t in
  let all_trans= List.map (function (q,symtable) -> (q, (Common.hashtbl_to_list symtable))) all_trans1 in
  let rec to_string_symboltable q l=
    match l with 
	[] -> ""
      | (f, table)::rem -> (table_to_string f q table)^(to_string_symboltable q rem) in 
  let rec to_string_aux l =
    match l with
	[] -> ""
      | (q, symtable)::rem -> (to_string_symboltable q symtable)^(to_string_aux rem) in
    to_string_aux all_trans


(* trans_to_list only produce regular transitions and no epsilon transition *)

let trans_to_list (t: transtable): external_trans list=
  let all_trans1= Common.hashtbl_to_list t in
  let all_trans= List.map (function (q,symtable) -> (q, (Common.hashtbl_to_list symtable))) all_trans1 in
  let rec to_list_symboltable q l=
    match l with 
	[] -> []
      | (f, table)::rem -> (List.append (table_to_list f q table) (to_list_symboltable q rem)) in 
  let rec to_list_aux l =
    match l with
	[] -> []
      | (q, symtable)::rem -> (List.append (to_list_symboltable q symtable) (to_list_aux rem)) in
    to_list_aux all_trans
    

let to_string (d: delta): string=
  (trans_to_string d.trans)^"\n"^(epsilon_to_string d.epsilon)

let to_list (d: delta): external_trans list=
  (List.append (trans_to_list d.trans) (epsilon_to_list d.epsilon))

let external_to_string (t: external_trans): string=
  match t with
      Trans ((sym,sl,dest)) ->  (Symb.to_string sym)^"("^(state_list_to_string sl)^") -> "^(State.to_string dest) 
    | Epsilon((q1,q2)) -> 
	(State.to_string q1)^" -> "^(State.to_string q2)

let list_to_string (l: external_trans list): string=
  (String.concat "\n" (List.map external_to_string l))^"\n"

let symlist_to_string (l: external_symtrans list): string=
  (String.concat "\n" (List.map symext_to_string l))^"\n"
  

let copy (d: delta): delta=
  {trans= Common.hashtbl_copy_level2 d.trans;
   epsilon= epsilontable_copy d.epsilon;
   satable= Common.hashtbl_copy_level3 d.satable;
   ftable= Hashtbl.copy d.ftable;
   stmt= Hashtbl.copy d.stmt}

IFDEF TEST THEN
let delta_equal (d1:delta) (d2:delta): bool=
  let trans= (Common.hashtbl_equal_level2 d1.trans d2.trans Statetable.equal) in
  let epsi= (epsilontable_equal d1.epsilon d2.epsilon) in
  let satable= (Common.hashtbl_equal_level3 d1.satable d2.satable Statetable.equal) in
  let ftable= (Common.hashtbl_equal_level1 d1.ftable d2.ftable Stateset.equal) in
  let _=
    begin
      print_string "\nDisequality:\n";
      if not trans then print_string "trans\n";
      if not epsi then print_string "epsi\n";
      if not satable then print_string "satable\n";
      if not ftable then print_string "ftable\n"
    end in 
  trans && epsi && satable && ftable
END

(* Table creation *)
(* for constant symbols, state vector is empty, dispo and filled are empty *)

let new_table (): table= Statetable.empty

(* Transition table creation *)
(* A transition table maps deststate->symb->state vector *)
(* for constant symbols, state vector is empty *)

let new_transtable (): transtable=
  Hashtbl.create Common.max_state_symb  (* as many as distinct states, initial value *)

let new_symtable (): symtable=
  Hashtbl.create average_nb_symb_by_state

(* low level adding of a transition to a transtable, we do not check if
   transition is already in the table. For constant transitions, a check is 
   made *) 

(* Only update the transtable, no symtable and satable update *)
let add_fast ((symb,sl,dest): external_symtrans) (t: transtable): unit= 
  try let symtable= Hashtbl.find t dest in
    (try (let table= Hashtbl.find symtable symb in 
	     if sl=[] then () (* add a constant transition, already present *)
	     else 
	       Hashtbl.replace symtable symb (Statetable.add sl table)
	 )
       (* dest state exists but not for that symbol yet *)
     with Not_found -> 
	 Hashtbl.add symtable symb (Statetable.add sl (new_table ()))
    )
      (* dest state and symbol does not exist *)
  with Not_found -> 
    let new_table= (Statetable.add sl (new_table ())) in
    let new_sym= new_symtable () in
      begin
	Hashtbl.add new_sym symb new_table;
	Hashtbl.add t dest new_sym
      end


(* add with copy checking, returns true if transition is not present
   in the transtable, and false if the transition is present. *)
(* Only update the transtable, no symtable and satable update *)

let add_trans ((symb,sl,dest): external_symtrans) (t: transtable): bool=
  try(
    let symtable= Hashtbl.find t dest in
    let table= Hashtbl.find symtable symb in
      if (sl=[] || (Statetable.mem sl table))
      then false (* The transition is already in the table *)
      else 
	begin 
	  Hashtbl.replace symtable symb (Statetable.add sl table);
	  true
	end
  )with
      Not_found -> begin (add_fast (symb,sl,dest) t); true end


(* members for transition tables *)

let mem ((symb,sl,dest): external_symtrans) (t: transtable): bool=
  try(
    let symtable= Hashtbl.find t dest in
    let table= Hashtbl.find symtable symb in
      (sl=[] || (Statetable.mem sl table))
  )with
      Not_found -> false
	 



IFDEF TEST THEN
let satable_to_string s=
  Common.hashtbl_to_string_level3 s ""
  
let ftable_to_string s=
  Common.hashtbl_to_string s ""

let to_string_check (d: delta): string=
  (trans_to_string d.trans)^(epsilon_to_string d.epsilon)^"\n\nSATABLE\n\n"^(satable_to_string d.satable)^"\n\nFTABLE\n\n"^(ftable_to_string d.ftable)^"\n"
 ^"\n\nEPSILON\n\nFilled="^(Dumper.dump d.epsilon.!efilled)^"\nDispo="^(Dumper.dump d.epsilon.!edispo)^"\n\nTable size="^(string_of_int (Array.length d.epsilon.!closuretable))^"\n"^"\nTable="^(epsilon_to_string d.epsilon)^"\n\n" 
END


(* Get all states where a symbol is defined *)

let get_states_symb (s: symbol) (d: delta): stateset=
  try (Hashtbl.find d.ftable s) with Not_found -> Stateset.empty


(* get the table for a given destination state, symbol (without modulo epsilon) *)

let simple_get_table (dest: state) (symb: symbol) (d: delta): table=
  Hashtbl.find (Hashtbl.find d.trans dest) symb 


(* get all transitions for a given destination state, symbol (without modulo epsilon) *)

let simple_get_trans (dest: state) (symb: symbol) (d: delta): external_symtrans list=
  try(
    let table= simple_get_table dest symb d in
      table_to_symlist symb dest table
  )
  with Not_found -> []


(* get all transitions for a given symbol (without modulo epsilon) *)

let simple_get_trans_f (symb: symbol) (d: delta): external_symtrans list=
  try
    (let stateset= Hashtbl.find d.ftable symb in
       List.flatten (List.map (function dest -> simple_get_trans dest symb d)
		       (Stateset.elements stateset))
    )
  with Not_found -> []

(* get all transitions for a given destination state, symbol (without modulo epsilon) *)

let simple_get_trans2 (dest: state) (symb: symbol) (d: delta): table list=
  try(
    [(Hashtbl.find (Hashtbl.find d.trans dest) symb)]
  )
  with Not_found -> []


(* Get all symbolic transitions whose destination is in dest *)

let simple_get_all_trans (dest: state list) (d: delta): external_symtrans list=
  let rec aux (dest: state list) (acc: external_symtrans list): external_symtrans list=
    match dest with
	[] -> acc
      | st::rem -> 
	  try (let list_symtable= Common.hashtbl_to_list (Hashtbl.find d.trans st) in
		 aux rem 
		   (List.append
		      (List.flatten (List.map (function (sym,t) -> table_to_symlist sym st t) list_symtable))
		      acc)
	      ) 
	  with Not_found -> 
	    (* If state in dest is not present in delta we ignore it *)
	    aux rem acc
  in
    aux dest []
    
(* A son of q2 is q if there exists a transition f(...q...) ->q *)

let all_sons (q: state) (d: delta): stateset=
  let rec aux (at: external_symtrans list) (res: stateset): stateset=
    match at with 
	[] -> res
      | (_,sl,_)::rem -> aux rem (Stateset.add_list sl res)
  in
    aux (simple_get_all_trans [q] d) Stateset.empty



let replace_trans (t: transtable) ((_,oldsl,_):external_symtrans) ((f, newsl, qdest): external_symtrans): unit=
  try(
    let symtable= Hashtbl.find t qdest in
    let table= Hashtbl.find symtable f in
      Hashtbl.replace symtable f (Statetable.add newsl (Statetable.remove oldsl table))
  )
  with Not_found -> 
    IFDEF TEST THEN
      failwith ("Transitions:replace_trans: incoherence in transtable. old trans="^(Dumper.dump (f,oldsl,qdest))^"\nnew trans="^(Dumper.dump (f, newsl, qdest))^"\nd="^(Dumper.dump t))
      ELSE
      failwith ("Transitions:replace_trans: incoherence in transtable.")
      ENDIF


(* Epsilon transitions *)

let new_epsilontable(): epsilontable=
  {edispo= ref (inclist 0 (Common.max_state_symb-1));
   efilled= ref [];
   closuretable= ref (Array.make_matrix Common.max_state_symb Common.max_state_symb false);
  dpred= Hashtbl.create Common.max_state_symb;
  dsucc= Hashtbl.create Common.max_state_symb}


(* Fill epsilon matrix m2 with content of m1 up to index l *)

let matrix_blit (m1: epsilonmatrix) (m2: epsilonmatrix) (l: int): unit=
  for i=0 to l-1 do
    Array.blit m1.(i) 0 m2.(i) 0 l
  done
  
let epsilon_extend (e: epsilontable): unit=
  if not(!(e.edispo)=[]) then failwith "Transitions:epsilon_extend: there are still free cells"
  else
    let nb_state= Array.length e.!closuretable in
    let new_array= Array.make_matrix (2*nb_state) (2*nb_state) false in
      (* Cannot use an Array.blit here because it copies arrays with
	 their initial length!... too bad *)
    let _= matrix_blit !(e.closuretable) new_array nb_state in
      begin
	e.edispo:= inclist nb_state ((2*nb_state)-1); 
	e.closuretable:= new_array
      end


(* State labels are integers. In the current implementation, new states
   are always new labels. Even if states are destroyed (by merging for instance)
   their label is not re-concerned. The reason is that substitution are
   not updated systematically and the correct value for a state is the value
   obtained by recursive applications of the state_merge table *)

(* Obtain a new state label *)

let new_state (d: delta): state=
  begin
    if d.epsilon.!edispo=[] 
    then 
	epsilon_extend d.epsilon;	
    match d.epsilon.!edispo with
      [] -> failwith "Transitions:new_state: extension has failed" 
    | i::rem -> 
	begin
	  d.epsilon.edispo:= rem;
	  d.epsilon.efilled:= i::(!(d.epsilon.efilled));
	  i
	end
  end

(* Check if a state is valid *)

let mem_state (d: delta) (s: state): bool=
  List.mem s d.epsilon.!efilled


(* Get the set of defined states *)

let get_all_states (d: delta): stateset=
  let rec aux (l: external_trans list) (set: stateset) : stateset=
    match l with
	[] -> set
      | Trans(f,sl,q)::rem -> 
	  aux rem
	    (Stateset.add q 
	       (List.fold_right (function e -> function s -> 			    
				   Stateset.add e s) sl set))
      | Epsilon(q1,q2)::rem ->
	  aux rem (Stateset.add q1 (Stateset.add q2 set))
  in
    aux (to_list d) Stateset.empty


(* (* Remove cycles in epsilontable *)
let remove_cycles (sl: state list) (e: epsilontable): unit=
  let table= e.closuretable in
    List.iter (function s -> !table.(s).(s)<-false) sl
*)

(* Remove a state in efilled and epsilon table only *)

let state_remove (sl: state list) (e: epsilontable): unit=
  let table= e.closuretable in
  let maxind= (Array.length !table)-1 in
    begin
      e.efilled:= Common.diff e.!efilled sl;
      (* We unset all edges from and towards every states of sl *)
      List.iter (function s -> 
		   for i=0 to maxind do 
		     begin
		       !table.(i).(s)<-false;
		       !table.(s).(i)<-false
		     end done) sl;
    (* e.edispo:= sl@e.!edispo  *)  (* A ajouter si on remet en jeu les states!! *)
  end

(* Specific transitive closure of epsilon graph, returns the list of states
   which are in a cycle, i.e. to be merged. Assumption by adding
   one transition, one can only produce a unique common merging cycle.
   Only if merging_loop option activated. *)
	  
(* Before adding the edge between nq1 and nq2, the graph is supposed to 
   be closed by transitivity *)

(* OPTIMISATION: ne prendre que les etats effectivement renseignes dans 
   efilled *)

let transitive_closure (e: epsilontable)(nq1: int)(nq2: int): int list=
  let t= !(e.closuretable) in
  let _= 
    IFDEF TEST THEN
      begin
	(* Diagonal should be empty (no cycles) if merge_loop activated
	   for efilled states, and states should belong to efilled field *)
	List.iter (function i->
		     if not (List.mem i e.!efilled) then failwith ("Transitions:transtitive_closure: merge on state "^(State.to_string i)^" that is no longer in matrix!")
	     else ();
		     if !(Common.merge_loop) && t.(i).(i)  then failwith ("Transitions:transitive_closure: epsilon graph has cycles on state: "^(State.to_string i)) else ()) e.!efilled
      end
      ELSE ()
      END
  in
  let maxind= (Array.length t)-1 in
  let predsetq2= Array.copy t.(nq1) in
  let _= (predsetq2.(nq1) <- true) in      (* nq1 is now a predecessor of nq2*)
  let cycle= ref [] in
    begin
      (* update all predecessors of nq2 *)
      for i=0 to maxind do 
	t.(nq2).(i)<- t.(nq2).(i) || predsetq2.(i)
      done;
      (* update all predecessors of successors of nq2 *)
      for i=0 to maxind do 
	if t.(i).(nq2) then (* i is a successor of nq2 *) 
	  for j=0 to maxind do (* we add all pred of nq2 to i *)
	    t.(i).(j)<- t.(i).(j) || predsetq2.(j)
	  done
      done;
      if !(Common.merge_loop) then
	for i=0 to maxind do  (* states on the diagonal are cycles to merge *)
	  (* found cycles are instantly removed from the matrix*)
	  if t.(i).(i) then begin cycle:= i::!cycle; t.(i).(i)<-false end
	done;
      !cycle
    end
      

(* get all direct pred of a state *)

let get_direct_pred (s: state) (d: epsilontable): stateset=
  try (Hashtbl.find d.dpred s) with Not_found -> Stateset.empty

(* get all direct succ of a state *)

let get_direct_succ (s: state) (d: epsilontable): stateset=
  try (Hashtbl.find d.dsucc s) with Not_found -> Stateset.empty


(* get all pred of a state *)

let get_pred (s: state) (d: delta): state list=
  let table= d.epsilon.closuretable in
  let maxind= (Array.length !table)-1 in
  let res= ref [] in
  try(
    for i=0 to maxind do
      if !table.(s).(i) then res:= i::!res;
    done;
    !res
  )
  with Not_found -> failwith "Transitions:get_pred: a state maybe not occuring in epsilon table!"


(* Get all pred of a set of states *)

let rec get_all_pred (s: stateset) (d: delta): stateset=
  Stateset.fold (function s -> function set1 -> Stateset.add_list (get_pred s d) set1)
    s s

(* get all succ of a state, it should not contain s itself (if there
   is no cycle in epsilon) *)

let get_succ (s: state) (d: delta): state list=
  let table= d.epsilon.closuretable in
  let maxind= (Array.length !table)-1 in
  let res= ref [] in
  try(
    for i=0 to maxind do
      if !table.(i).(s) then res:= i::!res;
    done;
    !res
  )
  with Not_found -> failwith "Transitions:get_succ: a state maybe not occuring in epsilon table!"



(* Add an epsilon transition + transitive closure of epsilon graph *)
(* add_epsilon returns false if the transition is already present in
   the graph *)

type add_res= Ok of (state list) | Present

(* add an epsilon transition *)
(* q1 and q2 are supposed to be states already present in e.edispo *)
(* add_epsilon does not merge states together! but if the Common.merge_loop
   is activated it will return the list of states on potential new loop created
   by the addition of the new trans *)

let add_epsilon ((q1,q2): external_epsilon) (e: epsilontable): add_res=
  let _= if (!Common.verbose>3) then 
    print_string ("Adding : "^(State.to_string q1)^" -> "^(State.to_string q2)^" to epsilon matrix. It was : "^(epsilonmatrix_to_string e)^"\n") in
  let _=     
    IFDEF TEST THEN
      begin
	print_string "Check add_epsilon\n"; 
	asserteq (List.mem q1 !(e.efilled)) true "";
	asserteq (List.mem q2 !(e.efilled)) true "";
      end
      ELSE ()
      END
  in

  (* BE CAREFUL we encode predecessors and not successors!
     so as to have all predecessors in a single row. Depending on the merge_loop
     flag we add or not cyclic epsilon trans and transitions already covered by
     the closure table (i.e. ->*epsilon) *)
  let finalres=
    if !(Common.merge_loop) then
      begin
	if q1=q2 then Present
	else 
	  if !(e.closuretable).(q2).(q1)
	  then Present
	  else 
	    let _= !(e.closuretable).(q2).(q1)<- true in
	    let res= (transitive_closure e q1 q2) in
	      Ok(res)
      end
    else 
      if Stateset.mem q1 (get_direct_pred q2 e) then Present
      else 
	begin 
	  !(e.closuretable).(q2).(q1) <- true;
	  ignore (transitive_closure e q1 q2);
	  Ok([])
	end
  in 
  let _= if (!Common.verbose>3) then 
    print_string ("... and it becomes: "^(epsilonmatrix_to_string e)^"\n") in
    finalres

(* ask if q1 ->*epsilon q2 *)

let is_pred (q1: state) (q2: state) (d: delta): bool=
  d.epsilon.!closuretable.(q2).(q1)

let new_satable(): satable= Hashtbl.create Common.max_state_symb

let new_ftable(): ftable= Hashtbl.create Common.max_symb

let new_merge_table (): state_merge_table= 
  Hashtbl.create Common.max_state_symb

let new_delta(): delta=
  {trans= new_transtable(); epsilon= new_epsilontable(); satable= new_satable();
  ftable= new_ftable();stmt= new_merge_table ()}


let add_ftable ((sym,sl,dest): external_symtrans) (fa: ftable): unit=
  try (
    let set= Hashtbl.find fa sym in
      if Stateset.mem dest set then ()
      else 
	Hashtbl.replace fa sym (Stateset.add dest set)
  )
  with Not_found -> 
    Hashtbl.add fa sym (Stateset.singleton dest)


(* s1 is the parameter state for which we want to add a table *)

let add_state_satable (sym: symbol) (dest: state) (sl:state list) (sa: satable) (s1:state): unit=
  try(
    let forstate= Hashtbl.find sa s1 in
      try(
	let forsym= Hashtbl.find forstate sym in
	  try(
	    let fordest= Hashtbl.find forsym dest in
	      Hashtbl.replace forsym dest (Statetable.add sl fordest)
	  )
	  with Not_found -> (* destination does not exist already *)
	    Hashtbl.add forsym dest (Statetable.singleton sl)
      )
      with Not_found -> (* symbol -> dest does not exist already *)
	let forsym= Hashtbl.create average_nb_symb_by_state in
	  begin
	    Hashtbl.add forsym dest (Statetable.singleton sl);
	    Hashtbl.add forstate sym forsym
	  end
  )
  with Not_found -> (* parameter -> symbol -> dest does not exist already *)
    let forsym= Hashtbl.create average_nb_symb_by_state in
    let forstate= Hashtbl.create average_nb_symb_by_state in
      begin
	Hashtbl.add forsym dest (Statetable.singleton sl);
	Hashtbl.add forstate sym forsym;
	Hashtbl.add sa s1 forstate
      end


(* Remove a state vector in a satable *)

let remove_state_satable (sym: symbol) (dest: state) (sl:state list) (sa: satable) (s1:state): unit= 
  try(
    let forstate= Hashtbl.find sa s1 in
    let forsym= Hashtbl.find forstate sym in
    let fordest= Hashtbl.find forsym dest in
    let new_fordest= Statetable.remove sl fordest in
      if Statetable.is_empty new_fordest 
      then 
	(* The set of transitions is empty for dest *)
	let _= Hashtbl.remove forsym dest in
	  if Hashtbl.length forsym = 0 
	  then 
	    (* The set of dest -> transitions is empty for sym *)
	    let _= Hashtbl.remove forstate sym in
	      if Hashtbl.length forstate = 0
	      then
		(* The set of sym -> dest -> transitions is empty for sym *)
		let _= Hashtbl.remove sa s1 in ()
	      else ()
	  else () 
      else Hashtbl.replace forsym dest new_fordest
  )with Not_found -> ()

(* Add a transition in a satable w.r.t. to all its state parameters. *)
(* Constant transitions are not added in this table! *)

let add_satable ((sym,sl,dest): external_symtrans) (sa: satable): unit=
  List.iter (function x -> (add_state_satable sym dest sl sa x)) sl


(* Remove a transition in a satable w.r.t. to all its state parameters. *)
(* Constant transitions are not part of this table! *)

let remove_satable ((sym,sl,dest): external_symtrans) (sa: satable): unit=
  List.iter (function x -> (remove_state_satable sym dest sl sa x)) sl


(* general adding of a f(q1..qn)->q transition Returns false if the transition
   is already present in the table. 
   The corresponding states are supposed to be already present in delta.epsilon *)

let add_regular_trans (t: external_symtrans) (d: delta): bool=
  match (add_trans t d.trans) with
      true -> 
	begin
	  add_satable t d.satable;
	  add_ftable t d.ftable;
	  true
	end
    | false -> false (* Transition is already present in the table *)



(* get all transitions for a given destination state, symbol, **modulo epsilon** *)

let get_trans (dest: state) (symb: symbol) (d: delta): external_symtrans list=
  List.append 
    (simple_get_trans dest symb d)
    (List.flatten
       (List.map (function state -> simple_get_trans state symb d) 
	  (get_pred dest d)))


(* get all transitions for a given destination state, symbol, **modulo epsilon** 
   return a table list *)

let get_trans2 (dest: state) (symb: symbol) (d: delta): table list=
  List.append 
    (simple_get_trans2 dest symb d)
    (List.flatten
       (List.map (function state -> simple_get_trans2 state symb d) 
	  (get_pred dest d)))  
      
(* Get all transitions by parameter state: i.e. get all f(...q2...) ->q  given q2 *)

let get_trans_parameter (par: state) (d: delta): external_symtrans list=
  try(
    let all_table= Common.hashtbl_to_list (Hashtbl.find d.satable par) in
      List.fold_right 
	(function (symb,t) -> function l ->
	   let trans_for_s= 
	     List.flatten
	       (List.map (function (dest,vect) -> 
			 List.map 
			 (function sl -> (symb,sl,dest) )
			 (Statetable.elements vect))
		  (Common.hashtbl_to_list t)) in
	     List.append trans_for_s l)
	all_table
	[]
  ) with
      Not_found -> [] (* states corresponding to constants are not in satable *)

let rec get_trans_parameter_list (par: state list) (d: delta): external_symtrans list=
  match par with
      [] -> []
    | s::rem -> List.append (get_trans_parameter s d) (get_trans_parameter_list rem d)


(* get all transitions by top symbol and parameter state (no epsilon resolving) *)
(* symb is supposed not to be a constant symbol *)

let get_trans_sym_parameter (symb: symbol) (par: state) (d:delta): external_symtrans list=
  try(
    let (all_trans: ((state * table) list))= 
      (Common.hashtbl_to_list (Hashtbl.find (Hashtbl.find d.satable par) symb)) in 
      try(
	List.flatten (List.map (function (dest,ts) -> 
		    (List.map (function sl -> (symb, sl, dest))
		       (Statetable.elements ts)))
	    all_trans)
      )
      with Not_found -> failwith "Transitions:get_trans_sym_parameter: incoherence between satable content and transition table"
		 
  )
  with
      Not_found -> []

(*
(* get all transitions by top symbol and state vector (no epsilon resolving) *)
(* symb is supposed not to be a constant symbol. Uses *)

let get_trans_sym_vect (symb: symbol) (vect: state list) (d:delta): external_symtrans list=
  match vect with
      first::rem ->
	(try(
	   let (all_trans: ((state * table) list))= 
	     (Common.hashtbl_to_list (Hashtbl.find (Hashtbl.find d.satable first) symb)) in 
	     List.fold_right 
	       (function (dest,ts) -> function l ->
		  if Statetable.mem vect ts then dest::l
		  else l)
	       all_trans [])
	 with Not_found -> [])
    | [] -> failwith "Transitions:get_trans_sym_vect:cannot be a constant!"
*)

(* get all transitions by top symbol for constant symbols (no epsilon resolving) *)

let get_trans_sym_constant (symb: symbol) (d:delta): external_symtrans list=
  try(
    let (all_dest: state list)= Stateset.elements (Hashtbl.find d.ftable symb) in
      List.flatten (List.map (function dest -> simple_get_trans dest symb d) all_dest)
  )
  with
      Not_found -> []



(* Merge: ne pas reprendre de nom d'etats anciens.
   Ainsi on peut garder l'info q0=10 et elle reste valable.
   Utile pour les substitutions solution (on les renomme une fois
   qu'on les parcourt/utilise. 
*)


(* This table gives for each state its cannonical state representative *)
(* for instance q1 -> q2 means that q1 and q2 can be merged and that
   q1 has to be renamed in q2 *)


(* We use a fixpoint since a state value maybe itself merged! *)

IFDEF TEST THEN
let merged_value (mt: state_merge_table) (s: state): state=
  let rec aux (s: state) (set: stateset): state=    
  try(
    let next=(Hashtbl.find mt s) in
      if Stateset.mem next set 
      then 
	begin
	  print_string ("State merge table="^(merge_table_to_string mt)^"\n");
	  failwith "Transitions:merged_value: Cycle in state merging table!"
	end
      else aux next (Stateset.add s set)
  )
  with
      Not_found -> s
  in
    aux s Stateset.empty
ELSE
let rec merged_value (mt: state_merge_table) (s: state): state=
  try(
    merged_value mt (Hashtbl.find mt s)
  )
  with
      Not_found -> s
ENDIF



(* l gives for each list of states its cannonical representative *)
(* l is supposed to be closed by transitivity +
   si q1 -> q3 et q2 -> q3 alors [q1,q2] -> q3 a la place. *)

(*
let rec add_merged (l: ((state list) * state) list) (mt: state_merge_table): ((state list * state) list)=
  match l with
      [] -> []
    | (sl,s)::rem -> 
	let new_s= merged_value mt s in
	let new_sl= Common.remove new_s sl in
	  if new_sl=[] then add_merged rem mt
	  else
	    begin
	      List.iter (function s2 -> (Hashtbl.add mt s2 new_s)) new_sl;
	      (new_sl,s)::(add_merged rem mt)
	    end
*)


let apply_rename_one_trans (t: transtable) (f: symbol) (dest: state) (mt: state_merge_table) (sl: state list) : state list=
  let newsl= List.map (merged_value mt) sl in
  let _= IFDEF TEST THEN
	   assertneq sl newsl ("state list and merged state lists should not be identical! "^(state_list_to_string sl)^" and "^(state_list_to_string newsl)^" Though state merge table is: "^(merge_table_to_string mt))
	 ELSE 
	   () 
         END
  in
  let _= replace_trans (t: transtable) (f, sl, dest) ((f, newsl, dest): external_symtrans) in
    newsl


(* rename on vector of state list, returns the renamed table 
   sll is the list of state vectors to be renamed *)

let rec apply_rename_translist (t: transtable) (f: symbol) (dest: state) 
(mt: state_merge_table) (sll: state list list): table=
  match sll with
      [] -> Statetable.empty
    | sl::rem -> Statetable.add (apply_rename_one_trans t f dest mt sl)
	(apply_rename_translist t f dest mt rem)


(* Apply rename on a list of symbols associated with their destination states 
   and tables. Side effects: update the satable (remove the old and add the
   new associations *) 

let rec apply_rename_dest_translist (t: transtable) (sa: satable) (mt: state_merge_table) (destlist: (symbol * ((state * table) list)) list): unit=
  match destlist with
      [] -> ()
    | (f,sld)::rem ->
	let newtrans_by_dest= 
	  List.map (function (dest,table) ->
		      begin (* remove all the old trans from the satable *)
			Statetable.iter (function sl -> 
				     remove_satable (f, sl, dest) sa)
			  table;
			(* obtain all the new transtable *)
			(dest, apply_rename_translist t f dest mt (Statetable.elements table))
		      end)
	    sld in
	  begin  (* Add all new transitions to satable *)
	    List.iter (function (dest, new_transtable)->
			 Statetable.iter (function sl -> add_satable (f,sl,dest) sa)
			 new_transtable)
	      newtrans_by_dest;
	    apply_rename_dest_translist t sa mt rem
	  end


(* Rename on transtable and satable (parameter states only) *)

let apply_rename_parameter (t: transtable) (sa: satable) (from: state list) (mt: state_merge_table): unit=
  let (temp: (symbol * ((state,table) Hashtbl.t)) list)=
    (List.flatten
       (List.map 
	  (function q ->
	     try 
	       (Common.hashtbl_to_list (Hashtbl.find sa q))
	     with
		 Not_found -> [])
	  from))
  in
  let (symtable_from: (symbol * ((state * table) list)) list)= 
    List.map (function (f,h) -> (f, Common.hashtbl_to_list h)) temp in
    begin
      (* eliminate all from states of the satable (they have disappeared) *)	 
      List.iter (Hashtbl.remove sa) from;
      apply_rename_dest_translist t sa mt symtable_from
    end


(* Merge trans and satable (parameter only) merge_policy is supposed to be already
   in mt *)

let merge_trans_parameter (t: transtable) (sa: satable) (mt: state_merge_table) (merge_policy: ((state list) * state) list) : unit= 
  List.iter (function (from, _) ->
	       apply_rename_parameter t sa from mt) merge_policy


(* Merge delta for destination states one from to one towards. From -> towards
   is supposed to be already in mt *)

let merge_trans_dest (d: delta) (from: state) (mt: state_merge_table): unit=
  let towards= merged_value mt from in
  try (
    begin
      Hashtbl.iter 
	((function sym -> function vects -> 
	    (Statetable.iter 
	       (function v -> 
		  begin
		    remove_satable (sym, v, from) d.satable;
		    (* add new transitions towards 'towards state' *)
			      let _=add_regular_trans (sym, v, towards) d in ()
			    end)
			 vects)): symbol -> table -> unit)
	(Hashtbl.find d.trans from); (* hashtbl giving for each symbol its set of vectors *)
      Hashtbl.remove d.trans from  (* remove old trans towards 'from' state *)
    end
  )
  with Not_found -> ()
    (* State 'from' does not occur in transition table (no regular transition
       towards this state *)
    



let rec stateset_map f s=
  try(
    let e= Stateset.choose s in
      Stateset.add (f e) (stateset_map f (Stateset.remove e s)))
  with Stateset.Empty_set -> Stateset.empty
    

(* Merge delta for destination states only merge_policy is supposed to be already
   in mt *)

let merge_trans_dest_list (d: delta) (mt: state_merge_table) (merge_policy: ((state list) * state) list): unit=
  begin
    List.iter (function (from, _) ->
		 (List.iter (function one_from -> 
			       (merge_trans_dest d one_from mt)) from))
      merge_policy;
    Common.hashtbl_map d.ftable (function sym -> function stateset ->
				   (stateset_map
				      (function s -> merged_value mt s) 
				      stateset))
  end
	

let merge_symext (mt: state_merge_table)((f,sl,q): external_symtrans) : external_symtrans=
  (f,(List.map (merged_value mt) sl), (merged_value mt q))

let merge_external (mt: state_merge_table)(t: external_trans) : external_trans=
  match t with
      Trans((f,sl,q)) -> Trans((f,(List.map (merged_value mt) sl), (merged_value mt q)))
    | Epsilon(q1,q2) -> Epsilon(merged_value mt q1, merged_value mt q2)


(* Extract the list of transitions concerned by the renaming performed in mt *)

let rec merge_external_symlist (mt: state_merge_table)(l: external_symtrans list)(res: external_symtrans list) : external_symtrans list=  
  match l with 
      [] -> res
    | t::rem -> 
	let new_t= merge_symext mt t in
	  if t=new_t 
	  then merge_external_symlist mt rem res
	  else merge_external_symlist mt rem (new_t::res)


(* Extract the list of transitions concerned by the renaming performed in mt *)

let rec merge_external_epsilist (mt: state_merge_table)(l: external_epsilon list)(res: external_epsilon list) : external_epsilon list=
  match l with 
      [] -> res
    | (q1,q2)::rem -> 
	let (q1b, q2b)= (merged_value mt q1, merged_value mt q2) in
	  if (q1,q2)=(q1b, q2b)
	  then merge_external_epsilist mt rem res
	  else 
	    if q1b=q2b then merge_external_epsilist mt rem res
	    else merge_external_epsilist mt rem ((q1b,q2b)::res)

let rec merge_external_list (mt: state_merge_table)(l: external_trans list)(res: external_trans list) : external_trans list=
  match l with 
      [] -> res
    | t::rem -> 
	let new_t= merge_external mt t in
	  if t=new_t 
	  then merge_external_list mt rem res
	  else 
	    match new_t with   (* Transitions of the form q->q are suppressed *)
		Trans(_) -> merge_external_list mt rem (new_t::res)
	      | Epsilon(q1,q2) -> 
		  if q1=q2 then merge_external_list mt rem res
		  else merge_external_list mt rem (new_t::res)



  
  (* closure of state merging equations. i.e.  [(q1,q2);(q4,q5);(q1,q3)] into
     [([q1;q2],q3);([q4],q5)] for instance *)
(*
  let rec propagate (q1: state) (q2: state) (l: state list list) (rem: state list): state list list=
    match l,rem with
	[],[] -> [[q1;q2]]
      | [],_ -> 
	  if List.mem q1 rem then [q2::rem]
	  else [q1::rem]
      | closure::rem2,[] ->
	  if List.mem q1 closure (* we have found the closure of q1 alone *)
	  then 
	    if List.mem q2 closure 
	    then l  (* q1 and q2 are already in the same closure *)
	    else propagate q1 q2 rem2 closure
	  else
	    if List.mem q2 closure (* we have found the closure of q2 alone *)
	    then propagate q1 q2 rem2 closure
	      (* neither q1 nor q2 *)
	    else closure::(propagate q1 q2 rem2 rem)
      | closure::rem2,_ ->  (* we have found one closure, looking for the other one *)
	  if (List.mem q1 closure)||(List.mem q2 closure) 
	  then 
	    (* q1 and q2 and all equivalent are put in the same closure *)
	    (closure@rem)::rem2 
	  else 
	    closure::(propagate q1 q2 rem2 rem)
*)

(*
(* trivial equations q1=q1 should not be present in me *)

let rec merging_equation_closure (me: (state * state) list) (res: state list list): ((state list) * state)list=
  match me with
      [] -> List.map (function x -> match x with 
			  s::rem -> (rem,s)
			| _ -> failwith "Transitions:merging_equation_closure: bad closure" ) res 
    | (q1,q2)::rem ->
	let _= IFDEF TEST THEN 
		 if q1=q2 then failwith "Transitions:merging_equation_closure: trivial equality present in me"
		 else ()
	       ELSE ()
	       ENDIF in
	merging_equation_closure rem (propagate q1 q2 res [])
*)

(* merging in epsilon matrix: returns the list of states on a cycle *)
(* Checking of cycle is done w.r.t efilled states, but transitive closure
   is done on the whole matrix, to maintain the coherence of epsilon transtions
   for states that have been (partially) merged by one equation application *)

let merge_matrix (e: epsilontable)(from: int list)(towards: int): int list=
  let t= !(e.closuretable) in
    (* We only keep states that are still in epsilon (avoid states already merged) *)
  let _= 
    IFDEF TEST THEN
      begin
	(* Diagonal of filled states should be empty (no cycles) *)
	print_string "Check Transitive closure\n";
	List.iter
	  (function i -> 
	     if not (List.mem i e.!efilled) then failwith ("Transitions:merge_matrix: merge on state "^(State.to_string i)^" that is no longer in matrix!")
	     else ();
	     if t.(i).(i) then failwith ("Transitions:merge_matrix: epsilon graph has cycles on state: "^(State.to_string i)) else ()) e.!efilled 
      end
      ELSE ()
      END 
  in  

  let maxind= (Array.length t)-1 in
  let lfrom= ref from in
    begin
      while not(!lfrom=[]) do
	let prems= List.hd !lfrom in
	  begin
	    t.(towards).(prems)<- true;
	    t.(prems).(towards)<- true;
	    (* update the 'towards' pred and succ with those of the first elt of from *)
	    for i=0 to maxind do 
	      begin
		t.(towards).(i)<- t.(towards).(i) || t.(prems).(i);
		t.(i).(towards)<- t.(i).(towards) || t.(i).(prems)
	      end
	    done;
	    (* update all predecessors of successors of the 'towards' state *)

	    for i=0 to maxind do
	      if t.(i).(towards) then (* i is a successor of 'towards' *) 
		for j=0 to maxind do
		  (* we add all pred of 'towards' to i *)
		  t.(i).(j)<- t.(i).(j) || t.(towards).(j)
		done
	    done;
	    lfrom:= List.tl !lfrom
	  end
      done;
      let cycle= ref [] in
	(* states on the diagonal (of filled states only are all on a common cycle
	   to merge, only if the Common.mergeloop flag is set *)
	if !(Common.merge_loop) then
	  List.iter 
	    (function i ->
	       if t.(i).(i) 
	       then 
		 begin
		   cycle:= i::!cycle;
		   t.(i).(i)<-false  
		 end
	       else ())
	    e.!efilled;
	!cycle
    end


(* Merging of two states and suppression of the 'from' state in epsilon table *)

let merge_states (from: state) (towards: state) (e: epsilontable) (mt: state_merge_table): unit=
  begin
    Hashtbl.add mt from towards;
    state_remove [from] e
  end

(* Propagation of mergings on epsilon. Results into a list of state merging
   (normalized) to perform on other transitions. s is the set of states that
   is built by the function and contains all states that will belong to the
   result. In the result they will be associated with the associated renamed state
   in the final mt *)


let rec propagate_merges (e: epsilontable) (l: (state * state) list) (s: stateset) (mt: state_merge_table): stateset=
  match l with 
      [] -> s
    | (q1,q2)::rem ->
	if q1=q2 then propagate_merges e rem s mt
	else 
	  let newq1,newq2= (merged_value mt q1),(merged_value mt q2) in
	    if q1=newq1 
	    then
	      if newq2=q1 (* cycle in preparation *)
	      then propagate_merges e rem s mt
	      else (* 1) *)
		let new_merges= Common.remove newq2 (merge_matrix e [q1] newq2) in
		  begin
		    merge_states q1 newq2 e mt;
		    propagate_merges e
		      (List.append 
			 (List.map (function elt -> (elt,newq2)) new_merges)
			 rem)
		      (Stateset.add q1 s)
		      mt
		  end
	    else
	      if newq1=q2 (* cycle in preparation *)
	      then propagate_merges e rem s mt
	      else
		if newq2=newq1 (* nothing to do, merge is already done *)
		then propagate_merges e rem s mt
		else 
		  if newq2=q2 
		  then (* symetrical case w.r.t. 1) *)
		    let new_merges= Common.remove newq1 (merge_matrix e [q2] newq1) in
		      begin
			merge_states q2 newq1 e mt;
			propagate_merges e
			  (List.append 
			     (List.map (function elt -> (elt,newq1)) new_merges)
			     rem)
			  (Stateset.add q2 s)
			  mt
		      end
		  else (* q1 and q2 are already renamed but different newq2/=newq1 *)
		    let new_merges= Common.remove newq2 (merge_matrix e [newq1] newq2) in
		      begin
			merge_states newq1 newq2 e mt;
			propagate_merges e
			  (List.append 
			     (List.map (function elt -> (elt,newq2)) new_merges)
			     rem)
			  (Stateset.add q1 (Stateset.add newq1 s))
			  mt
		      end


		
(* For a key a in an hash table t, add b to the list associated to a. *)

let myaddtable a b t=
  try(
    let l= Hashtbl.find t a in
      Hashtbl.replace t a (b::l)
  ) with Not_found -> Hashtbl.add t a [b]


(* merging in epsilon: saturates the epsilon relation w.r.t. merging and 
   results in a new merge policy. i.e. q1 -> q2, q3 -> q1, if q2=q3 then q2=q3=q1
   if Common.mergeloop is set *)  

let merge_epsilon (e: epsilontable) (me: (state * state) list) (mt: state_merge_table): ((state list)*state)list=
  let states_to_merge= propagate_merges e me Stateset.empty mt in
  let tbl= Hashtbl.create 10 in
  let statelist=
    begin
      (* For every state q1 renamed by q2 we add into entry q2, q1 to the list
	 of source states. q2 is the state obtained after renaming of q1. *)
      Stateset.iter (function s -> 
		       let news= merged_value mt s in
			 myaddtable news s tbl) states_to_merge;
      Common.hashtbl_to_list tbl
    end 
  in
    (* We produce a liste associating to each state list the destination state *)
    List.map (function (a,b) -> (b,a)) statelist


(* Simplify the symbolic version of epsilon using the state_merge_table
   t contains the new names of states. s is the state to rename *)

let rec simplify (e: epsilontable) (t: state_merge_table) (s: state): unit=
  let mv= merged_value t s in
    if mv=s then ()
    else 
      let pred= get_direct_pred s e in
      let succ= get_direct_succ s e in
	begin
	  (* merge succ and pred table of s into those of mv *)
	  Hashtbl.replace (e.dsucc) mv 
	    (Stateset.union  
	       succ (get_direct_succ mv e));
	  Hashtbl.replace e.dpred mv
	    (Stateset.union
	       pred (get_direct_pred mv e));

	  (* rename s by mv in all successors sets of predecessors of s *)
	  Stateset.iter 
	    (function s2 ->
	       let succ2= get_direct_succ s2 e in
(* to suppress loops on states
   let newsucc2= Stateset.remove mv (state_set_replace s mv succ2) in *)
 	       let newsucc2= (state_set_replace s mv succ2) in 
		 if succ2=newsucc2 then ()
		 else 
		   if Stateset.is_empty newsucc2 (* trivial epsilon transition *)
		   then Hashtbl.remove e.dsucc s2
		   else Hashtbl.replace e.dsucc s2 newsucc2)
	    pred;
	  (* rename s by mv in all predecessors sets of successors of s *)
	  Stateset.iter 
	    (function s2 ->
	       let pred2= get_direct_pred s2 e in
(* to suppress loops on states
	       let newpred2= Stateset.remove mv (state_set_replace s mv pred2) in *)
	       let newpred2= (state_set_replace s mv pred2) in
		 if pred2=newpred2 then ()
		 else 
		   if Stateset.is_empty newpred2 (* *)
		   then Hashtbl.remove e.dpred s2
		   else Hashtbl.replace e.dpred s2 newpred2)
	    succ
	end
    

(* Merging on transitions, returns the list of done mergings *)
(* modifies mt *)
let merge (d: delta) (mt: state_merge_table) (merge_list: (state * state) list): 
  ((state list) * state) list=
  (* merging of states in epsilon may add some merging. i.e. q1 -> q2, q3 -> q1, if
     q2=q3 then q2=q3=q1 *)

  (* fixpoint_mp is the transitive closure of merge_list: for every cycle of 
     states a canonical state name is chosen (q1, q2..) -> q0 for instance, for 
     merging states q1, q2, ... into q0. Merges are applied on epsilon and on mt *)
  let fixpoint_mp= merge_epsilon d.epsilon merge_list mt in
  let _ = if (!Common.verbose>3) then
    print_string (List.fold_right (function (sl,s) -> function str ->
					   "Fixpoint mp="^(state_list_to_string sl)^
					   "->"^
					   (State.to_string s)^"\n")
			  fixpoint_mp "") 
  in
    begin
      merge_trans_dest_list d mt fixpoint_mp;
      merge_trans_parameter d.trans d.satable mt fixpoint_mp;
      List.iter (function (from,towards) ->
		   List.iter (function s -> simplify d.epsilon mt s) from)
	fixpoint_mp;
      if (!Common.verbose>3) then 
	print_string ("State merge table= "^(merge_table_to_string mt)^"\n");
      fixpoint_mp
    end


(* general adding of a transtion to a delta. Returns false if the transition
   is already present in the table. The corresponding states are supposed to
   be already present in delta.epsilon. In case of an epsilon transition
   if the result is true, the list contains the list of states to be merged
   by adding the transition. It also performs the merging, since otherwise
   the next adding may add the states remaining on the diagonal of the 
   epsilon matrix to the new merging!

   Also update the stmt parameter of delta!  
   WARNING: it does not apply merging on final state set!
*)
    
let add (t: external_trans) (d: delta): (bool* ((state list * state) list))= 
  (* States of t have to be already defined in d *)
  let _= 
    IFDEF TEST THEN
      match t with
	  Trans((sym,sl,dest)) -> 
	    asserteq ((List.for_all (mem_state d) sl) &&
		      (mem_state d dest)) true "States of trans should belong to delta!"
	| Epsilon((q1,q2)) -> 
	    asserteq ((mem_state d q1) &&
		      (mem_state d q2)) true "States of trans should belong to delta!"
	      ELSE ()
	      END
  in
  match t with
      Trans((sym,sl,dest)) -> (add_regular_trans (sym,sl,dest) d, [])
    | Epsilon((q1,q2)) -> 
	match add_epsilon (q1, q2) d.epsilon with
	    Ok([]) -> 
	      begin
		(* *)
		direct_add d.epsilon q1 q2;
		(true,[])
	      end
	  | Ok([_]) -> failwith "Transitions:add: add_epsilon cannot return only one state!"
	  | Ok(towards::from) ->
	      (* If adding an epsilon leads to a new cycle, we merge *)
	      (* We first remove cycles that may interfere *)
(*	      let _= remove_cycles (towards::from) d.epsilon in *)
	      let _= direct_add d.epsilon q1 q2 in
	      let l= List.map (function elt -> (elt,towards)) from in
	      let res= merge d d.stmt l in
		begin
		  List.iter 
		    (function s -> simplify d.epsilon d.stmt s)
		    (towards::from);

		(* it is supposed here that merge (stabilizing the epsilon
		merging cannot stabilize to anything else than towards::from 
		since otherwise any additional state should be already in 
		the cycle, hence in towards::from *)
		  (true, res)
		end
	  | Present -> (false,[])
	      

(* This function is used to propagate epsilon on regular list of transitions
   so as to produce a t.a. without epsilon *)

let rec propagate_epsilon (l: external_symtrans list) (d: delta) acc: external_symtrans list=
  match l with 
      [] -> acc
    | (f,sl,dest)::rem -> let succ=get_succ dest d in
      let nextacc= List.fold_left (function l ->function s -> (f,sl,s)::l) ((f,sl,dest)::acc) succ 
      in
	propagate_epsilon rem d nextacc
	
	
  


(* ------------------------------------------------ Tests ----------------------------------------------*)


IFDEF TEST THEN 
  print_string "Transitions.ml\n\n" 
END

IFDEF TEST THEN 
  print_string "Test trans_to_string\n\n" 
END

IFDEF TEST THEN 
         let d= new_delta()
	 let s0= new_state d 
	 let s1= new_state d
	 let s2= new_state d

	 let t1= ("f", [s1;s2], s0)     (* f(q1, q2) -> q0 *)
	 let t2= ("g", [s1; s0], s2)  (* g(q1, q0) -> q2 *)
	 let t2b= ("g", [s2; s2], s2)  (* g(q2, q2) -> q2 *)
	 let t6= ("a", [], s0)        (* a -> q0 *)
	 let t6b= ("a", [], s2)        (* a -> q2 *)
	 let t5= ("f", [s2;s0], s1)   (* f(q2, q0) -> q1 *)
END

IFDEF TEST THEN
let _= add (Trans t1) d in
let _= add (Trans t2) d in
let _= add (Trans t6) d in
  begin
    (* should be equivalent (modulo ordering) to
       
       g(q1,q0) -> q2
       a -> q0
       f(q1,q2) -> q0
       
    *)
    print_string (to_string d);
    asserteq (to_string d) "g(q1,q0) -> q2\na -> q0\nf(q1,q2) -> q0\n\n" "";
    print_string ("To string check trans d= "^(to_string_check d)^"\n");
  end
ENDIF



(* the add function should avoid to add covered transitions *)

IFDEF TEST THEN
print_string "Test add and mem\n\n"
END

IFDEF TEST THEN 
  let trans3= new_transtable ()
  let trans4= new_transtable ()
  let trans5= new_transtable ()
END


IFDEF TEST THEN 
begin
  print_string "Transitions should not be added twice\n";
  asserteq (add_trans t1 d.trans) false "";
  print_string (trans_to_string d.trans);
  asserteq (trans_to_string d.trans) "g(q1,q0) -> q2\na -> q0\nf(q1,q2) -> q0\n" "";
  assertneq (add_trans t6 trans3) false "";
  asserteq (add_trans t6 trans3) false "";
  asserteq (trans_to_string trans3) "a -> q0\n" "";
  assertneq (add_trans t1 trans3) false "";
  assertneq (mem t1 trans3) false "";
end
END


(* adding the 3 first transitions should produce the same table with
add and add_fast *)

IFDEF TEST THEN
print_string "Test add and add_fast\n\n"
END

IFDEF TEST THEN
let _=add_fast t1 trans4 in
let _=add_fast t2 trans4 in
let _= add_fast t6 trans4 in
let _= add_trans t1 trans5 in
let _= add_trans t2 trans5 in
let _= add_trans t6 trans5 in
asserteq trans4 trans5 "";
END


IFDEF TEST THEN
print_string "Test epsilon transtitions\n"
  END

IFDEF TEST THEN
  let d2= new_delta()
  let d2b= new_delta()
  let s0= new_state d2 
  let s1= new_state d2
  let s2= new_state d2
  let s0b= new_state d2b 
  let s1b= new_state d2b
  let s2b= new_state d2b
	    END

	    IFDEF TEST THEN
	    begin
	      print_string "add_epsilon:1\n";
	      let a= add_epsilon (s1,s2) d2.epsilon in (* add q1 -> q2 *)
		asserteq a (Ok[]) "";
		print_string "add_epsilon:2\n";
		let b= add_epsilon (s2,s0) d2.epsilon in (* add q2 -> q0 *)
		  asserteq b (Ok[]) "";
		  let _= add_fast t2 d2.trans in
		  let _= add_fast t6 d2.trans in ();
		    (* predecessors of s0 are s2 and s1 *)
		    asserteq (is_pred s1 s0 d2) true "";
		    asserteq (get_pred s0 d2) [s2;s1] "";  
		    (* There should be a transition g(q1, q0) -> q2 -> q0 *)
		    asserteq (get_trans s0 "g" d2) [t2] "";
		    asserteq (get_trans s0 "a" d2) [t6] "";
		    ignore(add (Epsilon(s1b,s2b)) d2b);
		    ignore(add (Epsilon(s2b,s0b)) d2b);
		    ignore(add (Epsilon(s1b,s0b)) d2b);
		    ignore(add (Trans(t2)) d2b);
		    ignore(add (Trans(t6)) d2b);
		    (* add_epsilon does not perform the merge *)
		    let c= add_epsilon (s0,s1) d2.epsilon in  (* add q0->q1 *)
		      print_string ("d2="^(to_string d2));
		      (* Cycles are removed from the epsilon table but merges to
			 performed are given as a result (c here) *)
		      asserteq (to_list d2) [(Trans t2);(Trans t6);
					     (Epsilon (0,1));
					     (Epsilon (0,2));
					     (Epsilon (1,0));
					     (Epsilon (1,2));
					     (Epsilon (2,0));
					     (Epsilon (2,1))] "to_list"; 
		      asserteq c (Ok[s2;s1;s0]) "1) Produce equality between the 3";
		      (* Be careful d2 is finally not a valid epsilon table!
			 it contains cycles *)
		      print_string ("\nBefore add epsilon d2b="^(to_string_check d2b));
		      let c2= add (Epsilon (s0,s1)) d2b in (* add q0 -> q1 *)
			(* with add the merging should be performed and we should
			   obtain a valid epsilon table (empty here everything is
			   merged into q2) *)
			print_string ("After add q0 -> q1, d2b="^(to_string d2b));
			print_string ("State merge table of d2b="^(merge_table_to_string d2b.stmt)^"\n");
			asserteq c2 (true,[([s1;s0],s2)]) "2) Produce equality between the 3";
			asserteq (to_list d2b) [(Trans t2b);(Trans t6b)] "to_list2";
			let _= add (Epsilon (s0,s1)) d in
			let _= add (Trans t5) d in
			  print_string (to_string d);
			  print_string "get_trans_sym_constant\n";
			  asserteq (get_trans_sym_constant "a" d) [t6] "";  
			  print_string "get_trans_sym_parameter\n";
			  asserteq (get_trans_sym_parameter "f" s2 d) [t5;t1] ""
	    end
	    END

	    IFDEF TEST THEN
	      let mt= (new_merge_table ())
	      let s3= new_state d
	      let ml= [([s1;s3],s0)] 
	      let d3= copy d
END

IFDEF TEST THEN
		print_string ("copy and delta_equal\n");
		asserteq (delta_equal d d) true "";
		asserteq (delta_equal d d3) true "";
		print_string ("d3= "^(to_string_check d3)^"\n");
		asserteq (to_string d) (to_string d3) "Not necessarily, but maybe";
		merge_states  s1 s0 d3.epsilon d3.stmt;
		merge_states  s3 s0 d3.epsilon d3.stmt;
		merge_trans_parameter d3.trans d3.satable d3.stmt ml;
		
		asserteq (delta_equal d d3) false "d and d3 should be no longer equivalent"; 
		print_string ("merge_trans_parameter\n");
		print_string ("d avant merge (parametre) de q1,q3 vers q0=\n"^(to_string d)^"\n\n");
 	        print_string ("d apres merge (parametre) de q1,q3 vers q0=\n"^(to_string d3)^"\n\n");
		print_string "add_merged et merged_value\n";
		print_string ("replace "^(State.to_string s1)^" and "^(State.to_string s3)^" by: "^(State.to_string s0)^"\n");
		asserteq (merged_value d3.stmt s1) s0 "";
		asserteq (merged_value d3.stmt s3) s0 "";
		asserteq (merged_value d3.stmt s0) s0 "";
		asserteq (merged_value d3.stmt s2) s2 "";
		print_string "propagate_epsilon\n";
		print_string ("d avant propagate epsilon\n"^(to_string d)^"\n");
		let new_trans= propagate_epsilon 
				 (List.map (function t -> 
					      match t with 
						  Trans(t2) -> t2
						| _ -> failwith "Automaton:unepsilon:Cannot be an Epsilon trans") (trans_to_list d.trans))
				 d [] in
		  print_string ("Nouvelles transitions ares epsilon propagate="^(symlist_to_string new_trans));
		  asserteq (symlist_to_string new_trans) "f(q1,q2) -> q1\nf(q1,q2) -> q0\na -> q1\na -> q0\nf(q2,q0) -> q1\ng(q1,q0) -> q2\n" "Transitions to q0 are added to q1"
	
END


IFDEF TEST THEN 
         let d3= new_delta()
	 let d4= new_delta()
(*	 let d5= new_delta() *)
	 let d6= new_delta()

	 let ds0= new_state d3 
	 let _= new_state d4
	 let _= new_state d6
	 let ds1= new_state d3
	 let _= new_state d4
	 let _= new_state d6
	 let ds2= new_state d3
	 let _= new_state d4
	 let _= new_state d6
	 let ds3= new_state d3
	 let _= new_state d4
	 let _= new_state d6
	 let ds4= new_state d3
	 let _= new_state d6
	 let ds5= new_state d3


	 let dt1= ("a", [], ds3)            (* a -> q3 *)
	 let dt2= ("f", [ds1;ds2], ds3)     (* f(q1, q2) -> q3 *)
	 let dt3= ("f", [ds1; ds5], ds3)    (* f(q1, q5) -> q3 *)
	 let dt4= ("f", [ds1; ds2], ds4)    (* f(q1, q2) -> q4 *)
	 let dt5= ("g", [ds2; ds4], ds5)    (* g(q2, q4) -> q5 *)
	 let dt6= ("a", [], ds5)            (* a -> q5 *)
	 let dt7= ("b", [], ds5)            (* b -> q5 *)
	 let dt8= ("f", [ds1; ds4], ds5)    (* f(q1, q4) -> q5 *)

	 let dt9= ("f", [ds1; ds3], ds3)    (* f(q1, q3) -> q3 *)
	 let dt10= ("g", [ds2; ds3], ds3)   (* g(q2, q3) -> q3 *)
	 let dt11= ("b", [], ds3)           (* b -> q3 *)

	 let mt2= new_merge_table ()
	 let mp= [(ds4,ds5);(ds5,ds3)]
(*	 let mp2= merging_equation_closure [(ds1,ds2);(ds3,ds4)] [] *)
(*	 let mp2= merging_equation_closure [(ds1,ds2);(ds3,ds4)] [] *)
END

IFDEF TEST THEN	
  begin
    ignore (add (Trans dt1) d3);
    ignore (add (Trans dt2) d3);
    ignore (add (Trans dt3) d3);
    ignore (add (Trans dt4) d3);
    ignore (add (Trans dt5) d3);
    ignore (add (Trans dt6) d3);
    ignore (add (Trans dt7) d3);
    ignore (add (Trans dt8) d3);

(* d3 should be *)
(* a -> q3 *)
(* f(q1, q2) -> q3 *)
(* f(q1, q5) -> q3 *)
(* f(q1, q2) -> q4 *)
(* g(q2, q4) -> q5 *)
(* a -> q5 *)
(* b -> q5 *)
(* f(q1, q4) -> q5 *)

    ignore (add (Trans dt1) d4);
    ignore (add (Trans dt2) d4);
    ignore (add (Trans dt9) d4);
    ignore (add (Trans dt10) d4);
    ignore (add (Trans dt11) d4);

(* d4 should be *)
(* a -> q3 *)
(* f(q1, q2) -> q3 *)
(* f(q1, q3) -> q3 *)
(* g(q2, q3) -> q3 *)
(* b -> q3 *)

    ignore (add (Epsilon (ds1,ds3)) d6);
    ignore (add (Epsilon (ds4,ds2)) d6);

(* d6 should be *)
(* q1 -> q3, q4 -> q2 *)

    print_string ("before merge d3=\n"^(to_string_check d3));
    print_string ("merge q4 and q5 into q3 should give d4=\n"^(to_string_check d4));
(*    let d5= copy d3 in
    let _= add_merged mp mt2 in
    let _= merge_trans_dest_list d5 mt2 mp in
	print_string ("merge dest only on d3=\n"^(to_string_check d5)^"\n") *)
  end
END


IFDEF TEST THEN   
    let d5= copy d3 in
      begin
(*	print_string "\nmerging_equation_closure\n";
	asserteq (merging_equation_closure [(1,2);(1,3);(4,8);(2,6);(8,10);(7,9)] []) [([3; 1; 2], 6); ([4; 8], 10); ([9], 7)] "Closure of equations"; *)
(*	print_string "\nmerge_equivalence_classes\n"; *)
	(* asserteq (merge_equivalence_classes 2 [3;4] [] [(7,[6;8]);(4,[10;11]);(5,[12;13;14])]) [(2, [3; 4; 10; 11]); (7, [6; 8]); (5, [12; 13; 14])] "Merging of equivalence classes w.r.t. canonical states"; *)

	print_string "\nmerge general\n";
	asserteq (delta_equal d3 d4) false "d3 d4 Should not be equal before merge";
	ignore(merge d3 mt2 mp); (* merge q4 and q5 into q3 *)
	asserteq (delta_equal d3 d5) false "d3 and its copy d5 should not be equal after merge";
	print_string ("after merge d3=\n"^(to_string_check d3));
	(* Necessary to add and suppress q4 and q5 in d4 in order to have exactly the
	   same list of states (In the current implem destroyed states are not
	   considered again!) )*)
	ignore(new_state d4);
	ignore(new_state d4);
	state_remove [4;5] d4.epsilon;
	asserteq (delta_equal d3 d4) true "d3 d4 Should be equal after merge";  
(*	print_string "\nstabilize_epsilon\n";
	asserteq (stabilize_epsilon d6.epsilon mp2 [])  [([4; 1; 2], 3)] "q1=q2 and q3=q4 with q1->q3 and q4-> q2 should build a new cycle and thus merge more states" *)
      end 
END


IFDEF TEST THEN
  let d10= new_delta ()
  let s1= new_state d10
  let s2= new_state d10
  let s3= new_state d10
  let _= for i=1 to 97 do ignore(new_state d10) done (* Consume all states *)
  let _= ignore(add (Epsilon (s1,s2)) d10); ignore(add (Epsilon(s2,s3)) d10) 
  let e1= d10.epsilon 
  let ecopy= epsilontable_copy e1 
END

IFDEF TEST THEN
  let d11= new_delta()
  let s1= new_state d11
  let s2= new_state d11
  let s3= new_state d11
  let _= ignore(add (Epsilon (s1,s2)) d11)
  let set1= Stateset.add s2 Stateset.empty
  let set2= get_all_pred set1 d11
  let set3= Stateset.add s1 set1
END

IFDEF TEST THEN
    asserteq set2 set3 "get_all_pred on sets of states fails";
    asserteq set1 (state_set_replace s1 s2 set3) "replace replaces when state found";
    asserteq set1 (state_set_replace s1 s2 set1) "replace doesn't replace when state not found";
END

IFDEF TEST THEN
  let dbig= new_delta()
  let si= ref s1
  let sj= ref s2
  let ls= ref []
END

IFDEF TEST THEN
  print_string "Testing epsilon matrix automatic extension";

  (* Build a sequence of Common.max_states+4 states all related by epsilon
     transition, to check extension of epsilon matrix *)

  si:= (new_state dbig);
  ls:= !si::!ls;
  for i=0 to Common.max_state_symb+3 do
    begin
      sj:= (new_state dbig);
      ignore(add (Epsilon (!sj,!si)) dbig);
      ls:= !sj::!ls;
      si:= !sj;
    end
  done;
  (* We check that after extension all epsilon transitions and states
     are still there! *)

  si:= List.hd !ls;
  ls:= List.tl !ls;
  asserteq (mem_state dbig !si) true "Last added state is still there";
  List.iter (function sj ->
		 asserteq (mem_state dbig sj) true ("State "^(State.to_string sj)^" has desappeared");
	       asserteq (is_pred !si sj dbig) true 
		 ("Epsilon trans: "^
		  (external_to_string (Epsilon(!si,sj)))^
		  " has desappeared!");
	       si:=sj)
    !ls
END
